/*    
 *  Copyright 2000, 2001, 2002 Emmanuel Benazera  
 *
 *  This file is part of Koala.
 *
 *  Koala is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Koala is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Koala; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "Comprt.h"
#include <iostream>

using namespace std;

/* Functions for class Component */

//constructors
Comp::Comp() {
}
//for parsing...
Comp::Comp(string na) {
  name = na;
  state_variables = new vector<StateVar*>();
  constraints = new vector<Formula*>();
  component_transitions = new vector<Transition*>();
  atoms = new hash_map<const char*, Atom*, hash<const char*>, eqstr>();
  influences = new vector<influence>();
  variables = new vector<variable>();
  imported_comps = new vector<Comp*> ();
  atoms_number = 0;
  var_name = "";
}
//destructor.
Comp::~Comp() {
  purge(*state_variables);
  delete state_variables;
  state_variables = 0;
  purge(*constraints);
  delete constraints;
  constraints = 0;
  purge(*component_transitions);
  delete component_transitions;
  component_transitions = 0;
  delete atoms;
  delete influences;
  delete variables;
  delete imported_comps;
}
//Access to variables
//set the state_variables
void Comp::setStateVariables(vector<StateVar*>* sta) {
  state_variables = sta;
}
//return the vector of state variables
vector<StateVar*>* Comp::getStateVariables() {
  return state_variables;
}
//return a state variable, given its name.
StateVar* Comp::getStateVariable(string na) {
  for (unsigned int i=0; i<state_variables->size(); i++)
    if ((*state_variables)[i]->name == na)
      return (*state_variables)[i];
  return NULL;
}
//add a state variable to the vector of state variables
void Comp::addStateVar(StateVar* sta_var) {
  sta_var->setComp(this);
  state_variables->push_back(sta_var);
}
/* add a continuous variables. */
void Comp::addVariable(variable va) {
  variables->push_back(va);
}
//return all the atoms of the component.
vector<Atom*>* Comp::getAllAtoms() {
  vector<Atom*>* res = new vector<Atom*>();
  vector<Atom*>* temp = NULL;;
  for (unsigned int i=0; i<state_variables->size(); i++) {
    temp = (*state_variables)[i]->getAtomLinks();
    for (unsigned int j=0; j<temp->size(); j++)
      res->push_back((*temp)[j]);
  }
  delete temp;
  return res;
}
//return all current atoms.
/*vector<Atom*>* Comp::getAllCurrentAtoms() {
  vector<Atom*>* res = new vector<Atom*>();
  for (unsigned int i=0; i<state_variables->size(); i++)
    res->push_back((*state_variables)[i]->currentAtom);
    return res;
}*/
//compute nominal cover
void Comp::computeNominalCover() {
  int count = 0;
  for (unsigned int i=0; i<component_transitions->size(); i++) {
    if ((*component_transitions)[i]->type == "Nominal") {
      count++;
      nominal_cover = nominal_cover + (*component_transitions)[i]->proba;
    }
  }
  nominal_cover = (double)nominal_cover / (double)count;
}
//set the vector of feasible states
void Comp::setConstraints(vector<Formula*>* fea) {
  constraints = fea;
}
  //return the vector of feasible states
vector<Formula*>* Comp::getConstraints() {
  return constraints;
}
//add a Formula to feasible states
void Comp::addFormulaToCS(Formula* frm) {
  constraints->push_back(frm);
}

//add an influence
void Comp::addInfluence(influence inf) {
  influences->push_back(inf);
}

//set the vector of transitions
void Comp::setCompTransitions(vector<Transition*>* trans) {
  component_transitions = trans;
}
//return the vector of transitions
vector<Transition*>* Comp::getCompTransitions() {
  return component_transitions;
}
//add a Transition to the component's transitions
void Comp::addTransitionToC(Transition* trans) {
  component_transitions->push_back(trans);
}
//return the name of the component
const char* Comp::getName() {
  return name.c_str();
}
//set the name of the component
void Comp::setName(string na) {
  name = na;
}

/**
 * Tools & debug.
 */
void Comp::getTransitionsGuards() {
  for (unsigned int i=0; i<component_transitions->size(); i++) {
    if ((*component_transitions)[i]->proba == 1) {
      (*component_transitions)[i]->pre_at->guards_number++;
      (*component_transitions)[i]->modevar->guards_number++;
    }
    else {
      for (unsigned int j=i+1; j<component_transitions->size(); j++) {
	if ((*component_transitions)[i]->pre->eq((*component_transitions)[j]->pre)) {
	  (*component_transitions)[i]->pre_at->guards_number++;
	  (*component_transitions)[i]->modevar->guards_number++;
	}
      }
    }
  }
}

/**
 * debug & visu.
 */
//display the component properties
void Comp::showComp() {
  cout << "\n***************************************************" << endl;
  cout << "Comp : " << getName() << endl;
  cout << "***************************************************" << endl;
  cout << "nominal cover : " << nominal_cover << endl;
  cout << "States : \n";
  for (unsigned int i=0;i<state_variables->size();i++) {
    cout << (*state_variables)[i]->name << " - " << (*state_variables)[i]->getType() << " -" << endl;
    (*state_variables)[i]->showValues();
    (*state_variables)[i]->showCurrentState();
    /*if ((*state_variables)[i]->type == "Cond") {
      CondVar* cvtmp = (CondVar*)(*state_variables)[i];
      cout << cvtmp->condvar << endl;
      }*/
  }

  cout << endl;
/*  for (unsigned int i=0; i<variables->size(); i++) {
    ce_printVariable (stdout,(*variables)[i]);
  }
*/

  /*cout << endl;
  for (unsigned int i=0; i<variables->size(); i++) {
    printVariable ((*variables)[i]);
  }*/

  cout << "\n---------------------------------------------------";
  cout << "\nConstraints : \n";
  /*for (unsigned int i=0; i<constraints->size(); i++)
    (*constraints)[i]->showFormula();*/
  /*for (unsigned int i=0; i<constraints->size(); i++)
    (*external_clauses)[i]->showClause (0);*/ /* NO ! forward declaration... */
  cout << "---------------------------------------------------";
  cout << "\nInfluences : \n";
  for (unsigned int i=0; i<influences->size(); i++) {
    printInfluence ((*influences)[i]);
  }
  cout << "---------------------------------------------------";
  cout << "\nTransitions : \n";
  for (unsigned int i=0; i<component_transitions->size(); i++) {
    if (component_transitions == NULL) cout << "NULLL\n";
    (*component_transitions)[i]->showTransition();
  }
  cout << "***************************************************\n";
}
