/*    
 *  Copyright 2000, 2001, 2002 Emmanuel Benazera  
 *
 *  This file is part of Koala.
 *
 *  Koala is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Koala is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Koala; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "StateVar.h"
#include <iostream>
#include <iterator>

using namespace std;

/* Functions for class StateVar*/

//constructors
StateVar::StateVar() {
  values = set<string>();
  name = "";
  current_state = new string("");
  comp_lk = NULL;
  atoms_links = new vector<Atom*>();
  internal_clauses = new vector<Clause*>();
  guards_number = 0;
}
StateVar::StateVar(string ca) {
  name = ca;
  default_state = "";
  current_state = new string("");
  comp_lk = NULL;
  real_value = "";
  observable = false;
  atoms_links = new vector<Atom*>();
  internal_clauses = new vector<Clause*>();
}
StateVar::StateVar(string na, set<string> se) {
  name = na;
  values = se;
  comp_lk = NULL;
  real_value = "";
  observable = false;
  atoms_links = new vector<Atom*>();
  internal_clauses = new vector<Clause*>();
}
StateVar::StateVar(string na, set<string> se, string defa) {
  name = na;
  values = se;
  setDefault(defa);
  current_state = new string("");
  setCurrentState(defa);
  comp_lk = NULL;
  real_value = "";
  observable = false;
  atoms_links = new vector<Atom*>();
  internal_clauses = new vector<Clause*>();
}
StateVar::StateVar(string na, set<string> se, string defa, Comp* co) {
  name = na;
  values = se;
  setDefault(defa);
  current_state = new string("");
  setCurrentState(defa);
  comp_lk = co;
  type = "";
  real_value = "";
  atoms_links = new vector<Atom*>();
  observable = false;
  internal_clauses = new vector<Clause*>();
}
StateVar::StateVar(string na, set<string> se, string defa, Comp* co, bool obs) {
  name = na;
  values = se;
  setDefault(defa);
  current_state = new string("");
  setCurrentState(defa);
  comp_lk = co;
  type = "";
  real_value = ""; //do not set the 'real' value...
  atoms_links = new vector<Atom*>();
  observable = obs;
  internal_clauses = new vector<Clause*>();
}
//destructor.
StateVar::~StateVar() {
  delete internal_clauses;
  internal_clauses =0;
  delete atoms_links;
  atoms_links = 0;
  delete current_state;
  current_state = 0;
}
//test if the state variable is equal to another one... (name comparison).
/*bool StateVar::eq(StateVar* st) {
  if (name == st->name)
  return true;
  else return false;
}*/
//set the values
void StateVar::setValues(set<string> se) {
  values = se;
}
//return the set of values
set<string> StateVar::getValues() {
  return values;
}
//add a value to the set.
void StateVar::addValue(string str) {
  values.insert(str);
}
//set the default (start) value
void StateVar::setDefault(string defa) {
  if (inValues(defa))
    default_state = name + "=" + defa;
  else cout << "StateVar --> Wrong default value";
}
//get the default value
string StateVar::getDefault() {
  return default_state;
}
//return the current state
string StateVar::getCurrentState() {
  return *current_state;
}
//set the current_state
void StateVar::setCurrentState(string sta) { 
  *current_state = sta;
}
//set the name of the current var 
void StateVar::setName(string str) {
  name = str;
}
//return the name of the current var (with the name of the component for unicity)
string StateVar::getName() {
  if (comp_lk!=NULL)
    return name;
  else
    return name; //return the name anyway
}

void StateVar::setSymbName (string str) {
  symb_name = str;
}
string StateVar::getSymbName () {
  return symb_name;
}
const char* StateVar::getCharName() {
  return name.c_str();
}
//set a link to a component
void StateVar::setComp(Comp* co) {
  comp_lk = co;
}
//test if a value is in the values set
bool StateVar::inValues(const string& val) {
  set<string> temp;
  temp = set<string>();
  temp.insert(val);
  return includes(values.begin(), values.end(), temp.begin(), temp.end());
}
unsigned int StateVar::addAtomToLinks(Atom* at) {
  atoms_links->push_back(at);
  return atoms_links->size() - 1;
}
void StateVar::showValues() {
  cout << "\t\t" << name << endl;
  cout << "\t\tType: " << type << endl;
  cout << "\t\tValues: " << endl;
  cout << "\t\t{ ";
  copy(values.begin(), values.end(), ostream_iterator<string>(cout, "  ")); 
  cout << "}\n" << endl;
}
void StateVar::showCurrentState() {
  cout << name << " --->\t\t " << *current_state << "\n";
}
bool StateVar::matchObs() {
  if (observable && real_value!="") 
    return (*current_state==real_value);
  else return true; //because there can be no discrepency.
}
