/*    
 *  Copyright 2000, 2001, 2002 Emmanuel Benazera  
 *
 *  This file is part of Koala.
 *
 *  Koala is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Koala is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Koala; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* State variable */

#ifndef STATEVAR_H
#define STATEVAR_H

#include <string>
#include <set>
#include <algorithm>
#include <vector>
//#include <iostream>
//#include <iterator>
/*#include <hash_map.h>*/
#include "../system/purge.h"

using namespace std;

class Comp;
class Atom;
class Formula;
class Clause;

/**
 * \class StateVar
 * \brief A StateVar (state variable) is a discrete variable over a finite domain.
 *
 * A StateVar has discrete values that are <Atom> and belongs to a component (<Comp>).
 */

class StateVar {

 public:
   set<string> values;
   string type;
   string name;
   string symb_name;
  string* current_state;
  string default_state;
  Comp* comp_lk;                  /**< link to component the state variable belongs to */
  vector<Atom*>* atoms_links; /**< references to the atoms of the state variable */
  Atom* defaultAtom; /**< default atom as the default value of the state variable */
  bool observable; /**< if the state variable is measured */
  string real_value; //sensor value.
  vector<Clause*>* internal_clauses; /**< clauses that are internal to the state variable */
  int guards_number; /**< number of transitions guards related to this state variable (modevar) */

 public:
  StateVar();
  StateVar(string ca);
  StateVar(string name, set<string> se);
  StateVar(string name, set<string> se, string defa);
  StateVar(string name, set<string> se, string defa, Comp* co);
  StateVar(string name, set<string> se, string defa, Comp* co, bool obs); //define also observability.
  ~StateVar();
  /*bool eq(StateVar* st);*/ //test if the state variable is equal to another one... (name comparison).
  void setValues(set<string> se);
  set<string> getValues();
  void addValue(string str);
  void setDefault(string defa);
  string getDefault();
  string getCurrentState();
  void setCurrentState(string sta);
  void setName(string str);
  string getName();
  void setSymbName (string str);
  string getSymbName ();
  const char* getCharName();
  void setComp(Comp* co);
  bool inValues(const string& val);  //test if value is in the domain of the state variable
  unsigned int addAtomToLinks(Atom* at);  //return the current valid atom.
  void showValues();
  void showCurrentState();
  bool matchObs();

/* inline methods */
/**
 * @return the type of the variable (Mode, Command, Dependent, Observable).
 */
inline string StateVar::getType() {
  return type;
}
/**
 * @return the component the state variable belongs to.
 */
inline Comp* StateVar::getComp() {
  return comp_lk;
}
/**
 * @return the set of <Atom> the state variable is made of.
 */
inline vector<Atom*>* StateVar::getAtomLinks() {
  return atoms_links;
}
};

#endif
