
/*    
 *  Copyright 2000, 2001, 2002 Emmanuel Benazera  
 *
 *  This file is part of Koala.
 *
 *  Koala is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Koala is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Koala; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef TRANSITION_H
#define TRANSITION_H

#include "StateFormula.h"
#include "Formula.h"
#include "ModeVar.h"
#include "../system/purge.h"
#include "MapFunc.h"
#include <string>
#include <vector>
#if (__GNUC__ >= 3)
#include <ext/hash_map>
#else
#include <hash_map.h>
#endif

#if (__GNUC__ >= 3)
using __gnu_cxx::hash_map;
using __gnu_cxx::hash;
#else
using std::hash_map;
using std::hash;
#endif

/**
 * \class Transition.
 * \brief Transition between compoents modes : StateFormula =>Next StateFormula.
 */

class Transition {

 public:
  int trans_id;  /**< id of the transition */
  string type;   /**< type of the transition : nominal/fail */
  ModeVar* modevar;  /**< mode variable attached to the transition. */ 
  Atom* pre_at; /**< atom attached to the pre of the transition. */
  Atom* post_at; /**< atom attached to the post of the transition. */
  double proba;  /**< probability of the transition */
  double weight;
  double proba_guard; /**< probability normalized with number of modevar transition guards. */
  double cost;    /**< cost attached to the transition */
  StateFormula* pre;  /**< premisses */ 
  vector<StateFormula*>* compiled_pre; /** premisses replaced by their prime implicants*/
  StateFormula* post; /**< post changes */
  int time_length;   /**< time length of the transition (= st_post->size()) */
  hash_map<int, StateFormula*, hash<int> >* st_post; /**< for multiple commands output */
  vector<Formula*>* post_forms; /* post can contain formulas... the operator between both formulas and stateformulas is "and" */
  bool command_neq;  /**< negation on commands. */
  bool idle; /**< idle flag. */
  bool commanded;  /**< flag for commanded transitions. */
  vector<MapFunc*>* mf;  /**< mapping function */
  int compiled_num;  /**< compilation transition number */

 public:
  Transition(StateFormula* pr, StateFormula* po, string ty);
  Transition(StateFormula* pr, StateFormula* po, string ty, double prob);
  Transition(StateFormula* pr, StateFormula* po, string ty, double prob, bool idl);
  Transition(StateFormula* pr, StateFormula* po, string ty, double prob, double co, bool idl);
  Transition(StateFormula* pr, hash_map<int, StateFormula*, hash<int> >* s_post, string ty, double prob);
  ~Transition();
  void setID(int id);
  string getType(); /**< return the type of the transition : nominal/failed */
  void addRefToAtoms(); /**< reference the transition to its atoms */
  void addFormulaToTrans(Formula* form); /**< add a formula to the transition */
  void showTransition();
};

#endif
