/* koala-exec.c */

#include "koala-exec.h"
#include "../caen/caen-face.h"

//#define VERB_EXEC
#define VERB_PERF   /* performances verbose */

long accuTime=0, done=0, simu_multi=0, rt_num=0, disp_num=0;
double simTime, totalTime, precision;
char inputFile[40], inLogFile[40], measureFile[80], displayFile[40], scenarFile[40],
  modesFile[40], diagFile[40], auxFile[80], diagOut[40000];
FILE *dataIn, *logS, *dataOut, *measIn, *diagIn, *modesOut;
bool discrete = 0;
bool disc_init = 0;


/* line commands, for the control loop */
ceCom cde [] = {
  {"quit", ceQuit, "Quit exec"},
  {"q", ceQuit, "Shortcut to quit"},
  {"help", ceHelp, "Get Help"},
  {"#", ceComment, "Comment line"},
  {"detect", ceDetect, "Prediction and default detection"},
  {"gnuplot", ceGnup, "Display results"},
  {"gp", ceGnup, "Shortcut to gnuplot"},
  {"init", ceInit, "Run initializations"},
  {"print", cePrint, "Print data structures"},
  {"p", cePrint, "Shortcut to print"},
  {"simul", ceSimul, "Simulation"},
  {"disc", ceDisc, "Run discrete only models"},
  {"system", ceSystem, "Issue a shell command"},
  {"debug", ceDebug, "Debug mode"},
  {(char *) NULL, (LFunction *) NULL, (char *) NULL}
};

/* debug mode commands */
ceCom dde [] = {
  {"exit", deExit, "Return to normal mode"},
  {"x", deExit, "Shortcut to exit"},
  {"help", deHelp, "Get Help"},
  {"print", dePrint, "Print data structures"},
  {"p", dePrint, "Shortcut to print"},
  {(char *) NULL, (LFunction *) NULL, (char *) NULL}
};

static char *ligne=(char *)NULL;

multimap<long, runtime, less<long> >* nrt;                                 /* new runtimes */

int main (int argc, char *argv[]) {
  char *s, *t;
  ceCom *pc;
  char *migne;
  long k, iarg=0;
  int pos=0;
  hybNT = 0;

  cout << "appliName: " << model_name << endl;
  sMM = new Scheduler(model_name);
  startinit(sMM); /* model init */
  cout << "model successfully loaded.\n";
  sMM->rtInit(); /* itms init */
  /*if (h_pi)
    sMM->readPI (); */     /* load compiled prime implicants when possible */
  if (! isNull (koala_rt))  /* koala is its discrete mode */
    localInit(koala_rt, sMM); /* local vars init */
  else discrete = 1;
  ligneInit ();
  strcpy (auxFile, argv[iarg++]);                             /* runtime name */
  strcpy (inputFile, "inp/");
  strcpy (scenarFile, "inp/");
  strcpy (measureFile, "obs/");
  strcpy (displayFile, "disp/");
  strcpy (diagFile, "");
  strcpy (modesFile, "");
  
  if (discrete == 0) {
    cerr << "inputs: " << getRTNInput (koala_rt) << endl;
    if (getRTNInput (koala_rt)) {
      if (! isNull (argv[iarg])) {
	strcat (inputFile, argv[iarg++]);
	pos = string(inputFile).find(".");   /* to be changed ? */
	/*appliName = string(inputFile).substr(0, pos);*/
      }
      else {
	fprintf (stderr, "Not enough parameters (expecting input file) !\n");
	exit (0);
      }
    }
    if (getRTNMeasure (koala_rt)) {
      if (! isNull (argv[iarg])) {
	strcat (measureFile, argv[iarg++]);
	cout << argv[iarg] << endl;
      }
      else {
	fprintf (stderr, "Not enough parameters (expecting measure file) !\n");
	exit (0);
      }
    }
    if (getRTNDisplay (koala_rt)) {
      if (! isNull (argv[iarg])) {
	strcat (displayFile, argv[iarg++]);
	cout << argv[iarg] << endl;
      }
      else {
	fprintf (stderr, "Not enough parameters (expecting display file) !\n");
	exit (0);
      }
    }
    if (! isNull (argv[iarg])) {
      strcat (scenarFile, argv[iarg++]);
      cerr << "scenario file: " << scenarFile << endl;
      pos = string(scenarFile).find(".");     /* to be changed ? */
    }
  }
  else {  /* discrete: check for scenario file */
    cerr << "scenario file: " << argv[iarg] << endl;
    if (! isNull (argv[iarg])) {
      strcat (inputFile, argv[iarg++]);
      std::cout << "inputFile: " << inputFile << std::endl;
      pos = string(inputFile).find(".");     /* to be changed ? */
      /* appliName = string(inputFile).substr(0, pos);*/
    }
    else {
      fprintf (stderr, "Not enough parameters (expecting scenario file) !\n");
      exit (0);
    }
  }

  if (! isNull (argv[iarg]))    {       /* modes display file */
    strcpy (modesFile, argv[iarg++]);
    std::cout << "mode file: " << modesFile << std::endl;
  }
  else cout << "no mode files\n";
  
  s = auxFile;
  if (! isNull (t = strtok (auxFile, "/")))
    do
      s = t;
    while (! isNull (t = strtok (NULL, "/")));
  if (discrete == 0)
    fprintf (stderr, "Ready for executing %s with time step = %g;\n",
	     s, getRTTimeStep (koala_rt));
  fprintf (stderr, "Execution parameters: ");
  for (k=1; k<iarg; ++k) {
    fprintf (stderr, "%s", argv[k]);
    if (k < iarg-1)
      fprintf (stderr, ", ");
  }
  fprintf (stderr, ";\n");
  //strcpy (inLogFile, argv[0]);
  //strcat (inLogFile, ".in.log");
  
  /* control loop: reads command lines */
  while (! done ) {
    if (! isNull (migne = do_gets ("e> "))) {
      ligne = migne;                               /* para el caso #else GNU */
      if (*ligne == '\0')                             /* linea vacia         */
	continue;
      if (! (isatty (0) && isatty (1)))
	fprintf (stderr, "%s\n", ligne);              /* solo si redireccion */
      if (*ligne == '#')                              /* comentario          */
	continue;
      stripWhite (ligne);     /* tira los ' ' y '\t' del principio y del fin */
      t = strtok (ligne, " ");
      pc = whichCom (cde, t);
      if (isNull (pc)) {
	fprintf (stderr, "%s: Unknown Koala executer subcommand.\n", t);
	continue;
      }
      t = strtok (NULL, "");
      stripWhite (t);
      (*(pc->func)) (t);
    }
  }
  fprintf (stderr, "Leaving %s.\n", s);
  exit (0);
}


/* In GNU versions, use readline package */
void ligneInit (void) {

#ifdef HAVE_READLINE
  initialize_readline ();                             /* Bind our completer. */
#else /* GNU */
  ligne = (char *) malloc (81);
#endif /* GNU */
  return;
}


#ifdef HAVE_LIBREADLINE
/* Tell the GNU Readline library how to complete.  We want to try to complete
   on command names if this is the first word in the line, or on filenames
   if not. */

char *command_generator ();
char **ce_completion (const char *tex, int start, int end);

void initialize_readline (void) {
     
  /* Allow conditional parsing of the ~/.inputrc file. */
  rl_readline_name = "Koala";
  /* Tell the completer that we want a crack first. */
  /*rl_attempted_completion_function = ce_completion;*/
  return;
}
     
/* Attempt to complete on the contents of TEXT.  START and END show the
   region of TEXT that contains the word to complete.  We can use the
   entire line in case we want to do some simple parsing.  Return the
   array of matches, or NULL if there aren't any. */
char **ce_completion (const char *text, int start, int end) {
  char **matches=NULL;
     
  /* If this word is at the start of the line, then it is a command
     to complete.  Otherwise it is the name of a file in the current
     directory. */
  /*if (start == 0)
    matches = completion_matches (text, command_generator);*/
  return (matches);
}
     
/* Generator function for command completion.  STATE lets us know whether
   to start from scratch; without any state (i.e. STATE == 0), then we
   start at the top of the list. */
char *command_generator (char *text, long state) {
  static long list_index, len;
  char *name, *s=NULL;
     
  /* If this is a new word to complete, initialize now.  This includes
     saving the length of TEXT for efficiency, and initializing the index
     variable to 0. */
  if (!state) {
    list_index = 0;
    len = strlen (text);
  }
  /* Return the next name which partially matches from the command list. */
  while ((name = cde[list_index].name)) {
    list_index++;
    if (strncmp (name, text, len) == 0)
      s = name;
  }
  /* If no names matched, then return NULL. */
  return s;
}

/* Read a string, and return a pointer to it.  Returns NULL on EOF. */
char *do_gets (char *s) {
  /* If the buffer has already been allocated, return the memory
     to the free pool. */
  if (ligne != (char *)NULL) {
    free (ligne);
    ligne = (char *)NULL;
  }
  /* Get a line from the user. */
  ligne = readline (s);
  /* If the line has any text in it, save it on the history. */
  if (ligne && *ligne)
    add_history (ligne);
  return (ligne);
}

#else /* GNU */

char *do_gets (char *s) {
  long again;

  fprintf (stderr, "%s", s);
  do {
    again = 0;
    if (isNull (fgets (ligne, 81, stdin)))           /* EOF encontrado       */
      ligne = (char *) NULL;
    else {
      char *p;
      p = strtok (ligne, "\n");                      /* cambia '\n' por '\0' */
      if (isNull (p)) {
	again = 1;                                 /* ligne empieza por '\n' */
	continue;
      }
    }
  } while (again);
  return ligne;
}
    
#endif /* GNU */


/* Strip whitespace from the start and end of CE_STRING. (gnu) */
void stripWhite (char *s) {
  long i=0;

  if (! isNull (s)) {
    while (isspace (s[i])) /*whitespace (s[i]))*/
      i++;
    if (i)
      strcpy (s, s+i);
    i = strlen (s) - 1;
    while (i > 0 && isspace (s[i]))/* whitespace (s[i]))*/
      i--;
    s[++i] = '\0';
  }
  return;
}


/* Look up NAME as the name of a command, and return a pointer to that
   command.  Return a NULL pointer if NAME isn't a command name. */
ceCom *whichCom (ceCom *com, char *name) {
  long i;
  ceCom *pc=NULL;
     
  for (i = 0; com[i].name; i++)
    if (strcmp (name, com[i].name) == 0)
      pc = &com[i];
  return pc;
}


/* koala-quit -----------------------------------------------------*/

long ceQuit (char *s) {                      /* vuelve a caen con exit (0) */

  done = 1;
  return 0;
}


/* koala-help -----------------------------------------------------*/

long ceHelp (char *s) {

  printHelp (s);
  return 0;
}


/* Print out help for ARG, or for all of the commands if ARG is not present. */
void printHelp (char *arg) {
  long i;
  long printed = 0;
     
  for (i=0; cde[i].name != NULL; i++) {
    if (isNull (arg) || (*arg == '\0')
	|| (strstr (cde[i].name, arg) == cde[i].name)) {
      fprintf (stderr, "%-12s%s.\n", cde[i].name, cde[i].doc);
      printed++;
    }
  }
  if (! printed) {
    fprintf (stderr, "No commands match `%s'.  Possibilities are:\n", arg);
    for (i = 0; cde[i].name != NULL; i++) {
      if (printed == 6) {                           /* Print in six columns. */
	printed = 0;
	fprintf (stderr, "\n");
      }
      fprintf (stderr, "%-12s", cde[i].name);
      printed++;
    }
    if (printed)
      fprintf (stderr, "\n");
  }
}


/* koala-comment -------------------------------------------------*/

long ceComment (char *s) {

  return 1;
}


/* koala-init ------------------------------------------------------*/

long ceInit (char *s) {
  char *t;
  long withGC;
  
  if (! isNull (hybNT)) {
    fprintf (stdout, "The model has already been inited\n");
    return 0;
  }
  
  rt_num = 0;
  simTime = 0;
  withGC = 1;
  precision = 1e-2;
  accuTime = 0;
  if (! isNull (t = strtok (s, " "))) {
    stripWhite (t);
    if (! isdigit (*t)) {
      if (! isNull (strstr (s, "noglob")))
	withGC = 0;
      else {
	fprintf (stderr, "Unknown parameter %s (expecting ""noglob"") !\n", t);
	return 0;
      }
      t = strtok (NULL, " ");
    } /* endif ! isdigit t */
  } /* endif ! isNull t */
  if (! isNull (t)) {
    stripWhite (t);
    if (! isdigit (*t)) {
      fprintf (stderr, "Accumulation time should be a number !\n");
      return 0;
    }
    else
      accuTime = atol (t);
    t = strtok (NULL, " ");
  } /* endif ! isNull t */
  if (! isNull (t)) {
    stripWhite (t);
    if (! isdigit (*t)) {
      fprintf (stderr, "Precision should be a number !\n");
      return 0;
    }
    else
      precision = atof (t);
    t = strtok (NULL, " ");
  } /* endif ! isNull t */
  if (! isNull (t)) {
    stripWhite (t);
    fprintf (stderr, "Unexpected parameter %s\n", t);
  } /* endif ! isNull t */
  if (discrete) {     /* working on a discrete model */
    if (! disc_init) {
      disc_init = 1;
      return koala_initialize_discrete () - 1;
    }
    else fprintf (stderr, "Already inited !\n");
  }
  else if (prepareEI () && prepareMS () && prepareDO () && prepareMO ()) {
    if (1 == koala_initialize (withGC, precision, diagOut, accuTime)) {
      prepareSc ();  /* init scenario file */ 
      fprintf (stderr, "  Initializations done.\n");
      koala_setDisplays (getRTAppliName (hybNT->rt), getRTNDisplay (hybNT->rt),
			 getRTDispInfVal (hybNT->rt), getRTDispSupVal (hybNT->rt),
			 hybNT->rt->num);
    }
    else
      fprintf (stderr, "  Initializations failed.\n");
  } 
  fprintf (dataOut, "\n");  
  return 0;
}


/* simul -------------------------------------------------------*/

long ceSimul (char *s) {
  char *t;
  double resetTime, resetPeriod=0;
  /*hybTrajectory* current_htraj;
    runtime current_rt;
    Trajectory* current_traj;*/
  int answer = 0;
  long nin, nmes=0, mode=0;
  bool nonmisb = 1;
  
  if (isNull (hybNT)) {           /* the model has not been inited */
    fprintf (stderr, "Please initialize first !\n");
    return 0;
  }
  if (isNull (t = strtok (s, " "))) {
    fprintf (stderr, "Missing parameter ! (expecting total time)\n");
    return 0;
  }
  else {
    stripWhite (t);
    if (! isdigit (*t)) {
      fprintf (stderr, "Total time should be a number !\n");
      return 0;
    }
    else { /* ! isdigit t */
      totalTime = atof (t);
      if (totalTime <= 0)
	return 0;
      t = strtok (NULL, " ");
      if (! isNull (t)) {
	stripWhite (t);
	if (isdigit (*t)) {
	  resetPeriod = atof (t);
	  mode = 1;
	  if (isNull (measIn)) {
	    fprintf (stderr, "Resetting impossible (no measure file) !\n");
	    return 0;
	  } /* endif isNull measIn */
	  t = strtok (NULL, " ");
	  if (! isNull (t))
	    if (strcmp (t, "multi") == 0)
	      simu_multi = 1;
	} /* endif  isdigit t */
	else if (strcmp (t, "multi") == 0)
	  simu_multi = 1;
      } /* endif ! isNull t */
    } /* endif ! isdigit t */
  } /* endif isNull t */
 
  if (guards_int == 1 && simu_multi == 0) {
    fprintf (stderr, "Warning the model contains spontaneous transitions jumps on non-measured values: predicted bounds will drift away. Try simulation with \'multi\' flag instead.\nDo you want to continue anyway ? (y/n)\n");
    answer = tolower (fgetc (stdin));
    if (answer == 'n')
      return 0;
  }

  if (! isNull (measIn))
    nmes = getRTNMeasure (hybNT->rt);
  nin = getRTNInput (hybNT->rt);
  nrt->clear ();    /* in simulation starting trajs from init are erased */
  initDispAll (hybNT->rt, totalTime / getRTTimeStep (hybNT->rt));

  cout << "mode: " << mode << endl;

  for (simTime=getRTTimeStep (hybNT->rt),
	 resetTime=resetPeriod - getRTTimeStep (hybNT->rt);
       simTime<=totalTime;
       simTime += getRTTimeStep (hybNT->rt),
	 resetTime -= getRTTimeStep (hybNT->rt)) {

    fprintf (stderr, "-- time = %g --------------------------\n", simTime);

    if (! isNull (dataIn))
      koala_getInputs (model_name.c_str (), nin, inps);
    if (! isNull (measIn))
      koala_getMeasures (model_name.c_str (), nmes, meas);
    
    if (-1 == koala_doCycle (mode, resetTime, resetPeriod, simTime, 0, 0, simu_multi, nonmisb))
      fprintf (stderr, "Koala cycle failed!\n");
    
  } /* end for */
  
  writeDispAll (dataOut, tracked_trajs);
  clearMultiTrajs (tracked_trajs);
  hybNT = 0;
  simu_multi = 0;

  if (dataOut != stdout)
    fclose (dataOut);
  if (! isNull (dataIn))
    fclose (dataIn);
  if (! isNull (measIn))
    fclose (measIn);
  if (! isNull (modesOut))
    fclose (modesOut);
  cefree (inps);
  cefree (meas);
  return 0;
}


/* executer-detect (prediction and default detection) ------------------------*/

long ceDetect (char *s) {
  char *t;
  long runFI, althresh=1, nmes, nin, mode;
  bool nonmisb = 1;

  std::cout << "entering detect\n";

#ifdef VERB_PERF
  clock_t start, end;
  double elapsed;
#endif

  if ((nmes = getRTNMeasure(hybNT->rt)) == 0) {
    fprintf (stderr, "No measure variable: detection impossible !\n");
    return 0;
  }
  if (! isInited (hybNT->rt)) {
    fprintf (stderr, "Please initialize first !\n");
    return 0;
  }
  if (isNull (t = strtok (s, " "))) {
    fprintf(stderr, "\nMissing parameter (expecting total time) !\n");
    return 0;
  }
  else {
    stripWhite (t);
    if (! isdigit (*t)) {
      fprintf (stderr, "Total time should be a number !\n");
      return 0;
    }
    else { /* ! isdigit *t */
      totalTime = atof (t);
      if (totalTime <= 0)
	return 0;
    }
    t = strtok (NULL, " ");
  } /* endif isNull t */
  stripWhite (t);
  if (isNull (t)) {
    fprintf(stderr, "\nExpecting fault isolation parameter:\n");
    fprintf (stderr, "\n\t nofault\
\n\t fault1 ---> NO One-Fault-Exoneration AND    Multiple-Fault-Compensation\
\n\t fault2 --->    One-fault-Exoneration AND    Multiple-Fault-Compensation\
\n\t fault3 --->    One-fault-Exoneration AND NO Multiple-Fault-Compensation\
\n\n");
    return 0;
  }
  else { /* isNull t */
    if (! isNull (strstr (t, "nofault")))
      runFI = 2;
    else if (! isNull (strstr (t, "fault1")))
      runFI = 3;
    else if (! isNull (strstr (t, "fault2")))
      runFI = 4;
    else if (! isNull (strstr (t, "fault3")))
      runFI = 5;
    else {
      fprintf (stderr, "Unknown parameter %s. Expecting one of\n\t nofault\
\n\t fault1 ---> NO One-Fault-Exoneration AND    Multiple-Fault-Compensation\
\n\t fault2 --->    One-fault-Exoneration AND    Multiple-Fault-Compensation\
\n\t fault3 --->    One-fault-Exoneration AND NO Multiple-Fault-Compensation\
\n\n"
	       , t);
      return 0;
    } /* endif ! nofault */
    /*if (runFI > 2)*/
      t = strtok (NULL, " ");
  } /* endif ! isNull t */
  /*if (runFI > 2) {*/
  stripWhite (t);
  if (isNull (t)) {
    fprintf(stderr, "\nExpecting alarm threshold parameter:\n");
    return 0;
  }
  else { /* isNull t */
    if (! isdigit (*t)) {
      fprintf (stderr, "Alarm threshold should be a number !\n");
      return 0;
    }
    else /* ! isdigit *t */
      althresh = (long) atof (t);
  } /* endif isNull t */
    /*}*/ /* endif runFI > 2 */
  if (! isNull (t = strtok (NULL, " "))) {  /* look for merging option, default is yes (mode = 4) */
    if (! isNull (strstr (t, "nomerge")))
      mode = 4;
    else
      mode = 3;
  }
  else mode = 3;
  
  nin = getRTNInput (hybNT->rt);
  nrt->clear();
  initDispAll (hybNT->rt, totalTime / getRTTimeStep (hybNT->rt));
  cerr << "mode: " << mode << endl;

#ifdef VERB_PERF
  FILE* cyc_bench;
  if ((cyc_bench = fopen ("cycbench.bmk", "w")) == NULL) {
    cout << "Cannot create cycle benchmark file...\n";
    exit (1);
  }
#endif 

  for (simTime=getRTTimeStep (hybNT->rt); simTime<=totalTime;
       simTime += getRTTimeStep (hybNT->rt)) {
    fprintf (stderr, "-- time = %g --------------------------\n", simTime);

    koala_getInputs (model_name.c_str(), nin, inps);
    koala_getMeasures (model_name.c_str(), nmes, meas);
#ifdef VERB_PERF
    start = clock ();
#endif
    //debug
    //std::cerr << "starting koala cycle.\n";
    //debug
 
    if (-1 == koala_doCycle (mode, 0, 0, simTime, runFI, althresh, 0, nonmisb))
      fprintf (stderr, "Koala cycle failed!\n");
#ifdef VERB_PERF
    end = clock ();
    elapsed = ((double)(end - start))/CLOCKS_PER_SEC;
    cout << "cycle: " << elapsed << " seconds\n";
    fprintf (cyc_bench, "%i\t%g\n", tracked_trajs->size (), elapsed);
#endif
  } /* end for */
  
  writeDispAll (dataOut, tracked_trajs);
  graphTrajs (tracked_trajs);
  clearMultiTrajs (tracked_trajs);
  hybNT = 0;
  if (dataOut != stdout)
    fclose (dataOut);
  if (! isNull (dataIn))
    fclose (dataIn);
  fclose (measIn);
  if (! isNull (modesOut))
    fclose (modesOut);
#ifdef VERB_PERF
  fclose (cyc_bench);
#endif
  cefree (inps);
  cefree (meas);

/*#ifdef DEBUG
  fprintf (stderr, "Total dynamic memory allocated =  %ld  bytes\n",
	   totalAllo);
  fprintf (stderr, "Total dynamic memory deallocated =  %ld  bytes\n",
	   totalDeallo);
  fprintf (stderr, "Memory leak =  %ld  bytes\n",
	   totalAllo - totalDeallo);
#endif*/

  return 1;
}


/* create hitting sets -----------------------------------------------*/

char *tDiag;

long ceDiagnosis (char *s) {
  int procedure;

  if (isNull (getRTDiagOut (koala_rt)))
    setRTDiagOut (koala_rt, diagOut);
  if (isNull (tDiag = strtok (s, " "))) {
    fprintf(stderr, "\nUsage: diagnosis fault[1 2 3] misb-var1 misb-var2 ...\
\n\t fault1 ---> NO One-Fault-Exoneration AND    Multiple-Fault-Compensation\
\n\t fault2 --->    One-fault-Exoneration AND    Multiple-Fault-Compensation\
\n\t fault3 --->    One-fault-Exoneration AND NO Multiple-Fault-Compensation\
n\n");
    return 0;
  }
  stripWhite (tDiag);
  if (! isNull (strstr (tDiag, "fault1")))
    procedure = 6;
  else if (! isNull (strstr (tDiag, "fault2")))
    procedure = 7;
  else if (! isNull (strstr (tDiag, "fault3")))
    procedure = 8;
  else {
    fprintf (stderr, "Unknown parameter %s. Expecting one of\
\n\t fault1 ---> NO One-Fault-Exoneration AND    Multiple-Fault-Compensation\
\n\t fault2 --->    One-fault-Exoneration AND    Multiple-Fault-Compensation\
\n\t fault3 --->    One-fault-Exoneration AND NO Multiple-Fault-Compensation\
\n\n", tDiag);
    return 0;
  } /* endif ! fault1 */

  /* Set variables in the argument list as misbehaving */
  if (isNull (tDiag = strtok (NULL, " "))) {
    fprintf (stderr, "Missing misbehaving variables !\n");
    return 0;
  }
  stripWhite (tDiag);
  if (-1 == caen_doCycle (koala_rt, procedure, 0))
    fprintf (stderr, "Invalid model !\n");
  caen_close (koala_rt);
/*#ifdef DEBUG
  fprintf (stderr, "Total dynamic memory allocated =  %ld  bytes\n",
	   totalAllo);
  fprintf (stderr, "Total dynamic memory deallocated =  %ld  bytes\n",
	   totalDeallo);
  fprintf (stderr, "Memory leak =  %ld  bytes\n",
	   totalAllo - totalDeallo);
#endif*/
  return 0;
}

/* ceDisc ---------------------------------------------------------*/

long ceDisc (char* s) {
  if (disc_init) {
    std::cout << "inputFile: " << inputFile << std::endl;
    sMM->runSchD (inputFile);
    disc_init = 0;
    hybNT = 0;
  }
  else fprintf (stderr, "Should init first !\n");
  return 1;
}

/* executer-print -------------------------------------------------*/

long cePrint (char *s) {

  runtime current_rt;
  char* argu;

  if (hybNT != NULL)
    current_rt = hybNT->rt;
  else current_rt = koala_rt;	

  if (isNull (s))
    fprintf (stderr, "Missing argument for print.\n");
  else { /* isNull s */
    if (! isNull (strstr (s, "dic"))) {
      fprintf (stderr, "Dictionary:\n");
      ce_printDic (stderr, current_rt);
    }
    else if (! isNull (strstr (s, "cond"))) {
      fprintf (stderr, "Conditions:\n");
      ce_printConds (stderr, current_rt);
    }
    else if (!isNull (argu = strstr (s, "co"))) {
      s = strtok (argu, " ");
      s = strtok(NULL, " ");
      if (isNull (s))
	sMM->showComps ();
      else
	sMM->showComp (s);
      return 0;
    }
    else if (! isNull (strstr (s, "toms"))) {
      sMM->showAtoms ();
      return 0;
    }
    else if (! isNull (strstr (s, "spa"))) {
      fprintf (stderr, "Current space:\n");
      ce_printSpace (stderr, current_rt);
    }
    else if (! isNull (strstr (s, "inp"))) {
      if (getRTNInput (current_rt)) {
	fprintf (stderr, "Inputs:\n");
	ce_printInputs (stderr, current_rt);
      }
      else
	fprintf (stderr, "No inputs.\n");      
    }
    else if (! isNull (strstr (s, "dis"))) {
      fprintf (stderr, "Displays:\n");
      ce_printDisplays (stderr, current_rt);
    }
    else if (! isNull (strstr (s, "mea"))) {
      if (getRTNMeasure (current_rt)) {
	fprintf (stderr, "Measure:\n");
	ce_printMeasures (stderr, current_rt);
      }
      else
	fprintf (stderr, "No measure variables.\n");      
    }
    else if (! isNull (strstr (s, "inf"))) {
      fprintf (stderr, "Influences:\n");
      ce_printRules (stderr, current_rt);
    }
    else if (! isNull (strstr (s, "gra"))) {
      fprintf (stderr, "Causal graph:\n");
      ce_printCGraph (stderr, current_rt);
    }
    else if (! isNull (strstr (s, "upd"))) {
      fprintf (stderr, "Updating sequence: ");
      ce_printUpdateListcaen (stderr, current_rt);
    }
    else if (! isNull (strstr (s, "glo"))) {
      fprintf (stderr, "Global constraints variables:\n");
      ce_printGlobVs (stderr, current_rt);
      fprintf (stderr, "Global constraints:\n");
      ce_printGlobCs (stderr, current_rt);
    }
    else if (! isNull (strstr (s, "cla"))) {
      fprintf (stderr, "Clauses:\n");
      sMM->showClauses ();
    }
    else if (! isNull (strstr (s, "ini"))) {
      if (getRTNInit (current_rt)) {
	fprintf (stderr, "Initializations:\n");
	ce_printInits (stderr, current_rt);
      }
      else
	fprintf (stderr, "No initializations.\n");      
    }
    else
      fprintf (stderr, "Can't print %s\n", s);
  } /* endif isNull s */
  return 0;
}

long prepareMO (void) {
  long ok=1;

  if (modesFile[0] != '\0') {
    if (isNull (modesOut = fopen (modesFile, "w"))) {
      fprintf (stderr, "Can't open file %s\n", modesFile);
      ok = 0;
    }
  }
  else modesOut = NULL;
  
  if (ok) printf("modes file is ok\n");
  return ok;
}

long prepareDO (void) {
  long ok=1;

  if (displayFile[0] == '\0')
    dataOut = stdout;
  else
    if (isNull (dataOut = fopen (displayFile, "w"))) {
      fprintf (stderr, "Can't open file %s\n", displayFile);
      ok = 0;
    }
  if (ok) printf("display file is ok\n");
  return ok;
}


/* koala-system ---------------------------------------------------*/

long ceSystem (char *s) {
  long r=1;

  if (-1 == system (s)) {
    fprintf (stderr, "Shell command failed.\n");
    r = 0;
  }
  return r;
}


/* koala-debug ----------------------------------------------------*/

long ceDebug (char *s) {
  bool deb_done = 0;;
  char *migne;
  
  while (! deb_done) {
    if (! isNull (migne = do_gets ("debug>"))) {
      ligne = migne;
      if (*ligne == '\0')       /* empty line */
	continue;
      
    }
  
  }
  return 1;
}

long deExit (char *s) {

  return 1;
}

long deHelp (char *s) {

  return 1;
}

long dePrint (char *s) {
  
  return 1;
}

/* koala-gnuplot --------------------------------------------------*/

long ceGnup (char *s) {
  char *t;
  char u[40]="gnuplot ";

  if (! isNull (t = strtok (s, " "))) {
    stripWhite (t);
    strcat (u, t);
  } /* endif isNull t */
  if (-1 == system (u))
    fprintf (stderr, "Gnuplot failed");
 return 1;
}

/* enter inputs, measures -----------------------------------------*/

long prepareEI (void) {
  long ok=1;
  printf("inputfile : %s\n", inputFile);
  if (inputFile[0] == '\0')
    dataIn = NULL;
  else {
    if (isNull (dataIn = fopen (inputFile, "r"))) {
      fprintf (stderr, "Can't open file %s\n", inputFile);
      ok = 0;
    }
  }
  if (ok) printf("input file is ok\n");
  return ok;
}

long prepareSc (void) {
  long ok=1;
  printf("scenarFIle: %s\n", scenarFile);
  if (scenarFile[0] != '\0') {
    sMM->loadScenario ();
  }
  return ok;
}

long prepareMS (void) {
  long ok=1;

  if (measureFile[0] == '\0')
    measIn = NULL;
  else {
    if (isNull (measIn = fopen (measureFile, "r"))) {
      fprintf (stderr, "Can't open file %s\n", measureFile);
      ok = 0;
    }
  }
  if (ok) printf("measure file is ok\n");
  return ok;
}


/*----------------------------------------------------------------*/

/**
 * get inputs from the data in.
 */
long koala_getInputs (const char *s, long n, double *in) {
  long i;
 
  for (i=0; i<n; ++i) {
    fscanf (dataIn, "%lg", in+i);
  }
  fscanf (dataIn, "%*[^\n]");
/*#ifdef VERB_EXEC*/
  cout << "inputs: ";
  for (i=0; i<n; ++i)
    cout << in[i] << " ";
  cout << endl;
/*#endif*/
  return 1;
}

/**
 * get measures from the data in.
 */
long koala_getMeasures (const char *s, long n, double *ms) {
  long i;
 
  for (i=0; i<n; ++i)
    fscanf (measIn, "%lg", ms+i);
  fscanf (measIn, "%*[^\n]");
/*#ifdef VERB_EXEC*/
  cout << "measures: ";
  for (i=0; i<n; ++i)
    cout << ms[i] << " ";
  cout << endl;
/*#endif*/
  return 1;
}

long koala_setDisplays (const char *s, long n, double *inf, double *sup, long nu) {
  long i;
 
  if (dataOut == stdout)
    fprintf (dataOut, "Displays: ");
  for (i=0; i<n; ++i) {
    /*if (*(inf+i) != *(sup+i))*/
      fprintf (dataOut, "%g %g ", *(inf+i), *(sup+i));
    /*else fprintf (dataOut, "%g ", *(inf+i));*/
  }
  /*fprintf (dataOut, "\n");*/
  return 1;
}

void koala_setRTDisplays (long n, runtime rt, double stime) {
  long i=0, j=0, starti=2*n*((long) ceil (stime)-1), endi=(2*n*((long) ceil (stime)-1)+2*n)-1;
  
  for (i=starti; i<endi; i=i+2) {
    /*cout << "writing inf: " << getRTDispInfValNth (rt, j) << " at " << i << endl;
    cout << "writing sup: " << getRTDispSupValNth (rt, j) << " at " << i+1 << endl;*/
    setRTDispAllNth (rt, i, getRTDispInfValNth (rt, j));
    setRTDispAllNth (rt, i+1, getRTDispSupValNth (rt, j));
    j++;
  }
}

void initDispAll (runtime rt, double ttime) {
  if (rt->ndisp) {
    rt->ndispall = (long) (ttime * getRTNDisplay (rt) * 2);
    rt->dispall = (double*) new_malloc (rt->ndispall * sizeof (double));
  }
}

void writeDispAll (FILE* fd, multimap<double, hybTrajectory*, greater<double> >* ttrajs) {
  runtime rt;
  
  cout << "writing " << ttrajs->size() << " trajectories\n";
  /*cout << "ndispall " << getRTNDispAll ((*ttrajs->begin()).second->rt) << endl;*/
  
  for (long i=0; i<getRTNDispAll ((*ttrajs->begin()).second->rt); i=i+2*getRTNDisplay ((*ttrajs->begin()).second->rt)) {
    for (multimap<double, hybTrajectory*, greater<double> >::iterator it = ttrajs->begin(); it != ttrajs->end(); it++) {
      rt = (*it).second->rt;
      for (int j=i; j<i+2*getRTNDisplay ((*ttrajs->begin()).second->rt); j++)
	fprintf (fd, "%g ", getRTDispAllNth (rt, j));
    }
    fprintf (fd, "\n");
  }
}

long caen_setAlarms (const char *s, long n, long *bl, long *bh,
		     long *cl, long *ch) {
  long i;

	/*  if (dataOut == stdout)
    fprintf (dataOut, "Low alarms: ");
  for (i=0; i<n; ++i)
    fprintf (dataOut, "%ld ", *(bl+i));
  fprintf (dataOut, "\n");
  if (dataOut == stdout)
    fprintf (dataOut, "High alarms: ");
  for (i=0; i<n; ++i)
    fprintf (dataOut, "%ld ", *(bh+i));
    fprintf (dataOut, "\n"); */
  for (i=0; i<n; ++i) {
    if (*(cl+i) || *(ch+i)) {
#ifdef VERB_EXEC
      fprintf (stderr, "Variable ");
      ce_printVarNom (stderr, getRTMeasureNth (koala_rt, i));
      fprintf (stderr, " misbehaving !\n");
#endif
    }
  }
  return 1;
}


long caen_sendDiagnosis (const char *s, char *buf) {

  if ('\0' != *buf)
    fprintf (stderr, "Diagnostic Results:\n%s\n", buf);
  return 1;
}


long caen_getMisbs (const char *s, long n, long *misb) {
  long ok=1, i, goodName;
 
  for (i=0; i<n; i++)
    *(misb+i) = 0;                                     /* RAZ misb vector */
  stripWhite (tDiag);
  while (tDiag != NULL) {
    if (isdigit (*tDiag))
      break; /* while */
    for (i=0, goodName=0; i<n; i++) {
      if (strcmp (getVarName (getRTMeasureNth (koala_rt, i)),
		  tDiag) == 0) {
	*(misb+i) = 1;
	goodName = 1;
	break; /* for */
      } /* endif strcmp */
    } /* end for i */
    if (goodName) {
      tDiag = strtok (NULL, " ");
      stripWhite (tDiag);
    }
    else {
      fprintf(stderr, "Variable %s either is not measured or does not exist !\n", tDiag);
      for (i=0; i<n; i++)
	*(misb+i) = 0;                                  /* RAZ misb vector */
      tDiag = NULL;
      ok = 0;
      break; /* while */
    }
  } /* end while tDiag */
  return ok;
}


long caen_getConds (const char *s, long n, long *cval) {
  long i, ok=1;
  influence ifl;
 
  for (i=0; i<n; i++) {
    ifl = getRTInfluNth (koala_rt, i);
    if (isNull (getInfluCond(ifl)))
      *(cval+i) = 1;
    else { /* isNull */
      if (tDiag == NULL) {
	fprintf(stderr, "Not enough condition values !\n");
	ok = 0;
	break; /* for */
      }
      else { /* tDiag == NULL */
	if (*tDiag == '0')
	  *(cval+i) = 0;
	else if (*tDiag == '1')
	  *(cval+i) = 1;
	else {
	  fprintf(stderr, "Wrong condition value !\n");
	  ok = 0;
	  break; /* for */
	}
	tDiag = strtok (NULL, " ");
	stripWhite (tDiag);
      } /* endif tDiag == NULL */
    } /* endif isNull */
  } /* end for */
  return ok;
}

void localInit (runtime rt, Scheduler* sMM) {
  StateVar* stv;
  CondVar* cdv;
  ModeVar* mdv;

  /* initialize conditions based on Scheduler's atoms. */
  for (hash_map<const char*, Atom*, hash<const char*>, eqstr>::iterator it = sMM->active_atoms->begin(); it != sMM->active_atoms->end(); it++) {
    if ((*it).second->cond) {
      stv = (*it).second->getStateVar();
      if (stv->getType() == "Cond") {
	cdv = (CondVar*) stv;
	for (multimap<const char*, int>::iterator itcd = cdv->conditions->begin();
	     itcd != cdv->conditions->end(); itcd++)
	  if (strcmp ((*itcd).first, (*it).second->getId()) == 0)
	    setRTCondVal (rt, (*itcd).second, (*it).second->evalAtom());
      }

      else if (stv->getType() == "Mode") {
	mdv = (ModeVar*) stv;
	for (multimap<const char*, int>::iterator itcd = mdv->conditions->begin();
	     itcd != mdv->conditions->end(); itcd++)
	  if (strcmp ((*itcd).first, (*it).second->getId()) == 0)
	    setRTCondVal (rt, (*itcd).second, (*it).second->evalAtom());
      }
    }
  }
  nrt = new multimap<long, runtime, less<long> >();
}

void graphTrajs (multimap<double, hybTrajectory*, greater<double> >* tt) {
  string file_name = "htraj";
  char* s;
  char buf[20];
  s = buf;
  int i = 0;
  for (multimap<double, hybTrajectory*, greater<double> >::iterator it = tt->begin(); it != tt->end(); it++) {
    sprintf(buf, "%i", i);
    toDot((*it).second->traj, string(file_name + s));
    i++;  
  }
}
