/*    
 *  Copyright 2000, 2001, 2002 Emmanuel Benazera  
 *
 *  This file is part of Koala.
 *
 *  Koala is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Koala is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Koala; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* koala runtime */

#ifndef KOALARUNT
#define KOALARUNT
#endif

#include "caen-runt.h"
#include "Scheduler.h"
#include "../behav/hybTrajectory.h"
#include <time.h>

/* required variables in exec: */
extern string model_name;
extern hybTrajectory* hybNT;
extern Scheduler* sMM;
extern multimap<double, hybTrajectory*, greater<double> >* tracked_trajs;
extern multimap<double, hybTrajectory*, greater<double> >* ntrajs;
extern multimap<double, hybTrajectory*, greater<double> >* misbtrajs;
extern runtime koala_rt;
extern double* ms;
extern double* inps;
extern vector<condition>* existing_conflicts;

extern long koala_initialize_discrete ();
extern long koala_initialize (long, double, char*, long);
extern long koala_doCycle (long, double, double, double, long, long, long, bool&);
extern void koalaDiscreteCycle (long, double, bool&, long&, long);
extern void splitRuntime (runtime, Trajectory*, multimap<double, hybTrajectory*, greater<double> >*, long, long);
extern void splitTrajs (multimap<double, hybTrajectory*, greater<double> >*, long, long);
extern void koala_initMappingFuncs (runtime);
void applyMFBounds (runtime, variable, objet);
extern bool koala_applyMappingFuncs (runtime);
extern bool koala_reinitialize (runtime&, const double&, const long&, const long&);
extern void createNewTrajs (Trajectory*, multimap<double, hybTrajectory*, greater<double> >*);
extern void addBridgeTrans (Transition* tr, runtime rt);
extern void addBridgeConflict (runtime rt, ptConflict conf);
extern void parseCondConflicts (runtime rt, condition cd, bool add, vector<Atom*>* curr);
extern void clearMultiTrajs (multimap<double, hybTrajectory*, greater<double> >* mt);
extern void clearNrt ();
extern void clearRunTime (runtime);
extern bool isMisb (runtime rt);
extern void initNominalTraj (hash_map<const char*, Atom*, hash<const char*>, eqstr>*, runtime);
extern void lockFirstTrajState (Trajectory* tr);
extern void addBridgeEvent (condition cd, Atom* at, bool tr, runtime rt);
extern void removeBridgeEvent (condition cd, Atom* at, runtime rt);
extern void addBridgeCond (const char* at, boolean b, runtime rt);
extern void bridgeIncEvents (runtime rt);
extern void setRTCondVal (runtime crt, int pos, bool b);
extern void bridgeConditions (runtime crt);
extern bool intersectObsRt (runtime rt1, runtime rt2);
extern void ce_reduxTrajs (multimap<double, hybTrajectory*, greater<double> >* ttrajs, double simTime);
extern void mergeRuntimes (runtime rt1, runtime rt2, double simTime);
bool handleMisbTrajs (multimap<double, hybTrajectory*, greater<double> >* ntrajs, multimap<double, hybTrajectory*, greater<double> >* misbtrajs, long runFI, long althresh, double simTime);
void koala_diagGetInputs (runtime, double, double, double*);
void koala_diagGetMeasures (runtime, double, double, double*);
void addDiagBridgeTrans (Trajectory*, runtime, int); /* deprecated */
void updateDiagConds (Trajectory*, runtime&, const int&, const double&, const double&, const long&);
bool verifyAndUpdateTrajs (double simTime, long runFI, long althresh);
long koala_checkCycle (multimap<double, hybTrajectory*, greater<double> >*, double, double, double, long, long, long);
void koalaDiagDiscreteCycle (long, double, multimap<double, hybTrajectory*, greater<double> >*);
void insertDiagnosis (hybTrajectory* htr, multimap<double, hybTrajectory*, greater<double> >* tt);

/* diagnosis modes switching strategy */
multimap<double, hybTrajectory*, greater<double> >* copyHybTrajectories (multimap<double, hybTrajectory*, greater<double> >* mtrajs);
void switchToBST (multimap<double, hybTrajectory*, greater<double> >* ttrajs, multimap<double, hybTrajectory*, greater<double> >* mtrajs, double stime, double ldetect, long &runFI, long althresh);

/* required functions in exec: */
extern void koala_setRTDisplays (long, runtime, double);
extern long koala_getInputs (const char*, long, double*);
extern long koala_getMeasures (const char*, long, double*);
