/*#define DIAG_EXO*/

#include "itmsrt.h"

//#define DEBUG_ITMS

using namespace std;

/** 
 * constructor with debug flag.
 */
itms::itms() : mdlChk() {
  lock_comp = 0;
}
/* destructor. */
itms::~itms() {
  delete state_variables;
  state_variables = 0;
}

/* 
* get an elt. 
* set idle trans if trans_idle true.
*/
TrajEl* itms::getElt(const char* id, State* wkst, Trajectory* tr, 
		     const int &timestep, const bool &trans_idle) {
  TrajEl* res = (*wkst->state)[id];
  if (res != NULL)
    return res;
  else {
    res = tr->getElt(id, timestep, -1);
    /* add components to the current state. */
    tr->addComp(res->at->getVarComp(), wkst, timestep, trans_idle, lock_comp);
#ifdef USE_BDD
    addToStateBdd(wkst, res->at->getVarComp()->getName());
#endif
    res = (*wkst->state)[id];
    return res;
  }
}
//compatibility... set idle trans.
TrajEl* itms::getElt(const char* id, State* wkst, Trajectory* tr, 
		     const int &timestep) {
  TrajEl* res = (*wkst->state)[id];
  if (res != NULL)
    return res;
  else {
    res = tr->getElt(id, timestep, -1);
    /* add components to the current state. */
    tr->addComp(res->at->getVarComp(), wkst, timestep, 1, lock_comp);
#ifdef USE_BDD
    addToStateBdd(wkst, res->at->getVarComp()->getName());
#endif
    res = (*wkst->state)[id];
    return res;
  }
}

/*************************
 * Knowledge Compilation *
 *************************/

/* initialize a set of proposition */
void itms::initialize() {
  for (hash_map<const char*, Atom*, hash<const char*>, eqstr>::iterator i = active_atoms->begin(); 
       i != active_atoms->end(); i++) {
    compileTransitionsProb((*i).second);  /* set atoms probabilities from transitions (ModeVar only). */
    if ((*i).second->evalAtom()) {
      (*i).second->set();
      for (unsigned int k=0; k<(*i).second->stateVar->getAtomLinks()->size(); k++)
	if ((*i).second != (*(*i).second->stateVar->getAtomLinks())[k]) {
	  (*(*i).second->stateVar->getAtomLinks())[k]->setLabel("false");
	}
      (*i).second->propagation_number = 1;  /* init the propagation number to 1. */
    }
  }
}

/* load the atoms from clauses. */
hash_map<const char*, Atom*, hash<const char*>, eqstr>* itms::loadAtoms(vector<Clause*>* clauses) {
  /* load atoms from the set of clauses. */
  hash_map<const char*, Atom*, hash<const char*>, eqstr>* res = new hash_map<const char*, Atom*, hash<const char*>, eqstr>();
  for (unsigned int i=0; i<clauses->size(); i++)
    for (multimap<bool, Atom*, eqbool>::iterator it = (*clauses)[i]->cl->begin(); it != (*clauses)[i]->cl->end(); it++)
      if (res->count((*it).second->getId()) == 0) 
	(*res)[(*it).second->getId()] = (*it).second;
  /* load atoms from internal clauses. */
  for (unsigned int j=0; j<state_variables->size(); j++) {
    for (unsigned int k=0; k<(*state_variables)[j]->internal_clauses->size(); k++) 
      for (multimap<bool, Atom*, eqbool>::iterator it2 = (*(*state_variables)[j]->internal_clauses)[k]->cl->begin(); 
	   it2 != (*(*state_variables)[j]->internal_clauses)[k]->cl->end(); it2++)
      if (res->count((*it2).second->getId()) == 0) 
	(*res)[(*it2).second->getId()] = (*it2).second;
  }
  return res;
}

/**
 * set an id to each transition in a component.
 * Also update idle and fail not commanded transitions references into commands atoms. 
 * ( = universal quantifier).
 * @param co is a component.
 * @param curr_nu is the current index of transitions in the model.
 */
void itms::idTransitions(Comp* co, int &curr_nu) {
  multimap<bool, Atom*, eqbool>* allsigned;
  bool command;
  for (unsigned int i=0; i<co->component_transitions->size(); i++) {
    (*co->component_transitions)[i]->trans_id = curr_nu;
#ifdef VERBOSE_COMPIL
    (*co->component_transitions)[i]->showTransition();
#endif
    /* add the transition to global set. */
    (*active_trans)[curr_nu] = (*co->component_transitions)[i];
    /* set the transition weight. */
    (*co->component_transitions)[i]->weight = (*co->component_transitions)[i]->proba / (double)co->component_transitions->size();
#ifdef VERBOSE_COMPIL
    cout << "weight : " << (*co->component_transitions)[i]->weight << endl;
#endif
    curr_nu++;
    /* update references in case the transition is idle or fail and not commanded. */
    if ((*co->component_transitions)[i]->idle | (*co->component_transitions)[i]->getType() == "Fail") {
      command = false;
      /* search for a command. */
      allsigned = (*co->component_transitions)[i]->pre->getAllSignedAtoms();
      for (multimap<bool, Atom*, eqbool>::iterator it = allsigned->begin(); it != allsigned->end(); it++) {
	if ((*it).second->getVarType() == "Command") {
	  command = true;
	  break;
	}
      }
      delete allsigned;
      allsigned = 0;
      /* set the 'commanded' flag of the transition. */
      if (!command)
	(*co->component_transitions)[i]->commanded = false;
      if (!command & co->cmd_in != NULL) {
	/* references the transition into component 'in' commands atoms. */
	for (unsigned int j = 0; j < co->cmd_in->size(); j++)
	  for (unsigned int k=0; k < (*co->cmd_in)[j]->atoms_links->size(); k++) {
	    (*(*co->cmd_in)[j]->atoms_links)[k]->addRefTransPre((*co->component_transitions)[i]);
#ifdef VERBOSE_COMPIL
	    cout << "adding idle trans reference :\n";
	    cout<< (*(*co->cmd_in)[j]->atoms_links)[k]->getId() << endl;
	    (*co->component_transitions)[i]->showTransition();
#endif
	  }
      }
    }
  }
  /* get mode variables transitions guards (relies on ids). */
  co->getTransitionsGuards();
  for (unsigned int m=0; m<co->component_transitions->size(); m++) {
#ifdef VERBOSE_COMPIL
    cout << "guards number of " << (*co->component_transitions)[m]->pre_at->getId() << " : " << (*co->component_transitions)[m]->pre_at->guards_number << endl;
    cout << "guards number of " << (*co->component_transitions)[m]->modevar->name << " : " << (*co->component_transitions)[m]->modevar->guards_number << endl;
#endif
    /* set guarded probability. */
    (*co->component_transitions)[m]->proba_guard = (*co->component_transitions)[m]->proba / (double) (*co->component_transitions)[m]->modevar->guards_number;
#ifdef VERBOSE_COMPIL
    cout << "guard proba : " << (*co->component_transitions)[m]->proba_guard << endl;
#endif
  }
}

/* compile transitions probabilities w.r.t. modes vars. */
void itms::compileTransitionsProb(Atom* at) {
  int den = 0;
  double probs = 0;
  for (unsigned int i=0; i<at->ref_trans_post->size(); i++) {
    den++;
    probs = probs + (*at->ref_trans_post)[i]->proba;
  }
  if (den != 0) {
    at->post_prob = probs/(double)den;
#ifdef VERBOSE_COMPIL
    cout << "compiled probability : " << at->getId() << " -- " << at->post_prob << endl;
#endif
  }
}
/* 
 * intersect two sets of clauses for common atoms. 
 * true if disjoint.
 */
bool itms::intersectClauses(vector<Clause*>* cls1, vector<Clause*>* cls2) {
  /* create hash tables to speed up the search. */
  hash_map<const char*, Atom*, hash<const char*>, eqstr>* h1 = new hash_map<const char*, Atom*, hash<const char*>, eqstr>();
  hash_map<const char*, Atom*, hash<const char*>, eqstr>* h2 = new hash_map<const char*, Atom*, hash<const char*>, eqstr>();
  for (unsigned int i1=0; i1<cls1->size(); i1++)
    for (multimap<bool, Atom*, eqbool>::iterator it1 = (*cls1)[i1]->cl->begin(); it1 != (*cls1)[i1]->cl->end(); it1++)
      (*h1)[(*it1).second->getId()] = (*it1).second;
  for (unsigned int i2=0; i2<cls2->size(); i2++)
    for (multimap<bool, Atom*, eqbool>::iterator it2 = (*cls2)[i2]->cl->begin(); it2 != (*cls2)[i2]->cl->end(); it2++)
      (*h2)[(*it2).second->getId()] = (*it2).second;
  /* intersect. */
  for (hash_map<const char*, Atom*, hash<const char*>, eqstr>::iterator ith1 = h1->begin(); ith1 != h1->end(); ith1++)
    if ((*h2)[(*ith1).first] != NULL) {
      delete h1;
      h1 = 0;
      delete h2;
      h2 = 0;
      return false;
    }
  delete h1;
  h1 = 0;
  delete h2;
  h2 = 0;
  return true;
}

/* idem on an atom and a vector of atoms. */
bool itms::intersectClauses(Atom* at, vector<Atom*>* vec) {
  for (unsigned int i=0; i<vec->size(); i++)
    if (!intersectClauses(at->ref_clauses, (*vec)[i]->ref_clauses))
      return false;
  return true;
}

/* initialize the pb bdd (from mdlChk). */
#ifdef USE_BDD
void itms::initBddPb(State* wkst) {
  if (wkst->bdd_state != NULL)
    pb = bdd(*wkst->bdd_state);
  else
    pb = bdd(*makeStateBdd(wkst));
  delete wkst->pb;
  wkst->pb = 0;
  wkst->pb = new bdd(pb);
  if (wkst->observable) {
    delete wkst->no_obs;
    wkst->no_obs = 0;
    delete wkst->obs;
    wkst->obs = 0;
    wkst->obs = new bdd(pb);
    wkst->no_obs = new bdd(pb);
  }
#ifdef ENTROPY
  for (hash_map<const char*, TrajEl*, hash<const char*>, eqstr>::iterator it = wkst->state->begin(); 
       it != wkst->state->end(); it++) {
    if ((*it).second->lock == 0 || (*it).second->lock == 3) {
      pb &= (*corresp_atoms)[(*it).second->getId()];
    }
  }
  initStateProbas (wkst);
#endif
}
#else /* USE_BDD */
#ifdef PROBAS
void itms::initPb (State* wkst) {
  initStateProbas (wkst);
}
#endif
#endif

void itms::rtInit() {
  /* load state variables. */
  vector<StateVar*>* stav;
  int trans_num = 0;
  state_variables = new vector<StateVar*>();
  
  for (hash_map<const char*, Comp*, hash<const char*>, eqstr>::iterator it = active_comps->begin(); 
       it != active_comps->end(); it++) {
    stav = (*it).second->getStateVariables();
    for (unsigned int i=0; i<stav->size(); i++) {
      state_variables->push_back((*stav)[i]);
    }
    idTransitions((*it).second, trans_num);
  }
  
  active_atoms = loadAtoms (cls);
  
  for (hash_map<const char*, Comp*, hash<const char*>, eqstr>::iterator it = active_comps->begin(); 
       it != active_comps->end(); it++) {
    stav = (*it).second->getStateVariables();
    (*it).second->atoms->clear();
    for (unsigned int i=0; i<stav->size(); i++) {
      for (unsigned int j=0; j<(*stav)[i]->atoms_links->size(); j++) {
	(*(*it).second->atoms)[(*(*stav)[i]->atoms_links)[j]->getId()] = (*(*stav)[i]->atoms_links)[j];
	(*(*(*stav)[i]->atoms_links)[j]->ref_comps)[(*it).second->name.c_str()] = (*it).second;
	deadEndAtom ((*(*stav)[i]->atoms_links)[j]); /* check for dead end atoms */
      }
    }
  }
  initialize();
}

void itms::deadEndAtom (Atom* at) {
  for (unsigned int i=0; i<at->ref_trans_pre->size (); i++) {
    Transition* trans = (*at->ref_trans_pre)[i];
    if (trans->type == "Nominal" & ! trans->idle)
      return;
  }
  at->dead_end = true;   /* can't get out of that mode */
}

/*********************************
 * Truth Maintenance over States *
 *********************************/

hash_map<const char*, TrajEl*, hash<const char*>, eqstr>* itms::copyState(hash_map<const char*, TrajEl*, hash<const char*>, eqstr>* state, State* wkst) {
  hash_map<const char*, TrajEl*, hash<const char*>, eqstr>* n_state = new hash_map<const char*, TrajEl*, hash<const char*>, eqstr>();
  for (hash_map<const char*, TrajEl*, hash<const char*>, eqstr>::iterator st_it = state->begin(); st_it != state->end(); st_it++) {
    (*n_state)[(*st_it).second->getId()] = (*wkst->state)[(*st_it).second->getId()];
  }
  return n_state;
}

/* compute the max of the propagation_number of a state formula. */
int itms::maxPropagationNumber(Clause* cla, TrajEl* p, State* wkst, Trajectory* tr, 
			       const int &timestep) {
  TrajEl* curr_el;
  int res = 0;
  for (multimap<bool, Atom*, eqbool>::iterator it = cla->cl->begin(); it != cla->cl->end(); it++) {
    curr_el = getElt((*it).second->getId(), wkst, tr, timestep);
    if (curr_el->propagation_number > res)
      res = curr_el->propagation_number;
  }
  return res;
}

/* test if resupporting is possible. added atoms are positive. */
bool itms::isResupportable(TrajEl* p, Clause* cla, const bool &non, State* wkst, Trajectory* tr, 
			   const int &timestep) {
  for (multimap<bool, Atom*, eqbool>::iterator it = cla->cl->begin(); it != cla->cl->end(); it++) {
    if (getElt((*it).second->getId(), wkst, tr, timestep) != p) {
      if (!(p->propagation_number <= (*it).second->propagation_number) 
	  || (!(*it).first && evalAtomOnLabels(it, wkst, tr, timestep) == "true") 
	  || ((*it).first && evalAtomOnLabels(it, wkst, tr, timestep) == "false"))
	return false;
    }
    else if ((*it).first != non) return false;
  }
  return true;
}

/* test if at is in vec. */
bool itms::inVector(Atom* at, vector<Atom*>* vec) {
  for (unsigned int i=0; i<vec->size(); i++) {
    if ((*vec)[i] == at) {
      return true;
    }
  }
  return false;
}
bool itms::inVector(const char* id, vector<const char*>* vec) {
  for (unsigned int i=0; i<vec->size(); i++)
    if (strcmp(id, (*vec)[i]) == 0)
      return true;
  return false;
}
bool itms::inVector(TrajEl* at, vector<TrajEl*>* vec) {
  for (unsigned int i=0; i<vec->size(); i++)
    if ((*vec)[i] == at) return true;
  return false;
}
//test if a transitino is in a vector.
bool itms::inVector(Transition* tr, vector<Transition*>* vec) {
  for (unsigned int i=0; i<vec->size(); i++)
    if ((*vec)[i] == tr) return true;
  return false;
}
//test if a statevar is in a vector.
bool itms::inVectorStateVar(StateVar* st, vector<StateVar*>* vec) {
  for (unsigned int i=0; i<vec->size(); i++)
    if ((*vec)[i] == st) return true;
  return false;
}
//test if a clause is in a vector.
bool itms::inVectorClause(Clause* cl, vector<Clause*>* vec) {
  for (unsigned int i=0; i<vec->size(); i++)
    if ((*vec)[i] == cl) return true;
  return false;
}
//test in a multimap.
bool itms::inMultimap(Atom* at, multimap<bool, Atom*, eqbool>* mul) {
  for (multimap<bool, Atom*, eqbool>::iterator i = mul->begin(); i != mul->end(); i++)
    if ((*i).second == at) return true;
  return false;
}
//return the set of statevars from the set of elts. NOT USED !
hash_map<const char*, StateVar*, hash<const char*>, eqstr>* itms::getStateVars(vector<TrajEl*>* vec) {
  hash_map<const char*, StateVar*, hash<const char*>, eqstr>* res = new hash_map<const char*, StateVar*, hash<const char*>, eqstr>();
  for (unsigned int i=0; i<vec->size(); i++) {
    if ((*res)[(*vec)[i]->getStateVar()->getCharName()] == NULL)
      (*res)[(*vec)[i]->getStateVar()->getCharName()] = (*vec)[i]->getStateVar();
  }
  return res;
}
//set the clause as a support for an elt and update 'supported' for each elt corresponding to atom in cl.
void itms::setSupport(Clause* cla, TrajEl* at, State* wkst, Trajectory* tr, 
		      const int &timestep) 
{
  if (at->support != NULL) {
    
    //debug
    /* std::cout << "support clause: " << at->support << ":\n";
       at->support->showClause (true); */
    //debug

    for (multimap<bool, Atom*, eqbool>::iterator it1 = at->support->cl->begin(); it1 != at->support->cl->end(); ++it1) {
      /* std::cout << "elt: " << (*it1).second << std::endl;
	 std::cout << "id: " << (*it1).second->getId () << std::endl; */
      TrajEl *curr_el = getElt((*it1).second->getId(), wkst, tr, timestep);
      if (curr_el != at) {
	for (vector<TrajEl*>::iterator d = curr_el->supported->begin(); d != curr_el->supported->end(); d++) {
	  if ((*d) == at) {
	    curr_el->supported->erase(d);
	    break;
	    //d = curr_el->supported->end()-1;
	  }
	}
      }
    }
  }
  
  //debug
  /* std::cout << "set cla: ";
     cla->showClause (true);
     std::cout << "as support to: " << at->getId () << std::endl; */
  //debug

  at->support = cla;
  for (multimap<bool, Atom*, eqbool>::iterator it = cla->cl->begin(); it != cla->cl->end(); it++) {
    TrajEl *curr_el = (*wkst->state)[(*it).second->getId()];
    if (curr_el != at)
      curr_el->supported->push_back(at);
  }
}
//get from the working state the elts that are labelled 'unknown'.
multimap<bool, TrajEl*, eqbool>* itms::getNotUnknown(Clause* cla, State* wkst, Trajectory* tr, 
						     const int &timestep) {
  TrajEl* curr_el;
  multimap<bool, TrajEl*, eqbool>* res = new multimap<bool, TrajEl*, eqbool>();
  for (multimap<bool, Atom*, eqbool>::iterator it = cla->cl->begin(); it != cla->cl->end(); it++) {
    curr_el = getElt((*it).second->getId(), wkst, tr, timestep);
    if (curr_el != NULL)
      if (curr_el->label == "unknown")
	res->insert(pair<bool, TrajEl*>((*it).first, curr_el));
  }
  return res;
}
TrajEl* itms::getUnknown(Clause* cla, State* wkst, Trajectory* tr, const int &timestep) {
  TrajEl* curr_el;
  for (multimap<bool, Atom*, eqbool>::iterator it = cla->cl->begin(); it != cla->cl->end(); it++) {
    curr_el = getElt((*it).second->getId(), wkst, tr, timestep);
    if (curr_el->getLabel() == "unknown")
      return curr_el;
  }
  return NULL;
}

/* eval atom on label. */
string itms::evalAtomOnLabels(multimap<bool, Atom*, eqbool>::iterator it, State* wkst, 
			      Trajectory* tr, const int &timestep) {
  string lab = getElt((*it).second->getId(), wkst, tr, timestep)->getLabel();
  if ((*it).first && lab == "true") return "false";
  else if ((*it).first && lab == "false") return "true";
  else if (lab == "true") return "true";
  else if (lab == "false") return "false";
  else if (lab == "unknown") return "unknown";
  else return "";
}

/* test a clause, considering the working state. */
bool itms::testClause(Clause* cla, State* wkst, Trajectory* tr, const int &timestep) {
  TrajEl* curr_el;
  for (multimap<bool, Atom*, eqbool>::iterator it = cla->cl->begin(); it != cla->cl->end(); it++) {
    curr_el = getElt((*it).second->getId(), wkst, tr, timestep);
    if ((!(*it).first && curr_el->getLabel() == "true") 
	|| ((*it).first && curr_el->getLabel() == "false")) 
      return true;
  }
  return false; //return false if 'unknown'...
}

/* test if the clause is a unit clause. (all literal false but one 'unknown'). */
TrajEl* itms::unitClause(Clause* cla, State* wkst, Trajectory* tr, const int &timestep) {
  string lab;
  TrajEl* curr_el;
  TrajEl* res = 0;
  int unk = 0; //number of 'unknown' labels that have been found.
  for (multimap<bool, Atom*, eqbool>::iterator it = cla->cl->begin(); it != cla->cl->end(); it++) {
    curr_el = getElt((*it).second->getId(), wkst, tr, timestep);
    lab = evalAtomOnLabels(it, wkst, tr, timestep); //useful ??
    if (!(*it).first & lab == "true") return NULL;
    else if (lab == "unknown") {
      unk++;
      res = curr_el;
    }
  }
  if (unk >= 1) return res;
  else return NULL;
}
//return the first unit clause to be found in at's dependencies. DEPRECATED ??
Clause* itms::getFirstUnitClause(Atom* at, State* wkst, Trajectory* tr, const int &timestep) {
  /* search for the clauses that contain only 1 'unknown' atom,
   * and where the others are evaluated as 'false'.
   */
  for (unsigned int i=0; i<at->ref_clauses->size(); i++) {
#ifdef DEBUG_ITMS
    (*at->ref_clauses)[i]->showClause(0);
#endif
    if (unitClause((*at->ref_clauses)[i], wkst, tr, timestep) != NULL) {
      return (*at->ref_clauses)[i];
    }
  }
  return NULL;
}
//return the first unit clause from the entire set of clauses.
Clause* itms::getFirstUnitClause(State* wkst, Trajectory* tr, const int &timestep) {
  for (unsigned int i=0; i<cls->size(); i++) {
#ifdef DEBUG_ITMS
    (*cls)[i]->showClause(0);
#endif
    if (unitClause((*cls)[i], wkst, tr, timestep) != NULL) {
#ifdef DEBUG_ITMS
      (*cls)[i]->showClause(0);
#endif
      return (*cls)[i];
    }
  }
 return NULL;
}
Clause* itms::getFirstUnitClause(vector<TrajEl*>* deleted, State* wkst, Trajectory* tr, 
				 const int &timestep) {
  for (unsigned int i=0; i<deleted->size(); i++) {
    vector<Clause*>* refs = (*deleted)[i]->at->ref_clauses;
    for (unsigned int j=0; j<refs->size(); j++) {
#ifdef DEBUG_ITMS
      (*refs)[j]->showClause(0);
#endif      
      if (unitClause((*refs)[j], wkst, tr, timestep) != NULL) {
#ifdef DEBUG_ITMS
	(*refs)[j]->showClause(0);
#endif
	return (*refs)[j];
      }
    }
  }
  return NULL;
}

//resupport the dependence of an atom.
void itms::resupportDependence(TrajEl* at, hash_map<const char*, TrajEl*, hash<const char*>, eqstr>* resc, State* wkst, Trajectory* tr, const int &timestep) {
  multimap<bool, TrajEl*, eqbool>* not_unknown;
  vector<TrajEl*>* temp_frm;
  TrajEl* curr_el;
  vector<Clause*>* refs = at->at->ref_clauses; //get clauses which 'at' appears in.
  for (unsigned int i=0; i<refs->size(); i++) {
#ifdef DEBUG_ITMS
    (*refs)[i]->showClause(0);
#endif    
    if (testClause((*refs)[i], wkst, tr, timestep)) {
      not_unknown = getNotUnknown((*refs)[i], wkst, tr, timestep); //get dependences of at.
      for (multimap<bool, TrajEl*, eqbool>::iterator j = not_unknown->begin(); 
	   j != not_unknown->end(); j++) { //try to resupport atoms that were already 'true' or 'false'.
	//a proposition can be resupported only if its propagation number is greater than the propagation number of all other propositions occuring in the clause.
	if ((*j).second != at & isResupportable((*j).second, (*refs)[i], (*j).first, wkst, tr, timestep)) {
#ifdef DEBUG_ITMS
	  cout << "[resupportDependency] --> resupporting :\n";
	  (*j).second->showAtom(0);
#endif
	  if ((*j).second->support != NULL) {   /* erase the old support. */
	    for (multimap<bool, Atom*, eqbool>::iterator k = (*j).second->support->cl->begin(); k != (*j).second->support->cl->end(); k++) {
	      curr_el = getElt((*k).second->getId(), wkst, tr, timestep);
	      temp_frm = curr_el->supported;
	      for (vector<TrajEl*>::iterator it_frm = temp_frm->begin(); it_frm != temp_frm->end(); it_frm++)
		  if ((*it_frm) == (*j).second) {
		    temp_frm->erase(it_frm);
		    break;
		  }
	      }
	    }
	  (*j).second->support = (*refs)[i]; //update the support.
	  //update propagation number.
	  (*j).second->propagation_number = maxPropagationNumber((*refs)[i], (*j).second, wkst, tr, timestep) + 1;
	  //add the atom to resc, in order to later check for internal conflicts. (could be added twice ???).
	  resc->insert(pair<const char*, TrajEl*>((*j).second->getId(), (*j).second));
	}
      } 
      delete not_unknown;
      not_unknown = 0;
    }
  }
}
//check for internal conflicts.
void itms::checkForFocusedInternalConflicts(hash_map<const char*, TrajEl*, hash<const char*>, eqstr>* chg, queue<Clause*>* forced_cstrs, State* wkst, Trajectory* tr, const int &timestep) {
  Clause* cl_tmp;
  //must store statevars...
  vector<StateVar*>* stavars_tmp = new vector<StateVar*>();
  for (hash_map<const char*, TrajEl*, hash<const char*>, eqstr>::iterator i = chg->begin(); i != chg->end(); i++) {
    if (!inVectorStateVar((*i).second->getStateVar(), stavars_tmp)) {
      cl_tmp = checkForInternalConflicts((*i).second, wkst, tr, timestep);
      stavars_tmp->push_back((*i).second->getStateVar());
      if (cl_tmp != NULL)
	forced_cstrs->push(cl_tmp);
    }
  }
  delete stavars_tmp;
  stavars_tmp = 0;
}
Clause* itms::checkForInternalConflicts(TrajEl* at, State* wkst, Trajectory* tr, 
					const int &timestep) {
  StateVar* at_sta = at->getStateVar();
  Clause* res = NULL;
  for (unsigned int k=0; k<at_sta->internal_clauses->size(); k++) {
    if (!testClause((*at_sta->internal_clauses)[k], wkst, tr, timestep) 
	&& getUnknown((*at_sta->internal_clauses)[k], wkst, tr, timestep) == NULL) {
      res = (*at_sta->internal_clauses)[k];
    }
  }
  return res;
}
void itms::checkForFocusedConflicts(hash_map<const char*, TrajEl*, hash<const char*>, eqstr>* chg, queue<Clause*>* forced_cstrs, State* wkst, Trajectory* tr, const int &timestep) {
 for (hash_map<const char*, TrajEl*, hash<const char*>, eqstr>::iterator it = chg->begin(); it != chg->end(); it++) {
    for (unsigned int i=0; i<(*it).second->at->ref_clauses->size(); i++) {
#ifdef DEBUG_ITMS
      (*(*it).second->at->ref_clauses)[i]->showClause(0);
#endif      
      if (!testClause((*(*it).second->at->ref_clauses)[i], wkst, tr, timestep) 
	  && getUnknown((*(*it).second->at->ref_clauses)[i], wkst, tr, timestep) == NULL) { //to be optimized... (merging both procedures is not so useful anyway...).
	forced_cstrs->push((*(*it).second->at->ref_clauses)[i]);
      }
    }
  }
}
//check for false constraints and return them as conflicts.
void itms::checkForConflicts(vector<Clause*>* forced_cstrs, State* wkst, Trajectory* tr, 
			     const int &timestep) {
  if (forced_cstrs == NULL) 
    forced_cstrs = new vector<Clause*>();
  for (unsigned int i=0; i<cls->size(); i++) {
    if (!testClause((*cls)[i], wkst, tr, timestep)) {
      forced_cstrs->push_back((*cls)[i]);
    } 
  }
#ifdef DEBUG_ITMS
  cout << "*** detected conflicts ***\n";
  viewClauseVector(forced_cstrs);
#endif
}
//propagate through a unit clause.
void itms::propagateClause(Clause* cla, State* wkst, Trajectory* tr, 
			   const int &timestep) {
  TrajEl* curr_el;
  TrajEl* unknown = getUnknown(cla, wkst, tr, timestep);
#ifdef DEBUG_ITMS
  cout << "[propagateClause] --> propagating (setting to true) :\n";
  unknown->showAtom(1);
#endif
  unknown->setLabel("true");
#ifdef USE_BDD
  if (wkst->observable) {
    addNode(wkst->pb, unknown->getId(), "true");
    if (unknown->getVarType() != "Observable")
      addNode(wkst->no_obs, unknown->getId(), "true");
    else addNode(wkst->obs, unknown->getId(), "true");
  }
#endif
  /* set the atom's supported. */
  for (multimap<bool, Atom*, eqbool>::iterator it = cla->cl->begin(); it != cla->cl->begin(); it++) {
    curr_el = getElt((*it).second->getId(), wkst, tr, timestep);
    if (curr_el != unknown)
      curr_el->supported->push_back(unknown); 
  }
}

/* propagate through conflict. return the first "path"... the others are stored in agenda anyway... */
ItmsRes* itms::propagateThroughConflict(Clause* cla, ItmsRes* curr_res, stack<ItmsRes*>* ag, 
					const bool &nom, const double &rank, Trajectory* tr, const int &timestep) {
  TrajEl* curr_el;
  ItmsRes* itms_res;
  ItmsRes* resi = 0;
  hash_map<const char*, TrajEl*, hash<const char*>, eqstr>* resc;
  hash_map<const char*, TrajEl*, hash<const char*>, eqstr>* al_chg = curr_res->already_changed;
  queue<Clause*>* forced_cstrs = curr_res->forced_cstrs;
  State* wkst = curr_res->res;
  vector<TrajEl*>* res;
#ifdef DEBUG_ITMS
  cout << "[propagateThroughConflict] --> beginning\n";
  cout << "curr_conf :\n";
  cla->showClause(0);
  cout << "entering constraints size : " << forced_cstrs->size() << endl;
#endif
  /* in order to minimize the entropy. */
  multimap<double, TrajEl*, less<double> >* not_unknown = new multimap<double, TrajEl*, less<double> >();
  string lab;
  
  int count = 0;
  bool rank_test;
  for (multimap<bool, Atom*, eqbool>::iterator cl_it = cla->cl->begin(); cl_it != cla->cl->end(); cl_it++) {
    curr_el = getElt((*cl_it).second->getId(), wkst, tr, timestep);
    rank_test = true;
    if ((*cl_it).second->fail && curr_el->getLabel() == "false")
      /* evaluate the rank. */
      rank_test = (tr->rank - tr->getStateRank(wkst) + wkst->rank + 1 <= rank);
#ifdef DEBUG_ITMS
    cout << "fail : " << (*cl_it).second->fail << endl;
    cout << "rank_test : " << rank_test << endl;
    cout << "wkst rank : " << wkst->rank << endl;
    cout << "curr_el : " << curr_el->getId() << endl;
    cout << "curr_el label : " << curr_el->getLabel() << endl;
    cout << "curr_el lock : " << curr_el->lock << endl;
    cout << "curr_el support: " << curr_el->support << endl;
    if (curr_el->support) {
      curr_el->support->showClause (true);
      cout << endl;
    }
    cout << "count : " << al_chg->count((*cl_it).second->getId()) << endl;
    if (curr_el->prevTrans != NULL)
      curr_el->prevTrans->showTransition();
#endif
    if (rank_test && al_chg->count((*cl_it).second->getId()) == 0 
    && ((curr_el->getLabel() == "false" 
	 && (curr_el->lock != 1 && curr_el->lock != 4)) 
	|| (curr_el->getLabel() == "true" 
	    && curr_el->lock != 0 && curr_el->lock != 3))) {
      if (!nom) {
#ifdef DEBUG_ITMS
	  cout << "adding to not_unknown : " << curr_el->getId() << endl;
#endif
	  /* speed-up over commands => no change from a command to something else than none. */
	  if (curr_el->getVarType() != "Command" 
	      || (curr_el->at->value == "none" && curr_el->getLabel () == "false")
	      || curr_el->getLabel () == "true") {
#ifdef ENTROPY
	  not_unknown->insert(pair<double, TrajEl*>(entropy((*cl_it).second->getStateVar(), wkst, al_chg), curr_el));
#else   /* ENTROPY */
	    not_unknown->insert(pair<double, TrajEl*>(1, curr_el));
#endif  /* ENTROPY */
	  }
      }
      else if ((! ((*cl_it).second->getVarType () == "Cond" && curr_el->getLabel () == "false"))
	       && (*cl_it).second->getVarType() != "Mode" 
	       || ((*cl_it).second->getVarType() == "Mode" && curr_el->prevTrans == NULL)) { 

#ifdef DEBUG_ITMS
	cout << (*cl_it).second->getId() << endl;
	cout << (*cl_it).second->getVarType() << endl;
	if (curr_el->prevTrans != NULL)
	  curr_el->prevTrans->showTransition();
#endif
#ifdef ENTROPY
	not_unknown->insert(pair<double, TrajEl*>(entropy((*cl_it).second->getStateVar(), wkst, al_chg), curr_el));
#else 
	not_unknown->insert(pair<double, TrajEl*>(1, curr_el));
#endif      
      }
    }
  }
#ifdef DEBUG_ITMS
    cout << "not_unknown's size : " << not_unknown->size() << endl;
#endif
if (not_unknown->empty()) {
    curr_res->clear();
    delete curr_res;
    curr_res = 0;
    not_unknown->clear();
    delete not_unknown;
    not_unknown = 0;
    return NULL;
  }
 bool flag_del_pis = false;
 vector<vector<Atom*>*>* el_pis = NULL;
#ifdef DEBUG_ITMS
 for (multimap<double, TrajEl*, less<double> >::iterator t = not_unknown->begin(); t != not_unknown->end(); t++) {
   cout << (*t).first << " -- " << (*t).second->getId() << endl;
 }
 cout << "[propagateThroughConflict] --> beginning propagation\n";
#endif
  for (multimap<double, TrajEl*, less<double> >::iterator i = not_unknown->begin(); i != not_unknown->end(); i++) {
    flag_del_pis = false;
    /* try to use prime implicants (when turning a label to true) */
   if (pit != NULL && (*i).second->getLabel () == "false" 
       && (*i).second->getVarType () != "Command") {
     el_pis = pit->getPIs ((*i).second->at);
   }
   if (el_pis == NULL || (*i).second->getLabel () == "true") {   /* fake implicants */
     el_pis = new vector<vector<Atom*>*> ();
     el_pis->push_back (new vector<Atom*> ());
     (*el_pis)[0]->push_back ((*i).second->at);
     flag_del_pis = true;
   }
   for (unsigned int k=0; k<el_pis->size (); k++) {
     vector<Atom*>* loc_pis = (*el_pis)[k];
    count++;
    /* Copy the working space and make the changes. */
    itms_res = new ItmsRes(wkst, forced_cstrs);
    itms_res->already_changed = copyState(al_chg, itms_res->res);
    curr_el = getElt((*i).second->getId(), itms_res->res, tr, timestep);
    
    TrajEl* el_tmp = NULL;
    for (unsigned int l=0; l<loc_pis->size (); l++) {
      el_tmp = getElt ((*loc_pis)[l]->getId (), itms_res->res, tr, timestep);
      if (el_tmp->at->fail && el_tmp->getLabel () == "false") {
	rank_test = (tr->rank - tr->getStateRank(wkst) + wkst->rank + 1 <= rank);
	if (! (rank_test && itms_res->already_changed->count(el_tmp->getId ()) == 0 
	       && ((el_tmp->getLabel() == "false" 
		    && (el_tmp->lock != 1 && el_tmp->lock != 4)) 
		   || (el_tmp->getLabel() == "true"
		       && (el_tmp->lock != 0) && el_tmp->lock != 3)))) {
	  itms_res->clear ();
	  delete itms_res;
	  if (flag_del_pis) {
	    for (unsigned z=0; z<el_pis->size (); z++)
	      delete (*el_pis)[z];
	    delete el_pis;
	  }
	  el_pis = NULL;
	  break;
	}
	}
       if (el_tmp != curr_el) {
	 if (el_tmp->getLabel () == "false")
	   el_tmp->setLabel ("true");
	 (*itms_res->already_changed)[el_tmp->getId ()] = el_tmp;
       }
     }
    if (curr_el->getVarType () == "Mode" && curr_el->at->value == "unknown" && curr_el->getLabel () == "true")
      itms_res->res->setUnknown ();   /* mark state as 'unknown' */
    if (curr_el->at->fail) itms_res->res->rank++;
    setSupport(cla, curr_el, itms_res->res, tr, timestep);
    curr_el->propagation_number = maxPropagationNumber(cla, (*i).second, itms_res->res, tr, timestep) + 1;
#ifdef DEBUG_ITMS
    cout << "[propagateThroughConflict] --> clause to be added :\n";
    curr_el->showAtom(0);
#endif
    if (curr_el->getLabel() == "false") lab = "true"; /* compute the correct label. */
    else lab = "false";
#ifdef DEBUG_ITMS
    cout << "\n[propagateThroughConflict] --> adding a new clause. " << lab << endl;
#endif
    (*itms_res->already_changed)[curr_el->getId()] = curr_el;
#ifdef DEBUG_ITMS
    curr_el->showAtom(0);
#endif
    /* set the enforced proposition's label to lab. */
    curr_el->setLabel(lab);
#ifdef ENTROPY
    delete itms_res->res->trans_probs;
    itms_res->res->trans_probs = new hash_map<int, double, hash<int> >(*(*itms_res->res->temp_probs)[curr_el->getId()]);
#endif
#ifdef USE_BDD
    if (itms_res->res->observable) {
      addNode(itms_res->res->pb, curr_el->getId(), lab);
      if (curr_el->getVarType() != "Observable")
	addNode(itms_res->res->no_obs, curr_el->getId(), lab);
      else addNode(itms_res->res->obs, curr_el->getId(), lab);
    }
#endif    
    /* resupport dependencies. */
    resc = new hash_map<const char*, TrajEl*, hash<const char*>, eqstr>(*itms_res->already_changed);
    resupportDependence(curr_el, resc, itms_res->res, tr, timestep);
    res = new vector<TrajEl*>();  /* test all dependencies of the old label. */
    for (unsigned int j=0; j<curr_el->supported->size(); j++) {
      if ((*curr_el->supported)[j]->support != NULL) {
	if (!testClause((*curr_el->supported)[j]->support, itms_res->res, tr, timestep))
	  deleteClause((*curr_el->supported)[j], res, itms_res->res, tr, timestep);
      }
      else {
	deleteClause((*curr_el->supported)[j], res, itms_res->res, tr, timestep);
      }
    }
    itms_res->chg_tmp = resc;
    itms_res->deleted = res;
    if (count != 1)
      ag->push(itms_res);
    else {
      resi = itms_res;
    }  
   }
   if (flag_del_pis) {
     for (unsigned z=0; z<el_pis->size (); z++)
       delete (*el_pis)[z];
     delete el_pis;
   }
   el_pis = NULL;
  }
  not_unknown->clear();
  delete not_unknown;
  not_unknown = 0;
  curr_res->clear();
  delete curr_res;
  curr_res = 0;
  return resi;
}
/* unit propagation. */
void itms::unitPropagate(ItmsRes* curr_st, vector<ItmsRes*>* res, const bool &nom, 
			 const double &rank, Trajectory* tr, const int &timestep) {
  Clause* cla;
  Clause* curr_conf;
  ItmsRes* curr_res;
  State* wkst = curr_st->res;
  hash_map<const char*, TrajEl*, hash<const char*>, eqstr>* chg_tmp = curr_st->chg_tmp;
  hash_map<const char*, TrajEl*, hash<const char*>, eqstr>* already_changed = curr_st->already_changed;
  stack<ItmsRes*>* agenda = new stack<ItmsRes*>();
  bool flag = false;
  for (hash_map<const char*, TrajEl*, hash<const char*>, eqstr>::iterator add_it = already_changed->begin(); add_it != already_changed->end(); add_it++) {
    resupportDependence((*add_it).second, chg_tmp, wkst, tr, timestep);
  }
  checkForFocusedInternalConflicts(already_changed, curr_st->forced_cstrs, wkst, tr, timestep);
  checkForFocusedConflicts(already_changed, curr_st->forced_cstrs, wkst, tr, timestep); 
  agenda->push(curr_st);
 discarded: 
#ifdef DEBUG_ITMS
  cout << "a sol has been discarded...\n";
#endif
  while (!agenda->empty()) {
    if (flag) {
      checkForFocusedInternalConflicts(agenda->top()->chg_tmp, agenda->top()->forced_cstrs, agenda->top()->res, tr, timestep);
      checkForFocusedConflicts(agenda->top()->chg_tmp, agenda->top()->forced_cstrs, agenda->top()->res, tr, timestep);
      flag = false;
    }
    curr_res = agenda->top();
    agenda->pop();
    
    while ((curr_res->forced_cstrs->size() != 0) | ((cla = getFirstUnitClause(curr_res->deleted, curr_res->res, tr, timestep)) != NULL)) {
      if (cla != NULL) propagateClause(cla, curr_res->res, tr, timestep); //propagate through the unit clause. 
      while (!curr_res->forced_cstrs->empty()) {  
        curr_conf = curr_res->forced_cstrs->front();
	curr_res->forced_cstrs->pop();
	if (!testClause(curr_conf, curr_res->res, tr, timestep)) {  /* test conflict again in case the situation has changed. */
	  if ((curr_res = propagateThroughConflict(curr_conf, curr_res, agenda, nom, rank, tr, timestep)) != NULL) {
	  }
	  else {
	    if (!agenda->empty()) {
	      flag = true;   
	    }
	    goto discarded;
	  }
	}
      }
      checkForFocusedInternalConflicts(curr_res->chg_tmp, curr_res->forced_cstrs, curr_res->res, tr, timestep);
      checkForFocusedConflicts(curr_res->chg_tmp, curr_res->forced_cstrs, curr_res->res, tr, timestep);
    } //end while1.
#ifdef ENTROPY
    sortPostProbs(curr_res->res);
#endif    
    res->push_back(curr_res); //TODO : must return sols as soon as they are found.
    if (curr_res->res->observable) {
      /* compute p(Ot/si). */
      /*cout << "state atoms number : " << curr_res->res->atoms_number << endl;
      cout << "satcount pb : " << satcountset(*curr_res->res->pb, curr_res->res->atoms_number) << endl;
      cout << "satcount no_obs : " << satcountset(*curr_res->res->no_obs, curr_res->res->atoms_number) << endl;
      cout << "satcount bdd_state : " << satcountset(*curr_res->res->bdd_state, curr_res->res->atoms_number) << endl;*/
#ifdef PROBAS
#ifdef USE_BDD
      curr_res->res->obs_cond_prob = satcountset(*curr_res->res->pb, curr_res->res->atoms_number) / satcountset(*curr_res->res->no_obs, curr_res->res->atoms_number);
      curr_res->res->static_prob = satcountset(*curr_res->res->pb, curr_res->res->atoms_number) / satcountset(*curr_res->res->obs, curr_res->res->atoms_number);
#else
      computeObsCondRough (curr_res->res);
#endif
#endif
    }
#ifdef DEBUG_ITMS
    cout << "sol added !\n";
    cout << "agenda's size : " << agenda->size() << endl;
    cout << "already_changed's size end : " << curr_res->already_changed->size() << endl;
#endif
    flag = true;
  }
  delete agenda;
  agenda = 0;
}

vector<ItmsRes*>* itms::addPropositions(multimap<bool, TrajEl*, eqbool>* add_ats, hash_map<const char*, TrajEl*, hash<const char*>, eqstr>* already_changed, State* wkst, vector<ItmsRes*>* res, const bool &nom, const double &rank, const double &current_rank, Trajectory* tr, const int &timestep, const bool &comp_lock) {
#ifdef BENCHMK
  vector<const char*>* add_ats_names = new vector<const char*>();
  clock_t start, end;
  double elapsed;
  start = clock();
#endif
  
  if (comp_lock) lock_comp = 1;
  if (res == NULL) res = new vector<ItmsRes*>();

#ifdef USE_BDD
  if (!nom || wkst->observable || wkst->bdd_state == NULL)
    initBddPb (wkst); /* initialize bdd pb. */
#else
#ifdef PROBAS
  if (!nom || wkst->observable)
    initPb (wkst);
#endif
#endif

  ItmsRes* curr_st = new ItmsRes(wkst);
  for (multimap<bool, TrajEl*, eqbool>::iterator add_it = add_ats->begin(); 
       add_it != add_ats->end(); add_it++) { /* setting the labels of the propositions to be changed. */
#ifdef BENCHMK
    add_ats_names->push_back((*add_it).second->getId());
#endif
    if ((*add_it).first) {
#ifdef ENTROPY
#ifdef USE_BDD
      if (!nom)
	pb &= (*corresp_atoms)[(*add_it).second->getId()];
#else /* USE_BDD */
      /*if (!nom)*/
	
#endif
#endif      
      if ((*add_it).second->getLabel() == "false" && (*add_it).second->at->fail)
	if (tr->rank - tr->getStateRank(wkst) + wkst->rank + 1 > rank)
	  return res; /* abort because of rank */
#ifndef DIAG_EXO
      if ((*add_it).second->getStateVar ()->type == "Command" 
	  && ((CommandVar*)(*add_it).second->at->stateVar)->exo)
	return res;  /* abort because exo commands can't be changed */
#endif
      (*add_it).second->setLabel("true");
#ifdef USE_BDD
      if (wkst->observable) {
	addNode(wkst->pb, (*add_it).second->getId(), "true");
	if ((*add_it).second->getVarType() != "Observable")
	  addNode(wkst->no_obs, (*add_it).second->getId(), "true");
	else addNode(wkst->obs, (*add_it).second->getId(), "true");
      }
#endif
    }
    else {
#ifdef ENTROPY
#ifdef USE_BDD
      if (!nom)
	pb &= !(*corresp_atoms)[(*add_it).second->getId()];
#endif      
#endif
#ifndef DIAG_EXO
      if ((*add_it).second->getStateVar ()->type == "Command" 
	  && ((CommandVar*)(*add_it).second->at->stateVar)->exo)
	return res;  /* abort because exo commands can't be changed */
#endif
      (*add_it).second->setLabel("false");
#ifdef USE_BDD
      if (wkst->observable)
	addNode(wkst->pb, (*add_it).second->getId(), "false"); 
#endif    
    }
    (*curr_st->already_changed)[(*add_it).second->getId()] = (*add_it).second;
  }
  
  /*#ifdef ENTROPY
    curr_st->res->pb = new bdd(pb);
    #endif*/  

  unitPropagate(curr_st, res, nom, rank, tr, timestep);
#ifdef DEBUG_ITMS
  cout << "[addClause] --> res' size : " << res->size() << endl;
#endif
#ifdef BENCHMK
  end = clock();
  elapsed = ((double)(end - start))/CLOCKS_PER_SEC;
  addItmsBmkEntry(elapsed, res->size(), add_ats_names);
  delete add_ats_names;
  add_ats_names = 0;
#endif
  lock_comp = false;
  return res;
}

//delete a single elt and its dependencies.
vector<TrajEl*>* itms::deleteClause(TrajEl* del_at, vector<TrajEl*>* res, State* wkst, Trajectory* tr, const int &timestep) {
  vector<TrajEl*>* temp;
  /* set the atom to false. */
#ifdef DEBUG_ITMS
  cout << "[deleteClause] --> delete clause :\n";
  cout << del_at->getId() << " -- " << del_at->label << " -- " << del_at->lock << endl;
  if (del_at->support != NULL) {
    cout << "support : ";
    del_at->support->showClause(0);
  } 
  cout << endl;
#endif
  del_at->setLabel("unknown");
  del_at->support = NULL; //delete its support.
  del_at->propagation_number = 0; //set its propagation number to 0.
  res->push_back(del_at); //add the current deleted clause to res.
  for (unsigned int i=0; i<del_at->supported->size(); i++) { //do the same for all its dependences (supported).
    if ((*del_at->supported)[i]->support != NULL) 
      if (!testClause((*del_at->supported)[i]->support, wkst, tr, timestep)) {
	temp = deleteClause((*del_at->supported)[i], res, wkst, tr, timestep);
	for (unsigned int j=0; j<temp->size(); j++)
	  res->push_back((*temp)[j]);
      }
  }
#ifdef DEBUG_ITMS
  cout << del_at->supported->size() << endl;
  cout << del_at->supported << endl;
#endif
  del_at->supported->erase(del_at->supported->begin(), del_at->supported->end()); //erase the supported...
  return res;
}

//get the atoms that are predicted by an atom. TO BE OPTIMIZED.
vector<Atom*>* itms::predictedAtoms(Atom* p, vector<Atom*>* res) {
  if (res == NULL) 
    res = new vector<Atom*>();
  vector<Atom*>* temp;
  //seek through the ref_clauses to find opposite signed atoms.
  for (unsigned int i=0; i<p->ref_clauses->size(); i++) {
    temp = (*p->ref_clauses)[i]->getOpposites(p);
    //add results to res.
    for (unsigned int j=0; j<temp->size(); j++)
      if (!inVector((*temp)[j], res)) {
	res->push_back((*temp)[j]);
	//predictedAtoms((*temp)[j], res);
      }
  }
  //add p to res !
  if (!inVector(p, res)) res->push_back(p);
  return res;
}
//return an estimation of the conditional probability of the atom given a set of observations.
double itms::probaCondObsAtom(Atom* at, vector<Atom*>* obs) {
  int n = 0; //number of obs in predict(at).
  vector<Atom*>* pred;
  //if at is not already true, get the probability of the necessary transition =  p(at).
  
  //if predict(at) is not empty, evaluate the conditional probability.
  pred = predictedAtoms(at, NULL);
  if (!pred->empty()) {
    for (unsigned int i=0; i<pred->size(); i++)
      if (inVector((*pred)[i], obs))
	n++;
  }
  else n = 0;
  //cout << "[probaCondObsAtom --> n : " << n << endl;
  //cout << "[probaCondObsAtom --> obs : " << obs->size() << endl;
  //delete pred;
  return (double)((double)n / (double)obs->size());
}


/****************************
 * Probabilities management *
 ****************************/
#ifdef ENTROPY
/* heuristic to get p(v=vi). */
#ifdef USE_BDD
double itms::atomProba(Atom* at, State* wkst, hash_map<const char*, TrajEl*, hash<const char*>, eqstr>* alchg) {
  /* it is a prevision : values are inversed. */
  if (alchg->count(at->getId()) != 0) {
    if ((*wkst->state)[at->getId()]->getLabel() == "true")
      return computeProb(at);
    else return 1 - computeProb(at);
  }
  else {
    if ((*wkst->state)[at->getId()]->getLabel() == "true")
      return 1 - computeProb(at);
    else return computeProb(at);
  }
}
#endif

/* compute H(v=vi). */
/* beware sum of posteriors is no more equal to 1... */
double itms::entropy(Atom* at, State* wkst, hash_map<const char*, TrajEl*, hash<const char*>, eqstr>* alchg) {
  double entropy = 0;
  double p_at = 0;
  double cond_prob = 0;
  int id;
  Transition* curr_trans;
  for (hash_map<const char*, Comp*, hash<const char*>, eqstr>::iterator it1 = wkst->comps->begin(); it1 != wkst->comps->end(); it1++) {
    for (unsigned int i = 0; i<(*it1).second->component_transitions->size(); i++) {
      curr_trans = (*(*it1).second->component_transitions)[i];
      id = curr_trans->trans_id;
      if ((*wkst->trans_probs)[id] != 0) {
	cond_prob = 1;
	/*cond_prob = computeCondProb((*wkst->state)[at->getId()], curr_trans, wkst, alchg);*/
	/* store p(v=vi|tau)*pl. */
	(*(*wkst->temp_probs)[at->getId()])[id] = (*wkst->trans_probs)[id] * cond_prob;
	/* compute p(v=vi). */
	p_at = p_at + (*(*wkst->temp_probs)[at->getId()])[id];
      }
    }
  }
  /* normalize with p_at. */
  for (hash_map<int, double, hash<int> >::iterator it_norm = (*wkst->temp_probs)[at->getId()]->begin(); it_norm != (*wkst->temp_probs)[at->getId()]->end(); it_norm++) {
    (*it_norm).second = (*it_norm).second / p_at;
  }
  /* compute H(v=vi). */
  for (hash_map<const char*, Comp*, hash<const char*>, eqstr>::iterator it2 = wkst->comps->begin(); it2 != wkst->comps->end(); it2++) {
    for (unsigned int j = 0; j<(*it2).second->component_transitions->size(); j++) {
      curr_trans = (*(*it2).second->component_transitions)[j];
      id = curr_trans->trans_id;
      if ((*wkst->trans_probs)[id] != 0) {
	cond_prob = (*(*wkst->temp_probs)[at->getId()])[id];
	if (cond_prob != 0)
	  entropy = entropy + cond_prob * log(cond_prob);
      }
    }
  }
  //cout << "entropy for " << at->getId() << " : " << -entropy << endl;
  return -entropy;
}
/* compute He(v), i.e. the expected entropy of v. */
double itms::entropy(StateVar* st, State* wkst, hash_map<const char*, TrajEl*, hash<const char*>, eqstr>* alchg) {
  double res = 0;
  double p_at = 0;
  double cond_prob = 0;
  double entropy = 0;
  int id;
  Transition* curr_trans;
  for (unsigned int i=0; i<st->atoms_links->size(); i++) {
    p_at = 0;
    if (wkst->temp_probs->count((*st->atoms_links)[i]->getId()) == 0) {
      delete (*wkst->temp_probs)[(*st->atoms_links)[i]->getId()];
      (*wkst->temp_probs)[(*st->atoms_links)[i]->getId()] = 0;
    }
    (*wkst->temp_probs)[(*st->atoms_links)[i]->getId()] = new hash_map<int, double, hash<int> >();
    for (hash_map<const char*, Comp*, hash<const char*>, eqstr>::iterator it1 = wkst->comps->begin(); it1 != wkst->comps->end(); it1++) {
      for (unsigned int k = 0; k<(*it1).second->component_transitions->size(); k++) {
	curr_trans = (*(*it1).second->component_transitions)[k];
	id = curr_trans->trans_id;
	if ((*wkst->trans_probs)[id] != 0) {
#ifdef USE_BDD
	  cond_prob = computeCondProb((*wkst->state)[(*st->atoms_links)[i]->getId()], curr_trans, wkst, alchg);
#else /* USE_BDD */
	  cond_prob = computeCondProbRough ((*wkst->state)[(*st->atoms_links)[i]->getId()], curr_trans, wkst, alchg);
#endif
	  /* store p(v=vi|tau)*pl. */
	  (*(*wkst->temp_probs)[(*st->atoms_links)[i]->getId()])[id] = (*wkst->trans_probs)[id] * cond_prob;
	  /* compute p(v=vi). */
	  p_at = p_at + (*(*wkst->temp_probs)[(*st->atoms_links)[i]->getId()])[id];
	}
      }
    }
    /* normalize with p_at. */
    for (hash_map<int, double, hash<int> >::iterator it_norm = (*wkst->temp_probs)[(*st->atoms_links)[i]->getId()]->begin(); it_norm != (*wkst->temp_probs)[(*st->atoms_links)[i]->getId()]->end(); it_norm++) {
      (*it_norm).second = (*it_norm).second / p_at;
    }
    /* computing H(v=vi). */
    entropy = 0;
    for (hash_map<const char*, Comp*, hash<const char*>, eqstr>::iterator it2 = wkst->comps->begin(); it2 != wkst->comps->end(); it2++) {
      for (unsigned int j = 0; j<(*it2).second->component_transitions->size(); j++) {
	curr_trans = (*(*it2).second->component_transitions)[j];
	id = curr_trans->trans_id;
	if ((*wkst->trans_probs)[id] != 0) {
	  cond_prob = (*(*wkst->temp_probs)[(*st->atoms_links)[i]->getId()])[id];
	  //cout << "cond prob2 : " << cond_prob << endl;
	  if (cond_prob != 0)
	    entropy = entropy + cond_prob * log(cond_prob);
	}
      }
    }
    res = res - p_at * entropy;
  }
  return res;
}

void itms::computeObsCondRough (State* wkst) {
  double osbi = 0;
  wkst->obs_cond_prob = 1;

  for (hash_map<const char*, TrajEl*, hash<const char*>, eqstr>::iterator ocit = wkst->state->begin ();
       ocit != wkst->state->end (); ocit++) {
    /*if ((*ocit).second->getVarType () == "Observable" && (*ocit).second->getLabel () == "true") {*/
    if ((*ocit).second->lock == 0 || (*ocit).second->lock == 3) {
      osbi = 0;
       for (hash_map<const char*, TrajEl*, hash<const char*>, eqstr>::iterator cit = wkst->state->begin(); 
	    cit != wkst->state->end(); cit++) {
	 if ((*cit).second->getVarType () == "Mode" && (*cit).second->getLabel () == "true") { 
	   for (unsigned int i=0; i<(*cit).second->at->ref_trans_post->size(); i++) {
	     osbi += computeCondProbRough ((*ocit).second, (*(*cit).second->at->ref_trans_post)[i], wkst, NULL);
	   }
	 }
       }
       wkst->obs_cond_prob *= osbi;
    }
  }
}

/**
* evaluate the conditional probability : p(elt|trans).
*/
double itms::computeCondProbRough (TrajEl* elt, Transition* trans, State* wkst, hash_map<const char*, TrajEl*, hash<const char*>, eqstr>* alchg) {
  vector<vector<Atom*>*>* el_pis = NULL;
  multimap<bool, Atom*, eqbool>* postt = trans->post->getAtoms ();  /* posterior atoms of the transition */

  /* seek for elt compiled prime implicants */
  if (pit != NULL && (el_pis = pit->getPIs (elt->at)) != NULL) {

  }
  else {
    bool inopps = checkInOpposites (elt->at, postt);
    bool inimports = checkInImportedComps (elt->at, postt);
    
    if (inimports && inopps)  /* in opposites atoms and in imported components */
      return 1;
    else if (inimports && ! inopps)
      return 0;
    else /* ! inimports */
      return 1/(double) elt->getStateVar ()->atoms_links->size ();
  }
  cout << "rough proba computation - strange end\n";
  return 0;
}

/* check for opposite atoms (reduced to horn clauses). Depth is 1... */
bool itms::checkInOpposites (Atom* at, multimap<bool, Atom*, eqbool>* postt) {
  Atom* curr_at = 0;
  vector<Atom*>* opps;
  
  for (multimap<bool, Atom*, eqbool>::iterator it = postt->begin ();
       it != postt->end (); it++) {
    curr_at = (*it).second; 
    for (unsigned int i=0; i<curr_at->ref_clauses->size (); i++) {
      opps = (*curr_at->ref_clauses)[i]->getOpposites (curr_at);   /* horn clauses only... */
      if (inVector (at, opps))
	return true;
      delete opps;
    }
  }
  return false;
}

bool itms::checkInImportedComps (Atom* at, multimap<bool, Atom*, eqbool>* postt) {
  Comp* curr_co = 0;
  for (multimap<bool, Atom*, eqbool>::iterator it = postt->begin ();
       it != postt->end (); it++) {
    curr_co = (*it).second->getVarComp ();
    
    for (unsigned int i=0; i<curr_co->state_variables->size (); i++) { /* inside current component */
      if (at->getStateVar () == (*curr_co->state_variables)[i])
	return true;
    }
    
    for (unsigned int i=0; i<curr_co->imported_comps->size (); i++) {  /* inside imports */
      for (unsigned int j=0; j<(*curr_co->imported_comps)[i]->state_variables->size (); j++) {
	if (at->getStateVar () == (*(*curr_co->imported_comps)[i]->state_variables)[j])
	  return true;
      }
    }
  }
  
  return false;
}

#ifdef USE_BDD
/* compute the transition posterior probability using BDDs. */
double itms::condProbTrans(Transition* trans, TrajEl* elt, State* wkst, 
			   const double &atom_prob, 
			   hash_map<const char*, TrajEl*, hash<const char*>, eqstr>* alchg) {
  double res;
  if (atom_prob != 0) {
    /* get previous posterior probability. */
    res = (*wkst->trans_probs)[trans->trans_id];
    /* update bdd with elt. */
    res = res * (computeCondProb(elt, trans, wkst, alchg) / atom_prob);
    return res;
  }
  /* should'nt happen anyway. */
  else {
    return 0;
  }
}
#endif

/* initialize the state probabilities according to the current transitions assignments. */
void itms::initStateProbas(State* wkst) {
  for (hash_map<const char*, Comp*, hash<const char*>, eqstr>::iterator it1 = wkst->comps->begin(); it1 != wkst->comps->end(); it1++) {
    for (unsigned int i = 0; i<(*it1).second->component_transitions->size(); i++) {
      (*wkst->trans_probs)[(*(*it1).second->component_transitions)[i]->trans_id] = (*(*it1).second->component_transitions)[i]->proba;
    }
  }
}
/* store results into a sorted container. */
void itms::sortPostProbs(State* wkst) {
  wkst->sorted_trans_probs->clear();
  for (hash_map<int, double, hash<int> >::iterator it = wkst->trans_probs->begin();
       it != wkst->trans_probs->end(); it++) {
      wkst->sorted_trans_probs->insert(pair<double, int>((*it).second, (*it).first));
  }
}
#endif

/*******************************************************
 * Trajectories management and consistency maintenance *
 *******************************************************/
//eval a stateformula considering a trajectory's temporal state. recursive formulas ARE NOT handled yet.
bool itms::evalStateFormula(StateFormula* stfrm, Trajectory* tr, int timestep) {
  bool resAtoms;
  bool tmp;
  if (stfrm->operat == "and" | stfrm->operat == "") {
    if (stfrm->atoms!=NULL) {
      resAtoms=true;
      for (multimap<bool, Atom*, eqbool>::iterator i = stfrm->atoms->begin(); i != stfrm->atoms->end(); i++) {
	if ((*i).first)
	  tmp = tr->getElt((*i).second->getId(), timestep, 0)->evalTrajEl();
	else
	  tmp = !tr->getElt((*i).second->getId(), timestep, 0)->evalTrajEl();
	resAtoms = resAtoms & tmp;
      }
      return resAtoms;
    }
  }
  else if (stfrm->operat=="or") {
    if (stfrm->atoms!=NULL) {
      resAtoms=false;
      for (multimap<bool, Atom*, eqbool>::iterator i = stfrm->atoms->begin(); i != stfrm->atoms->end(); i++) {
	if ((*i).first)
	  tmp = tr->getElt((*i).second->getId(), timestep, 0)->evalTrajEl();
	else tmp = !tr->getElt((*i).second->getId(), timestep, 0)->evalTrajEl();
	resAtoms = resAtoms | tmp;
      }
      return resAtoms;
    }
  }
  return 0;
}

/************************************************************
 Read prime implicant file and build the fill in the table
************************************************************/

int itms::readPI () {
  const char* name = model_name.c_str ();
  const char* PISUFFIX=".pi";
  char FileName[strlen(name)+strlen(PISUFFIX)+1];
  vector<string>* table1;
  vector<string>* vec;
  string file_pi;
  char* bufptr;
  FILE* fd;
  int buf_size=1024;
  Atom *VD, *VD1;
  vector<vector<Atom*>*>* pis;

  pit = new PiTable ();

  strcpy(FileName, name);
  strcat(FileName, PISUFFIX);
  if ((fd = fopen(FileName, "r")) == NULL) {
    cout << "can't open " << FileName << endl;
    return 0;
  }

  cout << "*** File " << FileName << " found. Reading file...\n";
  file_pi = string::basic_string();
  bufptr = (char*)malloc(buf_size*sizeof(char));
  while (fgets(bufptr, buf_size, fd) != NULL) {
    file_pi+=string(bufptr); // copying the file in a string
  }
  fclose(fd);

  //spliting the string at "\n"s and put them in table1
  table1 = new vector<string> ();
  string::size_type pos;
  string::size_type debut = 0;
  string str, strname;
  pos = file_pi.find("\n");
  while (pos != file_pi.npos) {
    str =  string::basic_string(file_pi.substr(debut, pos - debut));
    table1->push_back(str);
    debut = pos+1;
    pos = file_pi.find("\n", debut);
  }

  // Processing each string in table1
  for(unsigned int i=0; i<table1->size(); i++) {
    pis = new vector<vector<Atom*>*> ();
    vec = new vector<string> ();
    debut = 0;
    pos=(*table1)[i].find(":"); // find the Id of dependent variable
    str =  string::basic_string((*table1)[i].substr(0, pos - debut));
    debut = pos+1;
    pos = str.find (".");
    strname = str.substr (0, pos);
    bool cp_f = false;
    
    for (hash_map<const char*, Comp*, hash<const char*>, eqstr>::iterator c_it = active_comps->begin ();
	 c_it != active_comps->end (); c_it++)
      if ((*c_it).second->var_name == strname) {
	strname = (*c_it).second->name + str.substr (pos+1, str.length ());
	cp_f = true;
	break;
      }
    if (! cp_f)
      cout << "Component not found: " << strname << endl;

    if((VD=getAtom(str.c_str()))==NULL) {
      cerr << "Atom not found: " << str << endl;
      delete vec;
      return -1;
    }
    // puts the prime implicants string in a vector
    pos = (*table1)[i].find(",",debut);
    while(pos != (*table1)[i].npos){
      str = string::basic_string((*table1)[i].substr(debut, pos - debut));
      vec->push_back(str);
      debut = pos+1;
      pos = (*table1)[i].find(",", debut);
    }
    str = string::basic_string((*table1)[i].substr(debut, pos - debut));
    vec->push_back(str);
    if ((*vec)[0].length()!=0) {
      // parse the prime implicants and link them to their respectiv Atom*
      for (unsigned int disj=0; disj<vec->size(); disj++) {
	pis->push_back(new vector<Atom*>);
	debut = 0;
	pos = (*vec)[disj].find(" ");
	while(pos != (*vec)[disj].npos){
	  str = string::basic_string((*vec)[disj].substr(debut, pos - debut));
	  debut = pos+1;
	  pos = (*vec)[disj].find(" ", debut);
	  if ((VD1=getAtom(str.c_str()))!=NULL) {
	    (*pis)[disj]->push_back(VD1);
	  }
	  else {
	    cerr << "Atom not found: " << str << endl;
	    delete vec;
	    return -1;
	  }
	}
	str = string::basic_string((*vec)[disj].substr(debut, pos - debut));
	if ((VD1=getAtom(str.c_str()))!=NULL) {
	  (*pis)[disj]->push_back(VD1);
	}
	else {
	  cerr << "Atom not found: " << str << endl;
	  delete vec;
	  return -1;
	}
      }
      pit->addPIAtom(VD, pis); // update the PiTable
     
    }
    delete vec;
  }

  pit->showAllPI();
  return 1;
}

/*************
 * debugging *
 *************/

//show vectors of atoms.
void itms::viewAtomVector(vector<Atom*>* vec) {
  for (unsigned int i=0; i<vec->size(); i++) {
    (*vec)[i]->showAtom(1);
    cout << "support : \n";
    if ((*vec)[i]->support != NULL)
      (*vec)[i]->support->showClause(0);
  }
  cout << endl;
}
void itms::viewAtomVectorId(vector<Atom*>* vec) {
  for (unsigned int i=0; i<vec->size(); i++) 
    cout << (*vec)[i]->getId() << endl;
}
//view a vector of clauses.
void itms::viewClauseVector(vector<Clause*>* vec) {
  for (unsigned int i=0; i<vec->size(); i++) {
    (*vec)[i]->showClause(0);
  }
  cout << endl;
}
//view a multimap
void itms::viewMultimapAtom(multimap<string, Atom*, eqstring>* mlt_map) {
  for (multimap<string, Atom*, eqstring>::iterator it = mlt_map->begin(); it != mlt_map->end(); it++) {
    (*it).second->showAtom(1);
    cout << "support : \n";
    if ((*it).second->support != NULL)
      (*it).second->support->showClause(0);
  }
  cout << endl;
}
//view a hash_map of atoms.
void itms::viewHashMapAtom(hash_map<const char*, Atom*, hash<const char*>, eqstr>* set) {
  for (hash_map<const char*, Atom*, hash<const char*>, eqstr>::iterator it = set->begin(); it != set->end(); it++) {
    /*cout << (*it).first << endl;
      cout << (*it).second << endl;*/
    (*it).second->showAtom(1);
    cout << "support : \n";
    if ((*it).second->support != NULL)
      (*it).second->support->showClause(0);
  }
  //cout << "********************* end **********************" << endl;
}
//view conflicts of a trajectory (hash_map of vectors of Clause).
void itms::viewTrajectoryConflicts(hash_map<int, vector<Clause*>*, hash<int> >* confs) {
  for (hash_map<int, vector<Clause*>*, hash<int> >::iterator it = confs->begin(); it != confs->end(); it++) {
    if ((*it).second != NULL) {
      cout << "conflict's timestep : " << (*it).first << endl;
      viewClauseVector((*it).second);
    }
  }
}

