
/*    
 *  Copyright 2000, 2001, 2002 Emmanuel Benazera  
 *
 *  This file is part of Koala.
 *
 *  Koala is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Koala is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Koala; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <iostream>
#include <strstream>
#include <fstream>
#include "DataManagerrt.h"

hash_map<const char*, Comp*, hash<const char*> , eqstr >* DataManager::active_comps = new hash_map<const char*, Comp*, hash<const char*>, eqstr >();
hash_map<const char*, Atom*, hash<const char*>, eqstr>* DataManager::active_atoms = new hash_map<const char*, Atom*, hash<const char*>, eqstr>();
hash_map<int, Transition*, hash<int> >* DataManager::active_trans = new hash_map<int, Transition*, hash<int> >();
vector<Clause*>* DataManager::cls = new vector<Clause*>();
multimap<double, Trajectory*, greater<double> >* DataManager::tracked_traj = new multimap<double, Trajectory*, greater<double> >();
double DataManager::traj_cover = 0;
hash_map<int, double, hash<int> >* DataManager::states_cover = new hash_map<int, double, hash<int> >();

//efficiency testing...
int DataManager::trajectories_checked = 0;
int DataManager::transitions_checked = 0;
int DataManager::trajectories_created = 1; //1 is for nominal.
int DataManager::trajectories_destroyed = 0;

string DataManager::model_name;

#ifdef BENCHMK
   ofstream DataManager::benchmk_stream("bench1.bmk", ios::trunc);
   ofstream DataManager::itms_stream("itmsbench.bmk", ios::trunc);
   unsigned int DataManager::itms_calls = 0;
#endif

/* Functions for class DataManager */

DataManager::DataManager() {
  init();
}
//destructor.
DataManager::~DataManager() {
  cout << "calling destructor of DataManager\n";
  for (hash_map<const char*, Comp*, hash<const char*>, eqstr>::iterator comps_it = active_comps->begin(); comps_it != active_comps->end(); comps_it++)
    delete (*comps_it).second;
  delete active_comps;
  active_comps = 0;
  for (hash_map<const char*, Atom*, hash<const char*>, eqstr>::iterator atoms_it = active_atoms->begin(); atoms_it != active_atoms->end(); atoms_it++)
    delete (*atoms_it).second;
  delete active_atoms;
  active_atoms = 0;
  for (unsigned int i=0; i<cls->size(); i++) {
    delete (*cls)[i];
    (*cls)[i] = 0;
  }
  delete cls;
  cls = 0;
  for (multimap<double, Trajectory*, greater<double> >::iterator tr_it = tracked_traj->begin(); tr_it != tracked_traj->end(); tr_it++)
    delete (*tr_it).second;
  delete tracked_traj;
  cout << "tracked_traj deleted\n";
  tracked_traj = 0;
  /*delete nominal_traj;
    nominal_traj = 0;*/
}
/* initialize the data structures. */
bool DataManager::init() {
#ifdef PROBAS
  /* init the states cover. */
  (*states_cover)[0] = 0;
#endif  
  return 1;
}

/************************** 
 *  Access to data. * 
 **************************/

/* add a component to the system */
void DataManager::addComp(Comp* co) {
  (*DataManager::active_comps)[co->getName()] = co;
}

Comp* DataManager::getComp(const char* st) {
  return (*DataManager::active_comps)[st];
}
/* add an atom to the atoms' table */
void DataManager::addAtom(Atom* at) {
  (*DataManager::active_atoms)[at->getId()] = at;
}
Atom* DataManager::getAtom(const char* id) {
  return (*DataManager::active_atoms)[id];
}
/* test if at is in vec. */
bool DataManager::inVector(Atom* at, vector<Atom*>* vec) {
  for (unsigned int i=0; i<vec->size(); i++)
    if ((*vec)[i] == at) return true;
  return false;
}
//test if a const char* is in a list...
bool DataManager::inList(const char* elt, list<const char*>* li) {
  for (list<const char*>::iterator i = li->begin(); i != li->end(); i++)
    if (strcmp(elt, (*i)) == 0)
      return true;
  return false;
}
/* test if an atom is in a multimap. */
bool DataManager::inMultimap(Atom* at, multimap<string, Atom*, eqstring>* mult_map) {
  for (multimap<string, Atom*, eqstring>::iterator it = mult_map->begin(); it != mult_map->end(); it++)
    if ((*it).second == at) return true;
  return false;
}

/* update trajectories mean cover. Made from scratch :(( could be optimized. */
void DataManager::updateTrajectoriesCover() {
  for (hash_map<int, double, hash<int> >::iterator it = states_cover->begin(); it != states_cover->end(); it++)
    traj_cover = traj_cover + (*it).second;
  traj_cover = traj_cover / (double)states_cover->size();
}
/* update mean cover of a state. */
void DataManager::updateStatesCover(int time_step, Transition* trans) {
  (*states_cover)[time_step] = (*states_cover)[time_step] + (trans->weight / (double)active_comps->size());
}

#ifdef BENCHMK
void DataManager::addBmkEntry(int t_step, double time, bool diag, unsigned int traj_num) {
  benchmk_stream << t_step << " " << time << " " << diag << " " << traj_num << endl;
  //cout << "[DataManager] --> added to benchmark_file : " << t_step << " " << time << " " << diag << " " << traj_num << endl;;
}
void DataManager::addItmsBmkEntry(double time, unsigned int res_size, vector<const char*>* add_ats_names) {
  itms_stream << itms_calls++ << " " << time << " " << res_size << " ";
  for (unsigned int i=0; i<add_ats_names->size(); i++) {
    itms_stream << (*add_ats_names)[i] << " ";
  }
  itms_stream << endl;
}
#endif

/********************************************* 
 *   Viewers for the data and for debugging. * 
 *********************************************/

void DataManager::showComps () {
  for (hash_map<const char*, Comp*, hash<const char*>, eqstr>::iterator i = active_comps->begin(); i != active_comps->end(); i++)
    cout << (*i).second->type << " -- " << (*i).second->getName() << endl;
  cout << endl;
}

void DataManager::showComp (const char* na) {
  if (active_comps->count (na) != 0)
    (*active_comps)[na]->showComp();
  else dataError (1, 1);
}

void DataManager::showClauses () {
  for (unsigned int i=0; i<cls->size (); i++)
    (*cls)[i]->showClause (0);
}

void DataManager::showAtoms () {
  for (hash_map<const char*, Atom*, hash<const char*>, eqstr>::iterator i = active_atoms->begin(); i != active_atoms->end(); i++) {
    (*i).second->showAtom(0);
    cout << " ";
  }
  cout << endl;
}

//cout for debug.
void DataManager::printDebug(int count, ...) {
  #ifdef DEBUG_MODE
  if (debug) {
    va_list ap;
    int i;
    va_start(ap, count);
    for (i=0; i<count; i++)
      cout << va_arg(ap, const char*);
    cout << endl;
    va_end(ap);
  }
  #endif
}
//show the current state of the model.
/*void DataManager::modelState() {
  cout << "*************************** model state (state variables) ******************************\n";
  hash_map<const char*, Comp*, hash<const char*>, eqstr >::iterator it = active_comps->begin();
  for (int i=0; i<active_comps->size(); i++) {
    for (int j=0; j<(*it).second->state_variables->size(); j++)
      (*(*it).second->state_variables)[j]->showCurrentState();
    it++;
  }
  }*/
//show the current state of the atoms.
/*void DataManager::modelStateAtoms() {
  cout << "*************************** model state (atoms) ******************************\n";
  ITMS->viewHashMapAtom(active_atoms);
  }*/
//debugging : view a multimap of vector of atoms... (operator is greater<double>).
void DataManager::viewMultimapAtomVector(multimap<double, vector<Atom*>*, greater<double> >* mapy) {
  cout << "*************************** debugging : multimap view ******************************\n";
  for (multimap<double, vector<Atom*>*, greater<double> >::iterator it = mapy->begin(); it != mapy->end(); it++) {
    cout << (*it).first << endl;
    viewAtomVector((*it).second);
  }
}
//debugging : view active atoms.
void DataManager::viewActiveAtoms(bool bb) {
  cout << "*************************** DEBUGGING : active atoms ******************************\n";
  if (bb) {
    hash_map<const char*, Atom*, hash<const char*>, eqstr >::iterator it = active_atoms->begin();
    for (unsigned int i=0; i<active_atoms->size(); i++) {
      (*it).second->showAtom(1);
      it++;
    }
  }
  cout << active_atoms->size() << " atoms.\n";
}
 
//show vectors of atoms.
void DataManager::viewAtomVector(vector<Atom*>* vec) {
  for (unsigned int i=0; i<vec->size(); i++) 
    cout << (*vec)[i]->getId() << endl;
  cout << endl;
}
void DataManager::viewAtomVector2(vector<vector<Atom*>*>* vec) {
  for (unsigned int i=0; i<vec->size(); i++) {
    for (unsigned int j=0; j<(*vec)[i]->size(); j++)
      (*(*vec)[i])[j]->showAtom(1);
    cout << endl;
  }
  cout << endl;
} 
//show a multimap with atoms.
void DataManager::viewMultimapAtom(multimap<string, Atom*, eqstring>* mlt_map) {
  for (multimap<string, Atom*, eqstring>::iterator it = mlt_map->begin(); it != mlt_map->end(); it++) {
    (*it).second->showAtom(1);
    cout << "support : \n";
    if ((*it).second->support != NULL)
      (*it).second->support->showClause(1);
  }
  cout << endl;
}
//view the 'true' atoms from the sols.
void DataManager::viewTrueAtoms(multimap<string, Atom*, eqstring>* mlt_map) {
  cout << "size : " << mlt_map->size() << endl;
  for (multimap<string, Atom*, eqstring>::iterator it = mlt_map->begin(); it != mlt_map->end(); it++) {
    if ((*it).first != "unknown") {
      cout << (*it).first << endl;
      (*it).second->showAtom(1);
      cout << "support : \n";
      if ((*it).second->support != NULL)
	(*it).second->support->showClause(1);
    }
  }
}
/* view a hash_map of TrajEl. */
void DataManager::viewHashMapTrajEl(hash_map<const char*, TrajEl*, hash<const char*>, eqstr>* set) {
  for (hash_map<const char*, TrajEl*, hash<const char*>, eqstr>::iterator it = set->begin(); it != set->end(); it++) {
    (*it).second->showTrajEl(1);
  }
}
//view a hash_map of TrajEl.
void DataManager::viewHashMapTrajElLow(hash_map<const char*, TrajEl*, hash<const char*>, eqstr>* set) {
  for (hash_map<const char*, TrajEl*, hash<const char*>, eqstr>::iterator it = set->begin(); it != set->end(); it++) {
    cout << (*it).first << " -- " << (*it).second->label << " -- " << (*it).second->lock << " -- " << (*it).second->proba << endl;
  }
}
/* view the sols (conflicts resolution). */
void DataManager::viewTrueSol(multimap<double, multimap<string, Atom*, eqstring>*, greater<double> >* mlt_map) {
  for (multimap<double, multimap<string, Atom*, eqstring>*, greater<double> >::iterator it = mlt_map->begin(); it != mlt_map->end(); it++) {
    cout << (*it).first << endl;
    viewTrueAtoms((*it).second);
  }
}
void DataManager::viewTrueSol(multimap<double, hash_map<const char*, TrajEl*, hash<const char*>, eqstr>*, greater<double> >* state) {
  for (multimap<double, hash_map<const char*, TrajEl*, hash<const char*>, eqstr>*, greater<double> >::iterator it = state->begin(); it != state->end(); it++) {
    cout << (*it).first << endl;
    viewHashMapTrajEl((*it).second);
  }
}
void DataManager::viewTrajectoriesTransitionsMultimap(multimap<double, Trajectory*, greater<double> >* trajs) {
  for (multimap<double, Trajectory*, greater<double> >::iterator it = trajs->begin(); it != trajs->end(); it++)
    (*it).second->showTrajectoryTransitions();
}
void DataManager::viewTrajectoriesMultimap(multimap<double, Trajectory*, greater<double> >* trajs) {
  for (multimap<double, Trajectory*, greater<double> >::iterator it = trajs->begin(); it != trajs->end(); it++)
    for (int i=0; i<=(*it).second->getCurrentTime(); i++)
      (*it).second->showTrajectory(i);
}
void DataManager::viewTrajectoriesMultimapFull(multimap<double, Trajectory*, greater<double> >* trajs) {
  
  for (multimap<double, Trajectory*, greater<double> >::iterator it = trajs->begin(); it != trajs->end(); it++) {
    (*it).second->showTrajectoryTransitions();
    for (int i=(*it).second->low_time; i<=(*it).second->getCurrentTime(); i++)
      (*it).second->showTrajectory(i);
  }
}

/* tools */
bool DataManager::compareTrajsStates(State* st1, State* st2) {
  TrajEl* curr_el;
  for (hash_map<const char*, TrajEl*, hash<const char*>, eqstr>::iterator it1 = st1->state->begin(); it1 != st1->state->end(); it1++) {
    curr_el = st2->getElt((*it1).second->getId());
    if (curr_el == NULL || (curr_el->getLabel() != (*it1).second->getLabel()))
      return false;
  }
  for (hash_map<const char*, TrajEl*, hash<const char*>, eqstr>::iterator it1 = st2->state->begin(); it1 != st2->state->end(); it1++) {
    curr_el = st1->getElt((*it1).second->getId());
    if (curr_el == NULL || (curr_el->getLabel() != (*it1).second->getLabel()))
      return false;
  }
  return true;
}

void DataManager::compareTrajs(multimap<double, Trajectory*, greater<double> >* res) {
  bool comp;
  int i = 1;
  int j = 1;
  cout << "Trajectories comparison :\n";
  for (multimap<double, Trajectory*, greater<double> >::iterator it1 = res->begin(); it1 != res->end(); it1++) {
    j = i;
    comp = false;
    for (multimap<double, Trajectory*, greater<double> >::iterator it2 = it1; it2 != res->end(); it2++) {
      if (i != j) {
	cout << "comparing traj " << i << " and " << j <<endl;
	/* compare all states. */
	for (unsigned int t=0; t<(*it1).second->traj->size(); t++) {
	  if (!compareTrajsStates((*it1).second->getState(t), (*it2).second->getState(t))) {
	    comp = false;
	    break;
	  }
	  comp = true;
	}
	if (comp) {
	  cout << "Traj " << i << " is the same as Traj " << j << " !!!\n";
	  (*it1).second->showTrajectoryTransitions();
	  cout << endl;
	}
      }
      j++;
    }
    i++;
  }
  cout << endl;
}

void DataManager::testTransitionsProbas() {
  cout << "[testTransitionsProbas] --> testing...\n";

  double* tabd;
  double* tabd_copy;
  unsigned int tabd_size = 1;
  unsigned int current_size = 1;
  unsigned int pos = 0;
  double res = 0;

  /* get the size. */
  for (hash_map<const char*, Comp*, hash<const char*>, eqstr>::iterator it1 = active_comps->begin(); it1 != active_comps->end(); it1++)
    tabd_size = tabd_size * (*it1).second->component_transitions->size();
  tabd = (double*) malloc(tabd_size * sizeof(double));

  cout << "[testTransitionsProbas] --> tabd size : " << tabd_size << endl;

  for (hash_map<const char*, Comp*, hash<const char*>, eqstr>::iterator it = active_comps->begin(); it != active_comps->end(); it++) {
    tabd_copy = (double*) malloc(current_size * sizeof(double));
    for (unsigned int i=0; i<(*it).second->component_transitions->size(); i++) {
      if (i == 0 & current_size != 1) {
	/* copy existing tabd. */
	for (unsigned j=0; j<current_size; j++) {
	  tabd_copy[j] = tabd[j];
	  tabd[j] = tabd[j] * (*(*it).second->component_transitions)[i]->proba_guard;
	  //tabd[j] = tabd[j] * (*(*it).second->component_transitions)[i]->proba;
	  //tabd[j] = tabd[j] * (*(*it).second->component_transitions)[i]->weight;
	  //tabd[j] = tabd[j] * (*(*it).second->component_transitions)[i]->proba / (double) (*(*it).second->component_transitions)[i]->modevar->values.size();
	}
	pos = 0;
      }
      else if (current_size == 1) {
	tabd[i] = (*(*it).second->component_transitions)[i]->proba_guard;
	//tabd[i] = (*(*it).second->component_transitions)[i]->proba;
	//tabd[i] = (*(*it).second->component_transitions)[i]->weight;
	//tabd[i] = (*(*it).second->component_transitions)[i]->proba / (double) (*(*it).second->component_transitions)[i]->modevar->values.size();
      }
      else {
	for (unsigned int k=pos; k<pos + current_size; k++)
	  tabd[k] = tabd_copy[k - pos] * (*(*it).second->component_transitions)[i]->proba_guard;
	  //tabd[k] = tabd_copy[k - pos] * (*(*it).second->component_transitions)[i]->proba;
	  //tabd[k] = tabd_copy[k - pos] * (*(*it).second->component_transitions)[i]->weight;
	  //tabd[k] = tabd_copy[k - pos] * (*(*it).second->component_transitions)[i]->proba / (double) (*(*it).second->component_transitions)[i]->modevar->values.size();
      }
      pos += current_size;
    }
    free(tabd_copy);
    if (current_size != 1)
      current_size = current_size * (*it).second->component_transitions->size();
    else current_size = (*it).second->component_transitions->size();
  }
  
  for (unsigned int l=0; l<tabd_size; l++) {
    cout << tabd[l] << " ";
    res += tabd[l];
  }
  cout << "test of transitions probabilities : " << res << endl;
  free(tabd);
}

void DataManager::dataError (long stade, long nerr) {
  static const char *msg[]= {
    "component doesn't exist.",                       /* 1 */
  };
  switch (stade) {
  case 0:
    fprintf (stderr, "Request error.\n ...%s\n",
	     msg[nerr-1]);
  case 1:
    fprintf (stderr, "Warning.\n  ...%s\n", msg[nerr-1]);
  }
}

void DataManager::printModes (FILE* fd, State* st) {
  if (fd != NULL) {
    for (hash_map<const char*, TrajEl*, hash<const char*>, eqstr>::iterator it = st->state->begin(); it != st->state->end(); it++) {
      if ((*it).second->getVarType () == "Mode" && (*it).second->getLabel() == "true")
	fprintf (fd, "%i ", (*it).second->at->dispid);
    }
    fprintf (fd, "\n");
  }
}
