/*    
 *  Copyright 2002 Emmanuel Benazera  
 *
 *  This file is part of Koala.
 *
 *  Koala is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Koala is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Koala; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "koaconf.h"

int koa_hybrid;
int koa_log;
int koa_evlps_opti;
int koa_code_opti;
int koa_code_debug;
int twc; /* temporal window continuous */
int twd; /* temporal window discrete */
int pi; /* prime implicants */
hash_map<const char*, int, hash<const char*>, eqstr>* config_pairs;

void config_init () {
  (*config_pairs)["hybrid"] = 1;
  (*config_pairs)["discrete"] = 0;
  (*config_pairs)["continuous"] = 0;
  (*config_pairs)["log"] = 0;
  (*config_pairs)["evlps_opti"] = 0;
  (*config_pairs)["code_opti"] = 0;
  (*config_pairs)["code_debug"] = 0;
  (*config_pairs)["diag_exo"] = 0;   /* diagnostic of exogeneous commands */
  (*config_pairs)["continuous_window"] = 10;  /* temporal window on continuous trajectories */
  (*config_pairs)["discrete_window"] = 5;  /* temporal window on discrete trajectories */
  (*config_pairs)["prime_implicants"] = 0; /* prime implicants compilation */

  for (hash_map<const char*, int, hash<const char*>, eqstr>::iterator it = config_pairs->begin(); 
       it != config_pairs->end(); it++)
    conf_cache_entry ((*it).first);
}

/* search for configuration in file (.koaconf) */
int load_conf_file () {
  uid_t curr_usr = getuid ();
  struct passwd* pwuid;
  char home_usr[64];
  FILE* conf_file;
  config_pairs = new hash_map<const char*, int, hash<const char*>, eqstr>();
  config_init ();

  if ((pwuid = getpwuid (curr_usr)) != NULL) {
    strcpy (home_usr, pwuid->pw_dir);
    strcat (home_usr, "/.koaconf");
    /*free (pwuid);*/
    if ((conf_file = fopen (home_usr, "r")) != NULL)
      return load_conf_file (conf_file);
    else {
      strcpy (home_usr, caenDir);
      strcat (home_usr, "/.koaconf");
      if ((conf_file = fopen (home_usr, "r")) != NULL)
	return load_conf_file (conf_file);
    }
  }
  return 0;
}

int load_conf_file (FILE* fd) {
  char* lineptr = 0;
  char *value;
  char* option;
  const char delimiters[] = " =\n";
  lineptr = (char*) malloc (1024 * sizeof (char));
  while (fgets (lineptr, 1024, fd) != NULL) {
    if (lineptr != NULL && lineptr[0] != '#' && lineptr[0] != '\n') {
      option = strtok (lineptr, delimiters);
      value = strtok (NULL, delimiters);
      (*config_pairs)[mkString (option)] = atoi (value);
    }
  }
  for (hash_map<const char*, int, hash<const char*>, eqstr>::iterator it = config_pairs->begin(); 
       it != config_pairs->end(); it++)
    conf_cache_entry ((*it).first);
  return 1;
}

void conf_cache_entry (const char* s) {
  if (strcmp (s, "hybrid") == 0)
    koa_hybrid = conf_get (s);
  else if (strcmp (s, "discrete") == 0)
    koa_log = conf_get (s);
  else if (strcmp (s, "continuous") == 0)
    koa_log = conf_get (s);
  else if (strcmp (s, "log") == 0)
    koa_log = conf_get (s);
  else if (strcmp (s, "evlps_opti") == 0)
    koa_evlps_opti = conf_get (s);
  else if (strcmp (s, "code_opti") == 0)
    koa_code_opti = conf_get (s);
  else if (strcmp (s, "code_debug") == 0)
    koa_code_debug = conf_get (s);
  else if (strcmp (s, "continuous_window") == 0)
    twc = conf_get (s);
  else if (strcmp (s, "discrete_window") == 0)
    twd = conf_get (s);
  else if (strcmp (s, "prime_implicants") == 0)
    pi = conf_get (s);
}

void conf_set (const char* s, int v) {
  if (config_pairs->count (s) != 0) {
    (*config_pairs)[s] = v;
    conf_cache_entry (s);
  }
  else fprintf (stderr, "option %s doesn't exist\n", s);
}

int conf_get (const char* s) {
  if (config_pairs->count (s) != 0)
    return (*config_pairs)[s];
  else {
    fprintf (stderr, "option %s doesn't exist\n", s);
    return -1;
  }
}

void print_conf () {
  for (hash_map<const char*, int, hash<const char*>, eqstr>::iterator it = config_pairs->begin(); it != config_pairs->end(); it++)
    fprintf (stdout, "%s -- %i\n", (*it).first, (*it).second);
}
