/*    
 *  Copyright 2002, 2003 Emmanuel Benazera  
 *
 *  This file is part of Koala.
 *
 *  Koala is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Koala is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Koala; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "koala-runt.h"

//#define VERB_RUNT

double *meas, *inps;
Scheduler* sMM;
hybTrajectory* hybNT = 0;
bool diag_flag = 0;  /* diagnosis flag (internal) */
bool bst_flag = 0; /* blind state tracking flag (internal) */
double last_detection_time = -1;
multimap<double, hybTrajectory*, greater<double> >* tracked_trajs;
multimap<double, hybTrajectory*, greater<double> >* ntrajs;
multimap<double, hybTrajectory*, greater<double> >* misbtrajs;
multimap<double, hybTrajectory*, greater<double> >* c_misbtrajs;
vector<condition>* existing_conflicts;   /* for reducing conflict sets when bridging to the discrete level */

long koala_initialize_discrete () {
  initNominalTraj (sMM->active_atoms, NULL);
  delete tracked_trajs;
  sMM->raz ();
  sMM->tracked_traj->insert (pair<double, Trajectory*>(1, hybNT->traj));
  return 0;
}

/* TODO : handle multiple first states due to splits */
long koala_initialize (long withGC, double precision, char* diagOut, long accuTime) {
  diag_flag = 0;
  initNominalTraj (sMM->active_atoms, koala_rt); /* trajs init. */
  ce_updateDico (hybNT->rt);
  fprintf (stderr, "-- time = 0 --------------------------\n");
  if (getRTNMeasure (hybNT->rt) != 0) {
    meas = (double*) new_malloc (getRTNMeasure (hybNT->rt) * sizeof (double));
    koala_getMeasures (model_name.c_str(), getRTNMeasure (hybNT->rt), meas);
  }
  if (getRTNInput (hybNT->rt) != 0) {
    inps = (double*) new_malloc (getRTNInput (hybNT->rt) * sizeof (double));
    koala_getInputs (model_name.c_str(), getRTNInput (hybNT->rt), inps);
  }
  sMM->raz();     
  caen_initialize (hybNT->rt, withGC, precision, diagOut, accuTime);            
  bridgeIncEvents(hybNT->rt);
  sMM->doCycle (hybNT->traj, hybNT->rt, 1, 0);   /* fake discrete cycle for first state configuration initialization. */
  lockFirstTrajState (hybNT->traj);
  /*sMM->printModes (modesOut, hybNT->traj->getLastNonEmptyState());*/
  bridgeConditions (hybNT->rt);
  if (! hybNT->rt->bridgeTransitions->empty ())
    koala_initMappingFuncs (hybNT->rt);  /* static loop (duplicate state pb is here) */
  clearRunTime (hybNT->rt);    /* clear events sets */
  return 1;
}

/**
 * Koala main cycle.
 * modes: 0 -> simulation only
 *        1 -> simulation and reset
 *        2 -> detection only
 *        3 -> detection and merge
 *
 * resetTime and resetPeriod are used in simulation (modes 1 & 2).
 */
long koala_doCycle (long mode, double resetTime, double resetPeriod, double simTime, long runFI, long althresh, long simu_multi, bool &nonmisb) {
  hybTrajectory* current_htraj;
  runtime current_rt;
  Trajectory* current_traj;
  ntrajs = new multimap<double, hybTrajectory*, greater<double> >();
  
  if (mode > 1)
    misbtrajs = new multimap<double, hybTrajectory*, greater<double> >();
 
#ifdef VERB_RUNT
  cerr << "cont cycle tracked trajs: " << tracked_trajs->size () << endl;
#endif
  for (multimap<double, hybTrajectory*, greater<double> >::iterator trajs_it = tracked_trajs->begin(); 
	trajs_it != tracked_trajs->end(); trajs_it++) {  /* continuous cycle */
      
    current_htraj = (*trajs_it).second;
    current_rt = current_htraj->rt;
    current_traj = current_htraj->traj;
    
    if (mode < 2) {
      if (resetPeriod == 0) {
	if (-1 == caen_doCycle2 (current_rt, 0, 0))              /* simul only */
	  fprintf (stderr, "Invalid model !\n");
      }
      else if (resetTime == 0) {
	if (-1 == caen_doCycle2 (current_rt, 1, 0))              /* simul and reset */
	  fprintf (stderr, "Invalid model !\n");
	resetTime = resetPeriod;
      }
      else {
	if (-1 == caen_doCycle2 (current_rt, 0, 0))
	  fprintf (stderr, "Invalid model !\n");
      }
    }
    else {                    /* detect only */
      if (-1 == caen_doCycle2 (current_rt, runFI, althresh))
	fprintf (stderr, "Invalid model !\n");
    }
  } /* end for tracked_trajs */
  
  splitTrajs (tracked_trajs, mode, simu_multi);  /* split the state-space */
  
#ifdef VERB_RUNT
  cerr << "discrete cycle tracked trajs: " << tracked_trajs->size () << endl;
#endif
  koalaDiscreteCycle (mode, simTime, nonmisb, runFI, althresh);
  return 1;
}

void koalaDiscreteCycle (long mode, double simTime, bool &nonmisb, long &runFI, long althresh) {
  hybTrajectory* current_htraj;
  runtime current_rt;
  Trajectory* current_traj;
  bool misb_flag = false;
  bool go_to_next = false;
  ntrajs = new multimap<double, hybTrajectory*, greater<double> >();
  multimap<double, hybTrajectory*, greater<double> >::iterator trajs_it = tracked_trajs->begin ();
  long curr_rt_num = rt_num;
  nonmisb = false;
  /*bool non_unknown = false;*/
  
  //debug
  //std::cerr << "tracked trajs size: " << tracked_trajs->size () << std::endl;
  //debug

  while (true) {  
    go_to_next = false;
    misb_flag = false;
    current_htraj = (*trajs_it).second;
    current_rt = current_htraj->rt;
    current_traj = current_htraj->traj;

    if (current_rt->num <= curr_rt_num) {
      
      //debug
      //std::cerr << "[Debug]: do discrete cycle\n";
      //debug

      bridgeIncEvents (current_rt);
      sMM->doCycle (current_traj, current_rt, 1, simTime);    /* discrete cycle */
      bridgeConditions (current_rt);
    }
    
    if (! current_rt->bridgeTransitions->empty ()) {

      //debug
      //std::cerr << "[Debug]: bridging transitions\n";
      //debug

      misb_flag = koala_reinitialize (current_rt, simTime, mode, althresh);   /* apply mapping functions */
      clearRunTime (current_rt);    /* clear runtime events */
      splitRuntime (current_rt, current_traj, tracked_trajs, 1, mode);  /* split over the remapped state */
      multimap<double, hybTrajectory*, greater<double> >::iterator trajs_it_tmp = trajs_it;
      trajs_it_tmp++;
      if (trajs_it_tmp == tracked_trajs->end()) {  /* kill (old) split runtime */
	tracked_trajs->erase (trajs_it);
	break;
      }
      tracked_trajs->erase (trajs_it);
      trajs_it = trajs_it_tmp;
      go_to_next = true;  /* TO BE FIXED */
      current_htraj = (*trajs_it).second;
      current_rt = current_htraj->rt;
      current_traj = current_htraj->traj;
    }
    else {  /* no transition have triggered */
      
      //debug
      //std::cerr << "[Debug]: no transition triggered\n";
      //debug

      ce_checkMs (current_rt, althresh);   /* verify measures are inside the computed bounds.
					      ebenazer (03/11/2005): changed seuil from 1 to althresh. */
      ce_yieldAlarms (current_rt);
      misb_flag = isMisb (current_rt);
      clearRunTime (current_rt);    /* clear runtime events */
    }

    if (! misb_flag && ! go_to_next) {

      //debug
      //std::cerr << "misb flag\n";
      //debug

      nonmisb = 1;
      ntrajs->insert (pair<double, hybTrajectory*>(-(double)current_rt->num, current_htraj));  /* last created runtime first */
    }
    else if (! go_to_next) misbtrajs->insert (pair<double, hybTrajectory*>(current_traj->rank, current_htraj));
    
    if (! go_to_next) {
      
      //debug
      //std::cerr << "setting display\n";
      //debug

      koala_setRTDisplays (getRTNDisplay (current_rt), current_rt, simTime / getRTTimeStep (current_rt));
      trajs_it++;
      if (trajs_it == tracked_trajs->end ())
	break;
    }
  } /* end while */
  
  /*cout << "nonmisb: " << nonmisb << endl;*/
  
  if (mode > 1) {                      /* detection modes */
    delete tracked_trajs;
    tracked_trajs = 0;
    if (nonmisb) {                     /* some trajectories behave correctly */
      clearMultiTrajs (misbtrajs);
      tracked_trajs = ntrajs;
      if (mode == 3 && tracked_trajs->size () > 1) 
	ce_reduxTrajs (tracked_trajs, simTime);
      if (mode > 1)
	hybNT = (*tracked_trajs->begin()).second;
    }
    else if (runFI > 2 && ! diag_flag) {   /* Diagnosis */
      /* TODO: copy misbtrajs... */
      last_detection_time = simTime;
      c_misbtrajs = copyHybTrajectories (misbtrajs);
      if (! handleMisbTrajs (ntrajs, misbtrajs, runFI, althresh, simTime)) {   /* there are abrupt faults diag */
	if (mode == 3 && tracked_trajs->size () > 1) 
	  ce_reduxTrajs (tracked_trajs, simTime);
	nonmisb = 0;
	hybNT = (*tracked_trajs->begin ()).second;
	diag_flag = 1;
      }
      else switchToBST (tracked_trajs, c_misbtrajs, simTime, last_detection_time, runFI, althresh);   /* switch to blind state tracking */
    }
    /*else if (! bst_flag && diag_flag && ! non_unknown) { */ /* only non-abrupt faults diag remaining */
      /*tracked_trajs = new multimap<double, hybTrajectory*, greater<double> > ();
      switchToBST (tracked_trajs, c_misbtrajs, simTime, last_detection_time, runFI, althresh);
    }
    else if (bst_flag && diag_flag && ! non_unknown) { */  /* automated variables decoupling with unknown mode */
      /* TODO */
    /*}*/
    else {
      cerr << "all trajs misbehave !\n";
      /*return;*/
      tracked_trajs = misbtrajs;
      nonmisb = 0;
    }
  }
  //cerr << "tracked trajectories size: " << tracked_trajs->size () << endl;
}

void splitRuntime (runtime current_rt, Trajectory* current_traj, multimap<double, hybTrajectory*, greater<double> >* ttrajs, long simu_multi, long mode) {
  ce_splitValiditySpc (current_rt);             /* split trajectories */
  if ((mode < 2 && simu_multi == 1)
      && ! nrt->empty()) {                     /* new runtimes have been created */
    createNewTrajs (current_traj, ttrajs);
    delete current_traj;  /* must delete the hybtraj too ? */
    nrt->clear ();
  }
  else if (mode > 1 && ! nrt->empty()) {
    createNewTrajs (current_traj, ttrajs);
    delete current_traj;
    nrt->clear ();
  }
  else clearNrt ();             /* simul */
}

/* split the state-space over the set of currently tracked trajectories */
void splitTrajs (multimap<double, hybTrajectory*, greater<double> >* tt, long mode, long simu_multi) {
  runtime current_rt;
  Trajectory* current_traj;
  for (multimap<double, hybTrajectory*, greater<double> >::iterator trajs_it = tracked_trajs->begin(); 
       trajs_it != tracked_trajs->end(); trajs_it++) {  /* split the state-space */
    current_rt = (*trajs_it).second->rt;
    current_traj = (*trajs_it).second->traj;
    splitRuntime (current_rt, current_traj, ntrajs, simu_multi, mode);
  }
  delete tracked_trajs;
  tracked_trajs = ntrajs;
}

bool isMisb (runtime rt) {
  for (int i=0; i<getRTNMeasure (rt); i++) {
    if (getRTMesLowMisbNth (rt, i) || getRTMesHighMisbNth (rt, i)) {
    /*if (isVarLowAlrm (getRTMeasureNth (rt, i)) || isVarHighAlrm (getRTMeasureNth (rt, i))) {*/
#ifdef VERB_RUNT
      cerr << "runtime " << rt->num << " misbehaving\n";
#endif
      return true;
    }
  }
  return false;
}

void lockFirstTrajState (Trajectory* tr) {
  State* fs = tr->getState (0);
  for (hash_map<const char*, TrajEl*, hash<const char*>, eqstr>::iterator it = fs->state->begin ();
       it != fs->state->end (); it++) {
    if ((*it).second->label == "true") {
      if (! isNull ((*it).second->at->idle)) {
	(*it).second->prevTrans = (*it).second->at->idle;
	fs->atoms_number++;
	(*fs->comps)[(*it).second->at->getVarComp ()->name.c_str ()] = (*it).second->at->getVarComp ();
      }
      sMM->lockTrajElsPerm ((*it).second, fs);
    }
  }
}

/* koala bridges procedures. */
/* hybrid trajectories creation for simulation only */
void createNewTrajs (Trajectory* tr, multimap<double, hybTrajectory*, greater<double> >* trajs) {
#ifndef PROBAS
  hybTrajectory* htr;
#endif
  for (multimap<long, runtime, less<long> >::iterator nrt_it = nrt->begin();
       nrt_it != nrt->end(); nrt_it++) {
#ifdef PROBAS
    trajs->insert(pair<double, hybTrajectory*>(tr->proba, new hybTrajectory (tr, (*nrt_it).second)));
#else
      htr = new hybTrajectory (tr, (*nrt_it).second);
      trajs->insert(pair<double, hybTrajectory*>(-(double)htr->rt->num, htr));
#endif
  }
}

void clearMultiTrajs (multimap<double, hybTrajectory*, greater<double> >* mt) {
  for (multimap<double, hybTrajectory*, greater<double> >::iterator it = mt->begin();
       it != mt->end(); it++) {
    delete (*it).second;
    (*it).second = 0;
  }
  delete mt;
  mt = 0;
}

void clearNrt () {
  /* the first runtime is not killed */
  for (multimap<long, runtime, less<long> >::iterator nrt_it = nrt->begin();
       nrt_it != nrt->end(); nrt_it++)
    freeRuntime ((*nrt_it).second);
}

void clearRunTime (runtime rt) {
  rt->bridgeEvents->clear();
  rt->bridgeConds->clear();
  rt->bridgeTransitions->clear();
}

void initNominalTraj (hash_map<const char*, Atom*, hash<const char*>, eqstr>* atoms, runtime rt) {
  tracked_trajs = new multimap<double, hybTrajectory*, greater<double> >();		
  hybNT = new hybTrajectory (atoms, rt);
  State* fst = hybNT->traj->getState (0);
  
  //debug
  /* std::cout << "fst:\n";
     fst->showFullState (); */
  //debug

  for (hash_map<const char*, Comp*, hash<const char*>, eqstr>::iterator it = sMM->active_comps->begin ();
       it != sMM->active_comps->end (); it++)
    (*fst->comps)[(*it).second->getName ()] = (*it).second;
  /*sMM->viewHashMapAtom(atoms);*/
  tracked_trajs->insert (pair<double, hybTrajectory*>(1, hybNT));
}

/* add a event down to the continuous level */
void addBridgeEvent (condition cd, Atom* at, bool tr, runtime rt) {
  if (!sMM->inMultimap (at, rt->bridgeEvents)) {
#ifdef VERB_RUNT
    if (tr)
      cerr << "\t\t" << at->getId() << " -- " << "1\n";
    else cerr << "\t\t" << at->getId() << " -- " << "0\n";
#endif
    rt->bridgeEvents->insert(pair<bool, Atom*>(tr, at));
    setCondValue (cd, (boolean) tr);
  }
}

void removeBridgeEvent (condition cd, Atom* at, runtime rt) {
  for (multimap<bool, Atom*, eqbool>::iterator it = rt->bridgeEvents->begin ();
       it != rt->bridgeEvents->end (); it++) {
    if ((*it).second == at) {
#ifdef VERB_RUNT
      cerr << "\t\tremoving " << at->getId () << endl;
#endif
      rt->bridgeEvents->erase (it);
      setCondValue (cd, (boolean) 0);
      return;
    }
  }
}

/* add event up to the discrete level */
void bridgeIncEvents (runtime rt) {
#ifdef VERB_RUNT
  if (! rt->bridgeEvents->empty ())
    cerr << "events to bridge: " << rt->bridgeEvents->size() << " on runtime " << rt->num << endl;
#endif
  for (multimap<bool, Atom*, eqbool>::iterator it = rt->bridgeEvents->begin(); it != rt->bridgeEvents->end(); it++) {
#ifdef VERB_RUNT
    cerr << (*it).first << " -- " << (*it).second->getId() << endl;
#endif
    if ((*it).second->ref_trans_pre->empty()) { /* the atom doesn't trigger any transition */
      sMM->contEvents->insert(pair<bool, Atom*>((*it).first, (*it).second));
    }
    else if ((*it).first) /* it commands some transitions and is true */
      /*sMM->addCommandEvent ((*it).second, getRTDTimeStep (rt));*/
      sMM->comEvents->push_back ((*it).second);
  }
}

void addBridgeTrans (Transition* tr, runtime rt) {
  rt->bridgeTransitions->push_back(tr);
}

void addBridgeCond (Atom* at, bool tr, runtime rt) {
#ifdef VERB_RUNT
  cerr << "bridgeCond: " << at->getId () << " -- " << tr << endl;
#endif
  rt->bridgeConds->insert(pair<bool, Atom*>(tr, at));
}

void addBridgeConflict (runtime rt, ptConflict conf) {
  condition cd = getInfluCond (getConflictInfluence (conf));

  for (unsigned int i=0; i<existing_conflicts->size (); i++)
    if ((*existing_conflicts)[i] == cd)
      return;

  parseCondConflicts (rt, cd, 1, 0);
  existing_conflicts->push_back (cd);
}

void parseCondConflicts (runtime rt, condition cd, bool add, vector<Atom*>* curr) {
  enum condType ct = getCondType (cd);
  Atom* at;
  vector<Atom*>* vec = 0;

  switch (ct) {
  case CE_EXPR :
  case CE_AT :
    at = getCondAtom (cd);
    if (! isNull (curr)) {
      if (add) {
	vec = new vector<Atom*> (*curr);
	vec->push_back (at);
      }
      else {
	curr->push_back (at);
      }
    }
    else {  /* curr is null */
      if (add) {
	vec = new vector<Atom*> ();
	vec->push_back (at);
	rt->bridgeConflicts->push_back (vec);
      }
      else {
	vec = curr;
	vec->push_back (at);
	rt->bridgeConflicts->push_back (vec);
      }
    }
    break;
  case CE_NON :
    parseCondConflicts (rt, getCondOp1 (cd), add, curr);
    break;
  case CE_OU :
    if (isNull (curr))
      vec = new vector<Atom*> ();
    else vec = curr;
    parseCondConflicts (rt, getCondOp1 (cd), 0, vec);
    parseCondConflicts (rt, getCondOp2 (cd), 0, vec);
    rt->bridgeConflicts->push_back (vec);
    break;
  case CE_ET :
    if (isNull (curr))
      vec = new vector<Atom*> ();
    else vec = new vector<Atom*> (*curr);
    parseCondConflicts (rt, getCondOp1 (cd), 1, vec);
    parseCondConflicts (rt, getCondOp2 (cd), 1, vec);
    break;
  default :
    break;
  }
}

void setRTCondVal (runtime crt, int pos, bool b) {
  crt->condi[pos]->val = (boolean) b;
}

void bridgeConditions (runtime crt) {
  CondVar* cdv = 0;
  ModeVar* mv = 0;
  DependentVar* dtv;

  for (multimap<bool, Atom*, eqbool>::iterator it = crt->bridgeConds->begin(); it != crt->bridgeConds->end(); it++) {
    
    if ((*it).second->getVarType() == "Cond") {
      cdv = (CondVar*) (*it).second->getStateVar();
      for (multimap<const char*, int>::iterator it2 = cdv->conditions->begin();
	   it2 != cdv->conditions->end(); it2++)
	if (strcmp ((*it2).first, (*it).second->getId()) == 0)
	  setRTCondVal (crt, (*it2).second, (*it).first);
    }
      
    else if ((*it).second->getVarType() == "Mode") {
      mv = (ModeVar*) (*it).second->getStateVar();
      for (multimap<const char*, int>::iterator it2 = mv->conditions->begin();
	   it2 != mv->conditions->end(); it2++)
	if (strcmp ((*it2).first, (*it).second->getId()) == 0)
	  setRTCondVal (crt, (*it2).second, (*it).first);
    }
    else if ((*it).second->getVarType() == "Dependent") {
      dtv = (DependentVar*) (*it).second->getStateVar();
      for (multimap<const char*, int>::iterator it2 = mv->conditions->begin();
	   it2 != mv->conditions->end(); it2++)
	if (strcmp ((*it2).first, (*it).second->getId()) == 0)
	  setRTCondVal (crt, (*it2).second, (*it).first);
    }
  }
}

void koala_initMappingFuncs (runtime rt) {
  variable v;
  objet ob;
  mapfVar mv;
  MapFunc* mft;
  long j;

  for (unsigned int i=0; i<rt->bridgeTransitions->size(); i++) {
    if (! isNull ((*rt->bridgeTransitions)[i]->mf)) {
      for (unsigned int j=0; j<(*rt->bridgeTransitions)[i]->mf->size (); j++) {
	mft = (*(*rt->bridgeTransitions)[i]->mf)[j];
	if (isNull (mft->cond) || ce_evalCond (mft->cond, rt)){
	  for (unsigned int k=0; k<mft->mvs->size (); k++) {
	    mv = (*mft->mvs)[k];
	    v = ce_getVarDico (rt, getMFVarName (mv));
	    ob = ce_evalExpr (rt, getRTMFNth (rt, getMFExpr (mv)));
	    /*ob = getMFObj (mv);*/
	    ce_putObjet (ob, varOldValue (v));
	    ce_fillHisto (v);   /* fill history of the variable with its 'old' value */
	    setDetermined (v);
	  }
	}
      }
    }
  }

  /* update other variables */
  for (j=0; j<getRTNTempo(rt); ++j) {
    v = getRTTempoNth (rt, j);
    if (! isDetermined (v)) {
      ce_compInitVar (rt, v);     /* result into old value */  
    } /* endif ! isDetermined v */
  } /* end for */
  for (j=0; j<getRTNUntemp(rt); ++j) {
    v = getRTUntempNth (rt, j);
    if (! isDetermined (v))
      ce_compInitVar (rt, v);        /* result into old value */
  } /* end for */
  if (ce_initValSpc (rt) != 1)
    fprintf (stdout, "Error while splitting init state\n");
  /*setInited (rt);*/

}

/* apply values from mapping function */
void applyMFBounds (runtime rt, variable v, objet omv) {
  objet vob = varOldValue (v);
  objet von = varNewValue (v);

  if (CE_REAL == getObjType (vob)) {
    setObjType (von, CE_REAL);
    setObjType (vob, CE_REAL);
    if (ce_isMeasure (rt, v))     {/* fake measures when needed by the mapping function */
      setObjType (varMesValue (v), CE_REAL);
      ce_putObjet (omv, varMesValue (v));
    }
    else {
      ce_putObjet (omv, vob);
      ce_putObjet (omv, von);
    }
  }
  else if (CE_RANGE == getObjType (vob)) { /* CE_REAL vob */
    if (CE_REAL == getObjType (omv)) {
      /*double od = getObjRealVal (omv);
	finf = min (vinf, od); fsup = max (vsup, od);
	range rr = {finf, fsup};
	if (ce_isMeasure (rt, v)) {
	setObjType (varMesValue (v), CE_RANGE);
	setObjRangeVal (varMesValue (v), rr);
	}
	else {
	setObjRangeVal (vob, rr);
	}*/
      if (ce_isMeasure (rt, v)) {
	setObjType (varMesValue (v), CE_REAL);
	ce_putObjet (omv, varMesValue (v));
      }
      else {
	setObjType (vob, CE_REAL);
	ce_putObjet (omv, vob);
      }
    }
    else {  /* CE_REAL omv */
      double finf, fsup, vinf = getRangeInf (getObjRangeVal (vob)),
	vsup = getRangeSup (getObjRangeVal (vob));
      double odinf = getRangeInf (getObjRangeVal (omv)),
	odsup = getRangeSup (getObjRangeVal (omv));
      finf = min (vinf, odinf); fsup = max (vsup, odsup);
      range rr = {finf, fsup};
      if (ce_isMeasure (rt, v)) {  /* fake measures when needed by the mapping function */
	setObjType (varMesValue (v), CE_RANGE);
	setObjRangeVal (varMesValue (v), rr);  /* beware... */
      }
      else {  /* isMeasure */
	setObjRangeVal (vob, rr);
      }
    }
  }
  setDetermined (v);
}

bool koala_applyMappingFuncs (runtime rt, double stime) {
  variable v;
  objet ob;
  mapfVar mv;
  MapFunc* mft;

  bool hasMF = false;

  /* current time-step static cycle */
  for (unsigned int i=0; i<rt->bridgeTransitions->size(); i++) {
    if (! isNull ((*rt->bridgeTransitions)[i]->mf)) {
      for (unsigned int j=0; j<(*rt->bridgeTransitions)[i]->mf->size (); j++) {
	mft = (*(*rt->bridgeTransitions)[i]->mf)[j];
	if (isNull (mft->cond) || ce_evalCond (mft->cond, rt)) {
	  for (unsigned int k=0; k<mft->mvs->size(); k++) {
	    mv = (*mft->mvs)[k];
	    v = ce_getVarDico (rt, getMFVarName (mv));
	    ob = ce_evalExpr (rt, getRTMFNth (rt, getMFExpr (mv)));
	    applyMFBounds (rt, v, ob);
	    hasMF = true;
	    /*ce_evalVarConds (v, rt);*/  /* BEWARE : can trigger more transitions */
	  }
	}
      }
    }
  }

  for (int j=0; j<getRTNTempo(rt); ++j) {
    v = getRTTempoNth (rt, j);
    if (! isDetermined (v)) {
      ce_compInitVar (rt, v);  
    } 
  } 
  for (int j=0; j<getRTNUntemp(rt); ++j) {
    v = getRTUntempNth (rt, j);
    if (! isDetermined (v))
      ce_compInitVar (rt, v);      
  }

    /* return to real measures, for next step evaluation */
  for (long k=0; k<getRTNMeasure (rt); k++) {
    /*ce_putObjet (varNewValue (getRTDicoNth (rt, k)), varMesValue (getRTDicoNth (rt, k)));*/
    affectRealMes (getRTMeasureNth (rt, k), getRTMesValNth (rt, k));
    /*cerr << "varMesValue: ";
      ce_printObjet (stdout, varMesValue (getRTMeasureNth (rt, k)));
      cerr << endl;*/
  }
  
  return hasMF;
}

bool koala_applyMappingFuncs2 (runtime rt, double stime) {
  variable v;
  objet ob;
  mapfVar mv;
  MapFunc* mft;
  runtime crt = ce_copyPartialFaultyRt (rt, 0, stime-1, stime, Trajectory::twd);   /* partial runtime copy */

  bool hasMF = false;

  /* previous time-step dynamic cycle */
  for (unsigned int i=0; i<crt->bridgeTransitions->size(); i++) {
    if (! isNull ((*crt->bridgeTransitions)[i]->mf)) {
      for (unsigned int j=0; j<(*crt->bridgeTransitions)[i]->mf->size (); j++) {
	mft = (*(*crt->bridgeTransitions)[i]->mf)[j];
	for (unsigned int k=0; k<mft->mvs->size (); k++) {
	  mv = (*mft->mvs)[k];
	  v = ce_getVarDico (crt, getMFVarName (mv));
	  ob = ce_evalExpr (crt, getRTMFNth (crt, getMFExpr (mv)));
	  applyMFBounds (crt, v, ob);
	  hasMF = true;
	}
      }
    }
  }
  ce_doStepTempo (crt);   /* dynamic cycle */
  ce_doStepUntemp (crt);

  /* current time-step static cycle */
  for (unsigned int i=0; i<rt->bridgeTransitions->size(); i++) {
    if (! isNull ((*rt->bridgeTransitions)[i]->mf)) {
      for (unsigned int j=0; j<(*rt->bridgeTransitions)[i]->mf->size (); j++) {
	mft = (*(*rt->bridgeTransitions)[i]->mf)[j];
	for (unsigned int k=0; k<mft->mvs->size(); k++) {
	  mv = (*mft->mvs)[k];
	  v = ce_getVarDico (rt, getMFVarName (mv));
	  ob = ce_evalExpr (rt, getRTMFNth (rt, getMFExpr (mv)));
	  applyMFBounds (rt, v, ob);
	  hasMF = true;
	}
      }
    }
  }
  
  /* update other variables with a static cycle */
  for (int j=0; j<getRTNTempo(rt); ++j) {
    v = getRTTempoNth (rt, j);
    if (! isDetermined (v)) {
      ce_compInitVar (rt, v);  
    } 
  } 
  for (int j=0; j<getRTNUntemp(rt); ++j) {
    v = getRTUntempNth (rt, j);
    if (! isDetermined (v))
      ce_compInitVar (rt, v);      
  }
  
  /* back to real measures, for next step evaluation */
  for (unsigned int i=0; i<rt->bridgeTransitions->size (); i++) {
    if (! isNull ((*rt->bridgeTransitions)[i]->mf)) {
      for (unsigned int j=0; j<(*rt->bridgeTransitions)[i]->mf->size (); j++) {
	mft = (*(*rt->bridgeTransitions)[i]->mf)[j];
	for (unsigned int k=0; k<mft->mvs->size(); k++) {
	  mv = (*mft->mvs)[k];
	  v = ce_getVarDico (rt, getMFVarName (mv));
	  if (ce_isMeasure (rt, v)) {
	    for (int l=0; l<getRTNMeasure (rt); l++) {
	      if (getRTMeasureNth (rt, l) == v) {
		affectRealMes (getRTMeasureNth (rt, l), getRTMesValNth (rt, l));
		/* cerr << "varMesValue: ";
		   ce_printObjet (stdout, varMesValue (getRTMeasureNth (rt, l)));
		   cerr << endl;*/
	      }
	    }
	  }
	}
      }
    }
  }

  /* merging results */
  for (int i=0; i<getRTNVar (rt); i++) {
    /*ce_putObjet (varOldValue (getRTDicoNth (crt, i)), varOldValue (getRTDicoNth (rt, i)));*/
    /*if (! ce_isMeasure (rt, getRTDicoNth (rt, i)) 
	&& ! ce_isInput (rt, getRTDicoNth (rt, i))
	&& ! isParameter (getRTDicoNth (rt, i))) {*/
    
    /*if (ce_isMeasure (rt, getRTDicoNth (rt, i))) {*/  /* hack */
    if (! isParameter (getRTDicoNth (rt, i))) {
      /*cerr << "crt: ";
      ce_printObjet (stderr, varNewValue (getRTDicoNth (crt, i)));
      cerr << "\nrt: ";
      ce_printObjet (stderr, varNewValue (getRTDicoNth (rt, i)));
      cerr << endl;*/
      ce_uniObjet (varNewValue (getRTDicoNth (rt, i)), varNewValue (getRTDicoNth (crt, i)));
    }
    /* hack */
    if (! isParameter (getRTDicoNth (rt, i))) {
    /*if (ce_isMeasure (rt, getRTDicoNth (rt, i)))*/
      ce_putObjet (varNewValue (getRTDicoNth (rt, i)), varOldValue (getRTDicoNth (rt, i)));
      ce_fillHisto (getRTDicoNth (rt, i));
    }
  }
  freeRuntime (crt);
  
  return hasMF;
}

bool koala_reinitialize (runtime &rt, const double &stime, const long &mode, 
			 const long &seuil) {
#ifdef VERB_RUNT
  cerr << "reinitializing state\n";
#endif
  for (long i=0; i<getRTNVar (rt); i++)
    if (! isParameter (getRTDicoNth (rt, i)))
      setUndetermined (getRTDicoNth (rt, i));
  bool hasMF = false;
  if (mode < 2)
    hasMF = koala_applyMappingFuncs (rt, stime);
  else hasMF = koala_applyMappingFuncs2 (rt, stime);
  //std::cerr << "hasMF: " << hasMF << " -- seuil: " << seuil << std::endl;
  if (hasMF)
    ce_checkMs (rt, 1);  /* ebenazer: seuil is 1... -> instant check. */
  else ce_checkMs (rt, seuil);
  //std::cerr << "ms check done\n";
  ce_yieldAlarms (rt);
#ifdef VERB_RUNT
  for (long i=0; i<getRTNVar (rt); i++) {
    if (! isParameter (getRTDicoNth (rt, i))) {
      cerr << getVarName (getRTDicoNth (rt, i)) << ": ";
      ce_printObjet (stderr, varOldValue (getRTDicoNth (rt, i)));
      cerr << endl;
    /*ce_printVarHistory (stdout, getRTDicoNth (rt, i), rt);
    cerr << endl;*/
    }
  }
#endif
  ce_yieldDisplays (rt, 0);
  return isMisb (rt);
}

/**
 * Reducing tracked trajectories by merging envelopes.
 * BEWARE:
 *        - only merge the current state and display values.
 *        
 */

bool intersectObsRt (runtime rt1, runtime rt2) {
  long n=getRTNVar (rt1);
  objet o1, o2, otmp;
  bool res = false;
  
  for (long i=0; i<n; i++) {
    if (ce_isMeasure (rt1, getRTDicoNth (rt1, i))) {
      o1 = varOldValue (getRTDicoNth (rt1, i));
      o2 = varOldValue (getRTDicoNth (rt2, i));
      otmp = copyObjet (o1);
      if (ce_interObjet (otmp, o2) 
	  || ce_eqObjet (o1, o2)) {       /* objects share prediction */
	ce_uniObjet (o1, o2);             /* values are merged */
	setDetermined (getRTDicoNth (rt1, i));
	res = true;
      }
      freeObjet (otmp);
    }
  }
  /*cerr << "intersectObsRt result: " << res << endl;*/
  return res;
}

void ce_reduxTrajs (multimap<double, hybTrajectory*, greater<double> >* ttrajs, double simTime) {
  runtime rtcurr, rtnext;
  State* stcurr;
  State* stnext;
  multimap<double, hybTrajectory*, greater<double> >::iterator tr_it_tmp;
  multimap<double, hybTrajectory*, greater<double> >::iterator tr_it = ttrajs->begin ();
  multimap<double, hybTrajectory*, greater<double> >::iterator tr_it_next = ttrajs->begin ();
  tr_it_next++;

  /*cerr << "reduxTrajs...\n";*/

  while (true) {
    rtcurr = (*tr_it).second->rt;
    rtnext = (*tr_it_next).second->rt;
    stcurr = (*tr_it).second->traj->getLastState ();
    stnext = (*tr_it_next).second->traj->getLastState ();
    tr_it_tmp = tr_it_next;
    
    if (sMM->compareTrajsStates (stcurr, stnext)
	&& intersectObsRt (rtcurr, rtnext)) {                      /* observations intersect */
      mergeRuntimes (rtcurr, rtnext, simTime);
      delete (*tr_it_next).second;
      (*tr_it_next).second = 0;
#ifdef VERB_RUNT
      cerr << "merged runtime\n";
      for (long i=0; i<getRTNVar (rtcurr); i++) {
	cerr << getVarName (getRTDicoNth (rtcurr, i)) << ": ";
	ce_printObjet (stderr, varOldValue (getRTDicoNth (rtcurr, i)));
	cerr << endl;
      }
#endif
      tr_it_tmp = tr_it_next;
      tr_it_tmp++;
      ttrajs->erase (tr_it_next);
    }
    if (tr_it_next != tr_it_tmp)
      tr_it_next = tr_it_tmp;
    else tr_it_next++;
    if (tr_it_next == ttrajs->end ()) {
      tr_it++;
      if (tr_it == ttrajs->end ())
	return;
      tr_it_next = tr_it;
      tr_it_next++;
      if (tr_it_next == ttrajs->end ())
	return;
    }
  }
}

/**
 * \brief merge past values of variables of both runtime.
 */
void mergeRuntimes (runtime rt1, runtime rt2, double simTime) {
  long n = getRTNVar (rt1);
  
  for (long i=0; i<n; i++) {
    ce_uniObjet (varOldValue (getRTDicoNth (rt1, i)), varOldValue (getRTDicoNth (rt2, i)));
  }
  ce_yieldDisplays (rt1, 0);
  
  for (long i=0; i<2*simTime*getRTNDisplay (rt1); i=i+2) { /* merge displays */
    setRTDispAllNth (rt1, i, min (getRTDispAllNth (rt1, i), getRTDispAllNth (rt2, i)));
    setRTDispAllNth (rt1, i+1, max (getRTDispAllNth (rt1, i+1), getRTDispAllNth (rt2, i+1)));
  }
}


/**
 * Hybrid diagnosis in the main loop.
 */

void insertDiagnosis (hybTrajectory* htr, multimap<double, hybTrajectory*, greater<double> >* tt) {
  multimap<bool, Atom*, eqbool>* obs;
  vector<vector<Atom*>*>* bc = htr->rt->bridgeConflicts;
  if (isNull (tt))
    tt = new multimap<double, hybTrajectory*, greater<double> >();
  for (unsigned int i=0; i<bc->size (); i++) {
    obs = new multimap<bool, Atom*, eqbool> ();
    for (unsigned int j=0;j<(*bc)[i]->size (); j++) {
      obs->insert (pair<bool, Atom*> (0, (*(*bc)[i])[j]));
    }
    hybTrajectory* chtr = new hybTrajectory ();
    chtr->traj = new Trajectory (htr->traj, htr->traj->getCurrentTime () + 1);
    chtr->rt = ce_copyRuntime (htr->rt);
    sMM->Diagnose (chtr->traj, chtr->rt, obs, tt);
    delete obs;
  }
  delete htr;
  htr = 0;
}

/**
 * perform diagnosis at both levels on a set of mibehaving tracked trajectories.
 * result are in tracked_trajs.
 * return true if all diagnosis have unknown modes.
 */
bool handleMisbTrajs (multimap<double, hybTrajectory*, greater<double> >* ntrajs, multimap<double, hybTrajectory*, greater<double> >* misbtrajs, long runFI, long althresh, double simTime) {
#ifdef VERB_RUNT
  cerr << "entering handleMisbTrajs\n";
#endif

  /* clear some stuff */
  delete tracked_trajs;
  tracked_trajs = 0;
  for (multimap<double, hybTrajectory*, greater<double> >::iterator it = ntrajs->begin(); 
       it != ntrajs->end(); it++)
    delete (*it).second;
  delete ntrajs;
  
  /* diagnose */
  tracked_trajs = new multimap<double, hybTrajectory*, greater<double> > ();
  for (multimap<double, hybTrajectory*, greater<double> >::iterator it = misbtrajs->begin();
       it != misbtrajs->end(); it++) {
    clock_t start = clock ();
    ce_detectAndIsolate ((*it).second->rt, runFI, althresh);
    clock_t end = clock ();
    sMM->viewAtomVector2 ((*it).second->rt->bridgeConflicts);
    cerr << "Diagnosis at continuous level in: " << ((double) (end - start) / CLOCKS_PER_SEC) 
	 << " seconds.\n";
    insertDiagnosis ((*it).second, tracked_trajs);
    //misbtrajs->erase (it);
  }
  delete misbtrajs;
  misbtrajs = 0;
  cerr << "Diagnosis completed\n";

  /* ebenazer (03/11/2005): added this check, but koala should switch back to
     nominal tracking in this case. */
  if (tracked_trajs->empty ())
    {
      std::cerr << "[Error]: Can't find any diagnose for this fault.\n";
      exit (-1);
    }

  return verifyAndUpdateTrajs (simTime, runFI, althresh); /* verify and update results */
}

void koala_diagGetInputs (runtime ref, double stime, double currtime, double* in) {
  int pos = (int) currtime - (int) stime - 1;
  if (pos >= 0) {
    for (int i=0; i<getRTNInput (ref); i++) {
      in[i] = getObjRealVal (varHistoNth (getRTInputNth (ref, i), pos));
#ifdef VERB_RUNT
      cerr << "check inputs: " << in[i] << endl;
#endif
    }
  }
  else { /* stime is current time */
    for (int i=0; i<getRTNInput (ref); i++) {
      in[i] = getObjRealVal (varOldValue (getRTInputNth (ref, i)));
#ifdef VERB_RUNT
      cerr << "check (current) inputs: " << in[i] << endl;
#endif
    }
  }
}

void koala_diagGetMeasures (runtime ref, double stime, double currtime, double *ms) {
  int pos = (int) currtime - (int) stime - 1;
  if (pos >= 0) {
    for (int i=0; i<getRTNMeasure (ref); i++) {
      ms[i] = getObjRealVal (varHistoMesNth (getRTMeasureNth (ref, i), pos));
#ifdef VERB_RUNT
      cerr << "check measures: " << ms[i] << endl;
#endif
    }
  }
  else { /* stime is current time */
    for (int i=0; i<getRTNMeasure (ref); i++) {
      ms[i] = getObjRealVal (varMesValue (getRTMeasureNth (ref, i)));
#ifdef VERB_RUNT
      cerr << "check (current) measures: " << ms[i] << endl;
#endif
    }
  }
}

void addDiagBridgeTrans (Trajectory* tr, runtime rt, int timestep) {
  for (multimap<int, Transition*>::iterator it = tr->trans_diag->begin ();
       it != tr->trans_diag->end (); it++) {
      if ((*it).first != timestep)
	break;
      else {
	addBridgeTrans ((*it).second, rt);
      }
  }
}

void updateDiagConds (Trajectory* tr, runtime &rt, const int &timestep, 
		      const double &fault_time, const double &simTime, const long &seuil) {
  int nc = getRTNCondi (rt);
  State* fst = tr->getState (timestep);
  Atom* cdat;
  TrajEl* tel;
  condition cd;
  for (int i=0; i<nc; i++) {
    cd = getRTCondiNth (rt, i);
    if (! isNull (cdat = getCondAtom (cd))) {
      if (! isNull (tel = fst->getElt (cdat->getId ()))) {
	if (tel->label == "true")
	  addBridgeCond (cdat, 1, rt);
	else addBridgeCond (cdat, 0, rt);
      }
    }
  }
  bridgeConditions (rt);
  addDiagBridgeTrans (tr, rt, timestep);
  if (! rt->bridgeTransitions->empty ()) {
    /*koala_diagGetInputs (rt, fault_time, simTime, inps);
    koala_diagGetMeasures (rt, fault_time, simTime, meas);*/
    koala_reinitialize (rt, fault_time, 3, seuil);
  }
  koala_setRTDisplays (getRTNDisplay (rt), rt, fault_time / getRTTimeStep (rt));
  /* reset alarms */
  for (int i=0; i<getRTNVar (rt); i++) {
    razVarLowAlrm (getRTDicoNth (rt, i));
    razVarHighAlrm (getRTDicoNth (rt, i));
    setVarLowOk (getRTDicoNth (rt, i));
    setVarHighOk (getRTDicoNth (rt, i));
    setVarOk (getRTDicoNth (rt, i));
  }
}

/* return true if all trajs have 'unknown' modes */
bool verifyAndUpdateTrajs (double simTime, long runFI, long althresh) {
  int discrete_fault_time;
  double rec_simTime;
  bool all_unknown = true;
  multimap<double, hybTrajectory*, greater<double> >* res = new multimap<double, hybTrajectory*, greater<double> > ();
  multimap<double, hybTrajectory*, greater<double> >* tt;
  
  /* first discrete cycle at discrete_fault_time */
#ifdef VERB_RUNT
  cerr << "verifying trajs: " << tracked_trajs->size () << endl;
#endif
  for (multimap<double, hybTrajectory*, greater<double> >::iterator it = tracked_trajs->begin ();
       it != tracked_trajs->end (); it++) {

    if (! (*it).second->traj->isUnknown ())
      all_unknown = false;

    if ((*it).second->traj->trans_diag == NULL) {  /* the trajectory is somewhat nominal */
      res->insert (pair<double, hybTrajectory*> ((*it).second->traj->rank, (*it).second));
      continue;
    }

   tt = new multimap<double, hybTrajectory*, greater<double> > ();  /* internal tracked trajectories */

   /*(*it).second->traj->showTrajectoryTransitions ();*/

   discrete_fault_time = (*(*it).second->traj->trans_diag->upper_bound (0)).first;
    hybTrajectory* htraj = new hybTrajectory ();
    htraj->traj = new Trajectory ((*it).second->traj, (*it).second->traj->getCurrentTime () + 1);
    double cont_fault_time = htraj->traj->getTimeCorr (discrete_fault_time);
    htraj->rt = ce_copyPartialFaultyRt ((*it).second->rt, discrete_fault_time, cont_fault_time, simTime, Trajectory::twd);
    
    updateDiagConds (htraj->traj, htraj->rt, discrete_fault_time, cont_fault_time, simTime, althresh);

    tt->insert (pair<double, hybTrajectory*> (htraj->traj->rank, htraj)); /* TODO: probas */
    /*koalaDiscreteCycle (3, cont_fault_time, tt);*/ /* TODO: focus on correct discrete time-step ! */

    for (rec_simTime=cont_fault_time+1; rec_simTime <= simTime; rec_simTime++) {
#ifdef VERB_RUNT
      cerr << "[koala-checkCycle] Time: " << rec_simTime << " (" << simTime << ")\n";
#endif
      koala_diagGetInputs ((*it).second->rt, rec_simTime, simTime, inps);
      koala_diagGetMeasures ((*it).second->rt, rec_simTime, simTime, meas);
      if (-1 == koala_checkCycle (tt, 0, 0, rec_simTime, runFI, althresh, 0))
	fprintf (stderr, "Koala recursive cycle failed!\n");
    }
    for (multimap<double, hybTrajectory*, greater<double> >::iterator ttit = tt->begin();
	 ttit != tt->end (); ttit++) {
      res->insert (pair<double, hybTrajectory*> ((*ttit).second->traj->rank, (*ttit).second));
    }
    delete tt;
  }
  delete tracked_trajs;
  tracked_trajs = res;
  return all_unknown;
}

/**
 * internal fast final diagnosis check loop 
 * return 0 if some trajectories are valid, 1 otherwise
 */
long koala_checkCycle (multimap<double, hybTrajectory*, greater<double> >* tt, double resetTime, double resetPeriod, double simTime, long runFI, long althresh, long simu_multi) {
  hybTrajectory* current_htraj;
  runtime current_rt;
  Trajectory* current_traj;
  multimap<double, hybTrajectory*, greater<double> >* temp_trajs = new multimap<double, hybTrajectory*, greater<double> >();
  /*bool nonmisb = 1;*/

  /* continuous cycle */
#ifdef VERB_RUNT
  cerr << "[koala-checkCycle] cont cycle tracked trajs: " << tt->size () << endl;
#endif
  for (multimap<double, hybTrajectory*, greater<double> >::iterator trajs_it = tt->begin(); 
	trajs_it != tt->end(); trajs_it++) {
      
    current_htraj = (*trajs_it).second;
    current_rt = current_htraj->rt;
    current_traj = current_htraj->traj;
    
    if (-1 == caen_doCycle2 (current_rt, 3, althresh))  /* re-compute bounds */
      fprintf (stderr, "Invalid model !\n");
    
    for (int i=0; i<getRTNVar (current_rt); i++) { /* reset alarms to follow measures */
      razVarLowAlrm (getRTDicoNth (current_rt, i));
      razVarHighAlrm (getRTDicoNth (current_rt, i));
      setVarLowOk (getRTDicoNth (current_rt, i));
      setVarHighOk (getRTDicoNth (current_rt, i));
      setVarOk (getRTDicoNth (current_rt, i));
    }
    
    temp_trajs->insert (pair<double, hybTrajectory*>(-(double)current_rt->num, current_htraj));
  } /* end for tt */

  delete tt;
  tt = temp_trajs;
  temp_trajs = new multimap<double, hybTrajectory*, greater<double> >();

  if (tt->size () > 1)    /* merge overlaping tracked trajectories */
    ce_reduxTrajs (tt, simTime);
  
  /* split the state space */
  for (multimap<double, hybTrajectory*, greater<double> >::iterator trajs_it = tt->begin(); 
       trajs_it != tt->end(); trajs_it++) {
    current_rt = (*trajs_it).second->rt;
    current_traj = (*trajs_it).second->traj;
    ce_splitValiditySpc (current_rt);             /* split trajectories */
    if (! nrt->empty()) {
      createNewTrajs (current_traj, temp_trajs);
      /*delete current_traj;*/ /* buggy */
      nrt->clear ();
    }
  }
  delete tt;
  tt = temp_trajs;
  koalaDiagDiscreteCycle (3, simTime, tt);
  return 1;
}

void koalaDiagDiscreteCycle (long mode, double simTime, multimap<double, hybTrajectory*, greater<double> >* ttrajs) {
  hybTrajectory* current_htraj;
  runtime current_rt;
  Trajectory* current_traj;
  
  if (mode > 1)
    hybNT = (*ttrajs->begin()).second;
  for (multimap<double, hybTrajectory*, greater<double> >::iterator trajs_it = ttrajs->begin();
       trajs_it != ttrajs->end(); trajs_it++) {
    current_htraj = (*trajs_it).second;
    current_rt = current_htraj->rt;
    current_traj = current_htraj->traj;
    /*koala_setRTDisplays (getRTNDisplay (current_rt), current_rt, simTime / getRTTimeStep (current_rt));*/
    /*bridgeIncEvents (current_rt);
      sMM->doCycle (current_traj, current_rt, 1, simTime);
      bridgeConditions (current_rt);
      if (! current_rt->bridgeTransitions->empty ())
      koala_reinitialize (current_rt); */  /* apply mapping functions */
    koala_setRTDisplays (getRTNDisplay (current_rt), current_rt, simTime / getRTTimeStep (current_rt));
    clearRunTime (current_rt);
  } /* end for ttrajs */
}

/* diagnosis modes switching strategy */
multimap<double, hybTrajectory*, greater<double> >* copyHybTrajectories (multimap<double, hybTrajectory*, greater<double> >* mtrajs) {
  hybTrajectory* hybt;
  multimap<double, hybTrajectory*, greater<double> >* res = new multimap<double, hybTrajectory*, greater<double> > ();
  for (multimap<double, hybTrajectory*, greater<double> >::iterator it = mtrajs->begin ();
       it != mtrajs->end (); it++) {
    hybt = new hybTrajectory ((*it).second, (*it).second->traj->getCurrentTime () + 1);
    res->insert (pair<double, hybTrajectory*> ((*it).first, hybt));
  }
  return res;
}

void switchToBST (multimap<double, hybTrajectory*, greater<double> >* ttrajs, multimap<double, hybTrajectory*, greater<double> >* mtrajs, double stime, double ldetect, long &runFI, long althresh) {
  multimap<double, hybTrajectory*, greater<double> >* tt;
  int discrete_fault_time;
  double rec_simTime;
  cerr << "switching to blind state tracking\n";
  runFI = 2;  /* switch to 'nofaults' mode */
  
  if (stime != ldetect) {
    for (multimap<double, hybTrajectory*, greater<double> >::iterator it = mtrajs->begin();
	 it != mtrajs->end (); it++) {
      tt = new multimap<double, hybTrajectory*, greater<double> > ();  /* internal tracked trajectories */
      discrete_fault_time = (*(*it).second->traj->trans_diag->upper_bound (0)).first;
      hybTrajectory* htraj = new hybTrajectory ();
      htraj->traj = new Trajectory ((*it).second->traj, (*it).second->traj->getCurrentTime () + 1);
      double cont_fault_time = htraj->traj->getTimeCorr (discrete_fault_time);
      htraj->rt = ce_copyPartialFaultyRt ((*it).second->rt, discrete_fault_time, cont_fault_time, stime, Trajectory::twd);
      
      updateDiagConds (htraj->traj, htraj->rt, discrete_fault_time, cont_fault_time, stime, althresh);
      
      tt->insert (pair<double, hybTrajectory*> (htraj->traj->rank, htraj)); /* TODO: probas */
      
      for (rec_simTime=cont_fault_time+1; rec_simTime <= stime; rec_simTime++) {
#ifdef VERB_RUNT
	cerr << "[switchToBST] Time: " << rec_simTime << " (" << stime << ")\n";
#endif
	koala_diagGetInputs ((*it).second->rt, rec_simTime, stime, inps);
	koala_diagGetMeasures ((*it).second->rt, rec_simTime, stime, meas);
	if (-1 == koala_checkCycle (tt, 0, 0, rec_simTime, runFI, althresh, 0))
	  fprintf (stderr, "Koala recursive cycle failed!\n");
      }
      for (multimap<double, hybTrajectory*, greater<double> >::iterator ttit = tt->begin();
	   ttit != tt->end (); ttit++) {
	ttrajs->insert (pair<double, hybTrajectory*> ((*ttit).second->traj->rank, (*ttit).second));
      }
      delete tt;
    }
  }
  
  delete mtrajs;   /* delete c_misbtrajs */
}
