/*    
 *  Copyright 2000, 2001, 2002 Emmanuel Benazera  
 *
 *  This file is part of Koala.
 *
 *  Koala is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  Koala is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Koala; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <iostream>
#include <fstream>
#include "model_tools.h"

/**
 * Make a graph of a component, i.e. a file in dot form.
 */
void toDot(Comp* co) {
  char* s = 0;
  char buf[10];
  std::cout << "[model_tools] --> graphing component " << co->getName() << std::endl;
  std::string gr;
  multimap<bool, Atom*, eqbool>* prea;
  multimap<bool, Atom*, eqbool>* posta;
  std::string com_name;
  
  /* add 'headers'. */
  std::string header = "digraph " + co->name + " {\n\trankdir=TB;\n\tsize=\"8,5\"\n\torientation=portrait;\n\tnode [shape = circle];\n\t";
  gr.append(header);
  /* add nodes. */
  for (unsigned int i=0; i<co->component_transitions->size(); i++) {
    com_name = "";
    prea = (*co->component_transitions)[i]->pre->getAllSignedAtoms();
    posta = (*co->component_transitions)[i]->post->getAllSignedAtoms();
    for (multimap<bool, Atom*, eqbool>::iterator it1 = prea->begin(); it1 != prea->end(); it1++) {
      if ((*it1).second->getStateVar() == (ModeVar*) (*co->component_transitions)[i]->modevar) {
	header.append((*it1).second->value + " -> ");
      }
      else if ((*it1).first & (*it1).second->getVarType() == "Command")
	com_name = (*it1).second->value;
    }
    for (multimap<bool, Atom*, eqbool>::iterator it2 = posta->begin(); it2 != posta->end(); it2++) {
      if ((*it2).second->getStateVar() == (ModeVar*) (*co->component_transitions)[i]->modevar) {
	header.append((*it2).second->value + " [ ");
      }
    }
    if ((*co->component_transitions)[i]->getType() == "Fail")
      header.append("style=dotted,");
    /* add label. */
    sprintf(s, "%3.2f", (*co->component_transitions)[i]->proba);
    header.append(" label = \"" + com_name + " (" + std::string(s) + ")\" ];\n\t");
    s = 0; s = buf;
    delete prea;
    delete posta;
  }
  header.append("}");
  /* create file : name + .dot */
  std::ofstream fgr(std::string(co->name + ".dot").c_str(), ios::trunc);
  fgr << header;
}
/**
 * Make a graph node name with specific label.
 */
std::string makeNodeNameWithLabel(Atom* at, const std::string &timestep) {
  return at->getVarComp()->name + "_" + at->value + timestep + " [label=\"" + at->value + "\"];";
}

/**
 * Make a graph node name without label (for search operations).
 */
std::string makeNodeName(Atom* at, const std::string &timestep) {
  return at->getVarComp()->name + "_" + at->value + timestep;
}

/**
 * Make a graph of a trajectory, i.e. a file in dot form.
 */
void toDot(Trajectory* tr, const std::string &name) {
  std::string graph;
  char* s;
  char buf[100];
  s = buf;
  char* sn;
  char bufn[100];
  sn = bufn;
  int next;

  tr->showTrajectoryTransitions ();

  /* add headers. */
  graph.append("digraph " + name + " {\n\trankdir=LR;\n\tranksep=1;\n\tsize=\"10,5\"\n\torientation=land;\n\t");

  /* add components' names. */
  graph.append("\n\t{\n\tnode [shape=plaintext];\n\t{rank=same; time;");
  for (hash_map<const char*, Comp*, hash<const char*>, eqstr>::iterator it0 = (*tr->traj)[tr->low_time]->comps->begin(); 
       it0 != (*tr->traj)[tr->low_time]->comps->end(); it0++)
    graph.append(" \"" + std::string((*it0).first) + "\";");
  graph.append("}\n\t}\n\t");

  
  /* create the time line. */
  graph.append("{\n\t\tnode [shape=plaintext];\n\t");
  for (int i=tr->low_time; i<=tr->getCurrentTime(); i++) {
    sprintf(s, "%i [label = \"%i (%g)\"];\n\t", i, i, tr->getTimeCorr (i));
    graph.append(std::string(s));
  }
  
  graph.append("\ttime");
  for (int i=tr->low_time; i<=tr->getCurrentTime(); i++) {
      sprintf (s, "%i", i);
      graph.append(" -> " + std::string(s));
  }

  /* add nodes. */
  //graph.append("\n\tnode [shape=box, fontsize=10];\n\t");
  //graph.append("\n\tratio=fill;");
  graph.append("}\n\tnode [ shape=ellipse, style=filled ];\n\t");
  for (hash_map<int, State*, hash<int> >::iterator it1 = tr->traj->begin(); it1 != tr->traj->end(); it1++) {
    sprintf(s, "%i", (*it1).first);
    graph.append("{rank = same; " + std::string(s) + ";");
    for (hash_map<const char*, TrajEl*, hash<const char*>, eqstr>::iterator it2 = (*it1).second->state->begin(); 
	 it2 != (*it1).second->state->end(); it2++) {
      if ((*it2).second->getVarType() == "Mode" && (*it2).second->getLabel() == "true") {
	graph.append(" " + makeNodeNameWithLabel((*it2).second->at, std::string(s)));
	//graph.append(" " + makeNodeName((*it2).second->at, string(s)) + ";");
      }    
    }
    s = 0; s = buf;
    graph.append("}\n\t");
  }

  /* connect nodes. */
  for (hash_map<int, State*, hash<int> >::iterator it3 = tr->traj->begin(); it3 != tr->traj->end(); it3++) {
    sprintf(s, "%i", (*it3).first);
    for (hash_map<const char*, TrajEl*, hash<const char*>, eqstr>::iterator it4 = (*it3).second->state->begin(); 
	 it4 != (*it3).second->state->end(); it4++) {
      if ((*it4).second->getVarType() == "Mode" && (*it4).second->getLabel() == "true") {
	/* link to components names.*/
	if ((*it3).first == tr->low_time) {
	  graph.append((*it4).second->at->getVarComp()->name 
		       + " -> " + makeNodeName((*it4).second->at, std::string(s)) + " [ style = invis ];\n\t");
	}
	next = tr->getNext((*it4).second->at->getStateVar(), (*it3).first);
	if (next != 0) {
	  sprintf(sn, "%i", next);
	  for (map<int, Transition*>::iterator it5 = (*it4).second->nextTrans->begin(); it5 != (*it4).second->nextTrans->end(); it5++) {
	    if ((*it5).second->post_at != NULL) {
	      graph.append(makeNodeName((*it5).second->pre_at, std::string(s)) + " -> " + makeNodeName((*it5).second->post_at, std::string(sn)));
	      if ((*it5).second->getType() == "Fail")
		//graph.append(" [ style = dotted ]");
		graph.append(" [ color = coral ]");
	      else if ((*it5).second->idle)
		graph.append(" [style = dotted ]");
	      graph.append(";\n\t");
	    }
	  }
	}
      }
      sn = 0; sn = bufn;
    }
    s = 0; s = buf;
  }

  graph.append("}\n");
  /* create file : name + .dot */
  ofstream fgr(std::string(name + ".dot").c_str(), ios::trunc);
  fgr << graph;
}
